Package: ShortForm
Type: Package
Title: Automatic Short Form Creation
Version: 0.5.8
Date: 2026-05-06
Authors@R: c(person("Anthony", "Raborn", email = "anthony.w.raborn@gmail.com", 
    role = c("aut", "cre"), comment = c(ORCID = "0000-0002-8083-4739")), 
    person("Walter", "Leite", email = "Walter.Leite@coe.ufl.edu", role = "aut"))
Description: Performs automatic creation of short forms of scales with an 
    ant colony optimization algorithm and a Tabu search. As implemented in the 
    package, the ant colony algorithm randomly selects items to build a model of 
    a specified length, then updates the probability of item selection according 
    to the fit of the best model within each set of searches. The algorithm 
    continues until the same items are selected by multiple ants a given number 
    of times in a row. On the other hand, the Tabu search changes one parameter at
    a time to be either free, constrained, or fixed while keeping track of the
    changes made and putting changes that result in worse fit in a "tabu" list
    so that the algorithm does not revisit them for some number of searches. 
    See Leite, Huang, & Marcoulides (2008) <doi:10.1080/00273170802285743> for
    an applied example of the ant colony algorithm, and Marcoulides & Falk (2018)
    <doi:10.1080/10705511.2017.1409074> for an applied example of the Tabu search.
License: LGPL (>= 2.0, < 3) | Mozilla Public License
LazyData: TRUE
Suggests: knitr, MplusAutomation (>= 0.7), rmarkdown, testthat
Imports: lavaan (>= 0.5-22), stringr, methods, doSNOW, parallel,
        foreach, rlang
Depends: R (>= 4.1.0)
URL: https://github.com/AnthonyRaborn/ShortForm
BugReports: https://github.com/AnthonyRaborn/ShortForm/issues
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-06 05:13:18 UTC; araborn
Author: Anthony Raborn [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8083-4739>),
  Walter Leite [aut]
Maintainer: Anthony Raborn <anthony.w.raborn@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-06 07:20:45 UTC
Built: R 4.6.0; ; 2026-05-06 09:56:09 UTC; unix

Package: ShortForm
Type: Package
Title: Automatic Short Form Creation
Version: 0.5.8
Date: 2026-05-06
Authors@R: c(person("Anthony", "Raborn", email = "anthony.w.raborn@gmail.com", 
    role = c("aut", "cre"), comment = c(ORCID = "0000-0002-8083-4739")), 
    person("Walter", "Leite", email = "Walter.Leite@coe.ufl.edu", role = "aut"))
Description: Performs automatic creation of short forms of scales with an 
    ant colony optimization algorithm and a Tabu search. As implemented in the 
    package, the ant colony algorithm randomly selects items to build a model of 
    a specified length, then updates the probability of item selection according 
    to the fit of the best model within each set of searches. The algorithm 
    continues until the same items are selected by multiple ants a given number 
    of times in a row. On the other hand, the Tabu search changes one parameter at
    a time to be either free, constrained, or fixed while keeping track of the
    changes made and putting changes that result in worse fit in a "tabu" list
    so that the algorithm does not revisit them for some number of searches. 
    See Leite, Huang, & Marcoulides (2008) <doi:10.1080/00273170802285743> for
    an applied example of the ant colony algorithm, and Marcoulides & Falk (2018)
    <doi:10.1080/10705511.2017.1409074> for an applied example of the Tabu search.
License: LGPL (>= 2.0, < 3) | Mozilla Public License
LazyData: TRUE
Suggests: knitr, MplusAutomation (>= 0.7), rmarkdown, testthat
Imports: lavaan (>= 0.5-22), stringr, methods, doSNOW, parallel,
        foreach, rlang
Depends: R (>= 4.1.0)
URL: https://github.com/AnthonyRaborn/ShortForm
BugReports: https://github.com/AnthonyRaborn/ShortForm/issues
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-06 05:13:18 UTC; araborn
Author: Anthony Raborn [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8083-4739>),
  Walter Leite [aut]
Maintainer: Anthony Raborn <anthony.w.raborn@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-06 07:20:45 UTC
Built: R 4.5.2; ; 2026-05-06 11:32:15 UTC; unix

# Generated by roxygen2: do not edit by hand

export(add.param)
export(antcolony.lavaan)
export(antcolony.mplus)
export(refit.model)
export(search.prep)
export(simulatedAnnealing)
export(tabu.sem)
export(tabuShortForm)
exportClasses(ACO)
exportClasses(SA)
exportClasses(TS)
exportClasses(modelCheck)
exportMethods(plot)
exportMethods(show)
exportMethods(summary)
import(lavaan)
import(utils)
importFrom(grDevices,hcl.colors)
importFrom(graphics,box)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(graphics,plot.new)
importFrom(graphics,polygon)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(methods,new)
importFrom(methods,show)
importFrom(rlang,.data)
importFrom(stats,runif)

#' Adds a parameter to the given search table. Checks whether parameter
#' is involved in any (in)equality constraints in a fitted lavaan model
#'
#' @param fitted.model fitted lavaan model
#' @param ptab search table
#' @param syntax model.syntax specifying the parameter to add to the current table
#' @param nullval optional numeric value specifying what the parameter should be fixed to (when fixed)
#' @param free optional logical value specifying whether the parameter should initially be set free (or not)
#' @param block optional numeric value specifying the group number to which the parameter corresponds
#'
#' @return A \code{data.frame} with lavaan-formatted parameter values.
#' @family Tabu Search
#' @import lavaan
#' @export
#'
#' @examples
#' \dontrun{
#' # load simulation data and select columns used in this example
#' data(simulated_test_data)
#' tabuData <- simulated_test_data[, c(1:10)]
#'
#' # specify an improper model (improper because data is unidimensional)
#' tabuModel <- "
#' Ability =~ Item1 + Item2 + Item3 + Item4
#' FakeAbility =~ Item5 + Item6 + Item7 + Item8
#' Ability ~ Outcome
#' FakeAbility ~ 0*Outcome"
#'
#' # run the initial misspecified model for Tabu
#'
#' init.model <- lavaan::lavaan(
#'   model = tabuModel, data = tabuData,
#'   auto.var = TRUE, auto.fix.first = FALSE, std.lv = TRUE, auto.cov.lv.x = TRUE
#' )
#'
#' # Use search.prep to prepare for the Tabu search
#' ptab <- search.prep(fitted.model = init.model, loadings = TRUE, fcov = TRUE, errors = FALSE)
#'
#' # add an additional (mispecified) parameter
#' additional.param <- "Item1 ~~ 0.5*Item3"
#' ptab <- add.param(fitted.model = init.model, ptab = ptab, syntax = additional.param)
#'
#' # Perform Tabu Search
#' trial <- tabu.sem(init.model = init.model, ptab = ptab, obj = AIC, niter = 2, tabu.size = 5)
#' }
#'
#' @author Carl F. Falk
#' @references \doi{10.1080/10705511.2017.1409074}

add.param <- function(fitted.model, ptab, syntax, nullval = NULL, free = NULL, block = NULL) {
  newpar <- lavaan::lavaanify(syntax)

  if (!is.null(block)) {
    newpar$block <- block
  }
  if (!is.null(free)) {
    newpar$free <- ifelse(free, 1, 0)
  }
  if (!is.null(nullval)) {
    newpar$nullval <- nullval
  } else {
    newpar$nullval <- newpar$ustart
  }

  newpar <- newpar[, c("lhs", "op", "rhs", "block", "free", "label", "nullval")]

  # Check whether newpar includes more than one row
  if (nrow(newpar) > 1) {
    newpar <- newpar[1, ]
    warning("lavaanify() results in more than one row of parameters, adding only the first row")
  }

  # Check whether parameter already exists in table
  matches <- par.matches(ptab, paste(newpar$lhs, newpar$op, newpar$rhs, sep = ""), block = block)
  if (any(matches)) {
    stop("Parameter already found in table")
  }

  # Check to see if parameter is part of constraint in existing model
  flag <- check.const(fitted.model, newpar)
  if (flag) {
    warning("Some parameters added to this table may be involved in (in)equality constraints")
  }

  # Add newpar to existing table
  ptab <- rbind(ptab, newpar)

  return(ptab)
}

#' Given a fitted lavaan model and a search table, refits the model using the search
#' table as specifying what changes should be done (parameters fixed/freed).
#'
#' This is not meant to be called explicitly as \link{tabu.sem} uses this
#' internally for model refitting.
#'
#' @param fitted.model fitted model of class lavaan
#' @param ptab search table
#'
#' @return An object of class \code{lavaan} if the new model fits, or an object of class \code{try-error} if the model update fails.
#' @family Tabu Search
#' @import lavaan
#' @export
#'
#' @author Carl F. Falk
#' @references \doi{10.1080/10705511.2017.1409074}

refit.model <- function(fitted.model, ptab) {

  # Extract parameter table
  tab <- lavaan::parTable(fitted.model)

  # Expand to get full table
  fulltab <- lav_partable_full(tab)
  mergetab <- lavaan::lav_partable_merge(tab, fulltab, remove.duplicated = TRUE, warn = FALSE)

  # Modify parameter table based on new input matrix

  # Obtain indices for matching
  mergetab$labels <- paste(mergetab$lhs, mergetab$op, mergetab$rhs, sep = "")
  ptab$labels <- paste(ptab$lhs, ptab$op, ptab$rhs, sep = "")
  midx <- vector("numeric")
  nidx <- vector("numeric")
  for (j in 1:nrow(ptab)) {
    idx <- which(ptab$labels[j] == mergetab$labels & ptab$block[j] == mergetab$block)
    if (length(idx) > 0) {
      midx <- c(midx, idx)
      nidx <- c(nidx, j)
    }
  }

  # Replace free parameters
  mergetab$free[midx] <- ptab$free[nidx]

  # If some parameters are now fixed; what are they fixed to?<-function
  mergetab$ustart[mergetab$free != 0] <- NA
  mergetab[midx, ]$ustart[mergetab[midx, ]$free == 0] <- ptab[nidx, ]$nullval[ptab[nidx, ]$free == 0]

  # Get rid of old parameter estimates and starting values
  mergetab$est <- NULL
  mergetab$se <- NULL
  mergetab$start <- NULL
  mergetab$labels <- NULL

  # Re-order values in mergetab
  lvnames <- lavaan::lavNames(mergetab, "lv")
  ovnames <- lavaan::lavNames(mergetab, "ov")
  mergetab <- mergetab[order(
    mergetab$block, # by group (or block)
    is.na(match(mergetab$lhs, c(lvnames))), # put latent vars first
    mergetab$op, # by op w/in lv and ov sections
    match(mergetab$lhs, c(lvnames, ovnames)), # then by lv and ov names on lhs
    match(mergetab$rhs, c(lvnames, ovnames))
  ), ] # then by same on rhs

  # Refit model
  prevmodel <- as.list(fitted.model@call)
  prevmodel$model <- mergetab
  newmod <- try(do.call(eval(parse(text = "lavaan::lavaan")), prevmodel[-1]), silent = TRUE)

  return(newmod)
}

setClassUnion("matrixORlist", c("matrix", "list"))
setClassUnion("lavaanORNULL", c("lavaan", "NULL"))

#' An S4 class for the modelCheck object
#'
#' @slot model.output A `lavaan` object.
#' @slot warnings A `character` vector of any warnings.
#' @slot errors A `character` vector of any errors.
#' @slot model.syntax A `character` vector of the modelCheck model syntax.
#'
#' @importFrom methods new show
#'
#' @return An S4 object of class `ACO`.
#' @export
#'
setClass('modelCheck',
         slots =
           list(
             model.output = 'lavaanORNULL',
             warnings = 'character',
             errors = 'character',
             model.syntax = 'character'
           )
)

#' An S4 class for the Simulated Annealing Algorithm
#'
#' @slot function_call The original function call.
#' @slot chains The number of chains used.
#' @slot chain_results A `matrix` (for multiple chains) or a `list` (for a single chain) of the chain results.
#' @slot all_fit A summary `vector` indicating the model fit results for
#' each iteration.
#' @slot best_fit The best model fit result using the selected `fitStatistic`. 
#' @slot best_model A `modelCheck` object of the final solution.
#' @slot best_syntax A `character` vector of the final solution model syntax.
#' @slot runtime A `difftime` object of the total run time of the function.
#'
#' @importFrom methods new show
#'
#' @return An S4 object of class `SA`.
#' @export
#'
setClass('SA',
         slots =
           list(
             function_call = 'call',
             chains = 'numeric',
             chain_results = 'matrixORlist',
             all_fit = 'vector',
             best_fit = 'numeric',
             best_model = 'modelCheck',
             best_syntax = 'character',
             runtime = 'ANY'
           )
)

#' Print method for class `SA`
#' 
#' @param object An S4 object of class `SA`.
#' 
#' @export
setMethod('show',
          signature = 'SA',
          definition = function(object) {
            line0 = c("Algorithm: Simulated Annealing")
            line1 = paste0(
              "Total Run Time: ",
              round(object@runtime[[1]], 3), 
              " ", 
              attr(object@runtime, "units"),
              " using ",
              object@chains,
              " chains. \n"
            )
            line2 = suppressWarnings(
              stringr::str_wrap(
                as.vector(c("Function call:\n", object@function_call, "\n")), 
                exdent = 2
              )
            )
            line3 = paste0(
              stringr::str_wrap(
                c("Final Model Syntax:", 
                  unlist(strsplit(object@best_syntax, '\n'))), 
                exdent = 2), 
              collapse = "\n"
            )
            to_console = paste0(c(line0, line1, line2, line3), collapse = "\n")
            cat(to_console)
          }
)

#' Plot method for class `SA`
#' 
#' @param x,y An S4 object of class `SA`.
#' @param ... Not used.
#' 
#' @export
#' @importFrom graphics legend lines par
setMethod('plot',
          signature = 'SA',
          definition = function(x, y, ...) {

            temp <- as.data.frame(x@all_fit)
            
            availableColors <-
              c("black", "#DF536B", "#61D04F", "#2297E6", "#28E2E5", "#D03AF5", "#EEC21F", "gray62")
            
            par(oma = c(0,0,0,5))
            plot(
              temp[,1], 
              col = availableColors[1],
              type = 'l', 
              ylim = c(min(temp, na.rm = T), max(temp, na.rm = T)),
              bty = "L",
              main = "Model Fit Results per Chain",
              ylab = "Fit Statistic",
              xlab = "Chain Step"
            )
            
            if (2 <= ncol(temp) & ncol(temp) <=8 ) {
              
              for (i in 2:ncol(temp)) {
                lines(temp[,i], col = availableColors[i])
              }
              legend(
                par()$usr[2], par()$usr[4],
                legend = paste0("Chain ", 1:ncol(temp)), 
                col = availableColors[1:ncol(temp)],
                lty = 1,
                bty = 'n',
                xpd = NA)
              }
            }
          )

#' Summary method for class `SA`
#' 
#' @param object An S4 object of class `SA`.
#' 
#' @export
setMethod('summary',
          signature = 'SA',
          definition = function(object) {
            line0 = c("Algorithm: Simulated Annealing")
            line1 = paste0(
              "Total Run Time: ",
              round(object@runtime[[1]], 3), 
              " ", 
              attr(object@runtime, "units"),
              "\n"
            )
            line2 = c(capture.output(print(object@best_model@model.output)), "\n")
            line3 = paste0(
              stringr::str_wrap(
                c("\nFinal Model Syntax:", 
                  unlist(strsplit(object@best_syntax, "\n"))), 
                exdent = 2), 
              collapse = "\n"
            )
            to_console = paste0(c(line0, line1, line2, line3), collapse = "\n")
            cat(to_console)
          }
)

#' A series of internal functions used by the Tabu search.
#' Directly accessing these functions is not advised unless you know what you are doing.
#'
#' @keywords internal
#'

# Internal function used to match parameters in label against
# those in a search table (ptab)
par.matches <- function(ptab, label, block = NULL) {

  # Collapse whitespace in label (if any)
  label <- gsub("[[:space:]]", "", label)

  # Look through lhsoprhs
  lhsoprhs <- paste(ptab$lhs, ptab$op, ptab$rhs, sep = "")

  # Look through labels
  matches <- label == lhsoprhs | label == ptab$label

  # Check against block
  if (!is.null(block)) {
    matches <- matches & ptab$block == block
  }
  return(matches)
}

# Internal function that modifies search table
manip.ptab <- function(ptab, label, task, nullval = NULL, block = NULL) {
  if (!task %in% c("remove", "free", "fix", "nullval")) {
    stop()
  }

  matches <- par.matches(ptab, label, block = block)

  nmatches <- sum(matches)
  if (nmatches > 1) {
    warning("More than one match for parameter found")
  } else if (nmatches < 1) {
    warning("No matches for parameter found")
  }

  if (task == "remove") {
    # Remove from table
    ptab <- ptab[!matches, ]
  } else if (task == "free") {
    # Set free
    ptab$free[matches] <- 1
  } else if (task == "fix") {
    # Fix
    ptab$free[matches] <- 0

    if (!is.null(nullval)) {
      ptab$nullval[matches] <- nullval
    }
  } else if (task == "nullval") {
    # Change nullval
    ptab$nullval[matches] <- nullval
  }

  return(ptab)
}

# Given a search table and a label, sets the parameter free in the search table
#
# ptab - search table
# label - combination of lhs, op, and rhs as would appear in lavaan model.syntax
#    or a parameter label from the label column of a lavaan parameter table
# block - optional numeric value specifying the group number to which the parameter corresponds
free.param <- function(ptab, label, block = NULL) {
  ptab <- manip.ptab(ptab, label, "free", block = block)
  return(ptab)
}

# Given a search table and a label, sets the parameter fixed in the search table
#
# ptab - search table
# label - combination of lhs, op, and rhs as would appear in lavaan model.syntax
#    or a parameter label from the label column of a lavaan parameter table
# nullval - optional numeric value specifying what the parameter should be set to when fixed (existing value in ptab is otherwise used)
# block - optional numeric value specifying the group number to which the parameter corresponds
fix.param <- function(ptab, label, nullval = NULL, block = NULL) {
  ptab <- manip.ptab(ptab, label, "fix", nullval = nullval, block = block)
  return(ptab)
}

# Given a search table and a label, removes the parameter from the search table
#
# ptab - search table
# label - combination of lhs, op, and rhs as would appear in lavaan model.syntax
#    or a parameter label from the label column of a lavaan parameter table
# block - optional numeric value specifying the group number to which the parameter corresponds
rm.param <- function(ptab, label, block = NULL) {
  ptab <- manip.ptab(ptab, label, "remove", block = block)
  return(ptab)
}

# Given a search table and a label, sets the value that the parameter would be
# fixed to if the parameter is fixed
#
# ptab - search table
# label - combination of lhs, op, and rhs as would appear in lavaan model.syntax
#    or a parameter label from the label column of a lavaan parameter table
# nullval - numeric value specifying what the parameter should be set to when fixed
# block - optional numeric value specifying the group number to which the parameter corresponds
nullval.param <- function(ptab, label, nullval, block = NULL) {
  ptab <- manip.ptab(ptab, label, "nullval", nullval = nullval, block = block)
  return(ptab)
}

# Given a fitted lavaan model and a search table, checks whether any of the
# parameters in the search table are involved in (in)equality constraints
check.const <- function(fitted.model, pars) {
  ptab <- lavaan::parTable(fitted.model)
  const <- ptab[ptab$op == "==" | ptab$op == ">" | ptab$op == "<", ]
  if (nrow(const) > 0) {
    plabels <- c(pars$label, pars$plabel)
    plabels <- plabels[plabels != ""]
    flag <- FALSE
    for (j in 1:nrow(const)) {
      var.lhs <- all.vars(parse(file = "", text = const$lhs[j]))
      var.rhs <- all.vars(parse(file = "", text = const$rhs[j]))
      if (any(plabels %in% c(var.lhs, var.rhs))) {
        flag <- TRUE
      }
    }
  } else {
    flag <- FALSE
  }
  return(flag)
}

#' A function to implement the ant colony optimization algorithm for short form
#' specification searches, either using MPlus directly via
#' \code{\link[base]{system}} calls or using Mplus indirectly with the package
#' \link[MplusAutomation]{MplusAutomation}.
#'
#' @description The Ant Colony Optimization (ACO) algorithm (Dorigo & Stutzle,
#' 2004) can produce short forms of scales that are optimized with respect to
#' characteristics selected by the developer, such as model fit and predictive
#' relationships with other variables. The algorithm is based on the foraging
#' behavior of a group of ants, which start searching for food in a variety of
#' directions and then eventually all ants converge to the shortest distance to
#' the food source. This behavior occurs because ants leave a pheromone trail
#' behind as they search for food and ants in shorter paths leave stronger
#' pheromone trails, which are detected by other ants and that will lead them to
#' follow the shortest trail.
#'
#' @details This function sends a specified number of ants
#' per iteration, which randomly select items to build a model, then evaluates
#' the model based on pheromone levels. The pheromone levels are updated after
#' each iteration according to the best-fitting model of that iteration. The
#' algorithm's stopping rule is to end the search when a certain solution is the
#' same for a given number of ants in a row. When constructing the mplus dataset
#' and when \code{Mplus.Automation=FALSE}, make sure that items in 'categorical
#' are' and 'usevariables' are specifications that take the same number of lines
#' per short form.
#'
#' PREPARATORY STEPS: For the ACO algorithm implementation for short
#' for selection, the following decisions are needed:
#'
#' 1. Determine the target size for the short form.
#'
#' 2. Determine which characteristics should be optimized.
#'
#' 3. Define how the pheromone level will be computed: This is a function of the
#' characteristics of the short form that will be optimized. In Leite, Huang and
#' Marcoulides (2008), the pheromone level was zero if model fit indices did
#' not meet Hu and Bentler's (1999) suggested thresholds, and equal to the sum of
#' path coefficients of a predictor variable if model fit indices met thresholds.
#' Currently, the package only implements pheromone calculation based on
#' regression coefficients or variance explained, with user-selected model fit
#' index thresholds.
#'
#' 4. Define how many short forms should be evaluated before the best-so-far
#' pheromone level is examined. Leite, Huang and Marcoulides (2008) used 10 short
#' forms.
#'
#' 5. Define the percentage of pheromone evaporation, if any. Leite, Huang and
#' Marcoulides (2008) used 5\%.
#'
#' 6. Define convergence criterion. Leite, Huang and Marcoulides (2008) set the
#' algorithm to converge if the short form did not improve in 100 x number of
#' short forms in step 4.
#'
#' IMPLEMENTATION: Once these decisions are made, the ACO algorithm
#' selects short forms with the following steps:
#'
#' Step 1. All items are assigned an initial weight of 1.
#'
#' Step 2. A set of n short forms is selected by sampling with probability
#' proportional to the items' weights.
#'
#' Step 3. Fit latent variable model to the n short forms.
#'
#' Step 4. Calculate the pheromone levels for the n short forms. Define the
#' best-so-far pheromone level (if iteration 1) or compare the current best
#' pheromone from the set of n short forms to the best-so-far pheromone.
#'
#' Step 5. If the pheromone level of the best short form from step 4 exceeds the
#' best-so-far pheromone level, update the best-so-far pheromone level and add it
#' to the current weight of the items of the best short form.
#'
#' Step 6. Return to step 2 until convergence criterion is reached.
#'
#' @param ants A numeric value indicating the number of ants to send send (short
#'  forms to evaluate) per iteration. Default value is 20.
#' @param evaporation A numeric value which sets the percentage of the pheremone
#'  that is retained after evaporation between steps of the algorithm. Default
#'  value is 0.9, indicating 10% evaporation. Should be within the range of
#'  (0,1), exclusive.
#' @param mplus When \code{Mplus.Automation=FALSE}, this is a character value
#'  indicating the name of the MPlus input file without the file extension
#'  ".inp". If not in the current working directory, include the full file path
#'  where it is located. This file will be changed during the ant colony search,
#'  so it's suggested to make a backup of the original file before running the
#'  function. When \code{Mplus.Automation=TRUE}, this is an object of class
#'  \link[MplusAutomation]{mplusObject} created by
#'  \link[MplusAutomation]{MplusAutomation} and containing the initial model.
#' @param list.items A list containing one or more character vectors of item
#'  names for each factor, where each factor is a separate element of the list.
#'  The items should be input in the order in which the factors are input in
#'  \code{i.per.f} and \code{factors}.
#' @param full A numeric value indicating the total number of unique items in the
#'  test or scale.
#' @param i.per.f A vector with number of items per factor (e.g. target number),
#'  in the same order of \code{list.items} and \code{factors}.
#' @param factors A character vector with the names of the factors in the same
#'  order of \code{list.items} and \code{i.per.f}.
#' @param steps A numeric value that sets the stopping rule, which is the number
#'  of ants in a row for which the model does not change.
#' @param max.run The maximum number of ants to run before the algorithm stops.
#'  This includes failed iterations as well. Default is 1000.
#' @param resultfile A character vector containing the file path where the MPlus
#'  results for the current ant model is saved. If the file is not in the
#'  current working directery, the full path must be specified. Not used when
#'  \code{Mplus.Automation=FALSE}.
#' @param min.CFI A numeric value indicating the minimum CFI for "acceptable"
#'  model fit. Models with CFI less than this value are automatically rejected.
#'  Default is 0.95.
#' @param min.TLI A numeric value indicating the minimum TLI for "acceptable"
#'  model fit. Models with TLI less than this value are automatically rejected.
#'  Default is 0.95.
#' @param max.RMSEA A numeric value indicating the maximum RMSEA for "acceptable"
#'  model fit. Models with RMSEA greater than this value are automatically
#'  rejected. Default is 0.06
#' @param summaryfile A character vector containing the name of the summary file
#'  generated. A .txt file is suggested. Default is "summary.txt" and writes
#'  into the current working directory. This file writes a line for each ant
#'  within each step and includes (a) a vector of a 0/1 value for each item
#'  indicating whether the item was selected by that ant, (b) the run number,
#'  (c) the count number, (d) the ant number, and (e) the current pheromone
#'  level.
#' @param feedbackfile A character vector containing the name of the feedback
#'  file generated. An .html file is suggested. Default is "iteration.html" and
#'  writes into the current working directory. This file saves the result of
#'  each run, which includes (a) the run number, (b) the count number, (c) the
#'  ant number, (d) the step number (if the current run is successful) or
#'  "Failure" (if the current run is unsuccessful), and for successful runs (f)
#'  the value of CFI, TLI, and RMSEA fit indices, the average of the gammas
#'  (standardized regression coefficients), and the overall variance explained
#'  of the current run.
#' @param loc.gammas A numeric vector with the line numbers where the regression
#'  coefficients of the MIMIC model start and end (locations). Not used with
#'  \code{Mplus.Automation=TRUE}
#' @param loc.variances A numeric vector with the line numbers of the residual
#'  variances of the latent factors. Not used with \code{Mplus.Automation=TRUE}
#' @param predictors Character vector with names of predictor variables, if any.
#' @param var.predictors A numeric vector with variances of the predictor(s), if
#'  any. Not used with \code{Mplus.Automation=TRUE}
#' @param Mplus.Automation Logical. If \code{TRUE}, uses the
#'  \code{MplusAutomation} package to modify the model as the algorithm
#'  procedes. The "mplus" option will then be used as Defaults to \code{FALSE}
#'  as it is in the process of being built.
#' @param dataOut A character vector specifying the location and name of the data
#'  file generated by \code{MplusAutomation} for each iteration of the
#'  algorithm. Default is "tempData.dat" and saves to the current working
#'  directory. When specifying the name, be sure to use a data format that Mplus
#'  can read. You must change the working directory to the location in which
#'  this file will be saved. Only used when \code{Mplus.Automation=TRUE}.
#' @param modelOut A character vector specifying the location and name of the
#'  Mplus model file generated by \code{MplusAutomation} for each iteration of
#'  the algorithm. Default is "tempModel.inp" and saves to the current working
#'  directory. When specifying the name of the model file, it must be a ".inp"
#'  extension. You must change the working directory to the location in which
#'  this file will be saved. Only used when \code{Mplus.Automation=TRUE}.
#' @return A named matrix containing final model's best RMSEA, CFI, and TLI
#'  values, the final pheromone level (the mean of the standardized regression
#'  coefficients (gammas)), and a series of 0/1 values indicating the items
#'  selected in the final solution.
#' @family Ant Colony Algorithms
#' @seealso \code{\link{antcolony.lavaan}}
#' @examples
#' \dontrun{
#' # use MplusAutomation to find a 15-item short form of a simulated 56-item unidimensional test
#' # first, create the list of the items by the factors
#' # in this case, all items load onto the general 'Ability' factor
#' list.items <- list(c(
#'   "Item1", "Item2", "Item3", "Item4", "Item5",
#'   "Item6", "Item7", "Item8", "Item9", "Item10",
#'   "Item11", "Item12", "Item13", "Item14", "Item15",
#'   "Item16", "Item17", "Item18", "Item19", "Item20",
#'   "Item21", "Item22", "Item23", "Item24", "Item25",
#'   "Item26", "Item27", "Item28", "Item29", "Item30",
#'   "Item31", "Item32", "Item33", "Item34", "Item35",
#'   "Item36", "Item37", "Item38", "Item39", "Item40",
#'   "Item41", "Item42", "Item43", "Item44", "Item45",
#'   "Item46", "Item47", "Item48", "Item49", "Item50",
#'   "Item51", "Item52", "Item53", "Item54", "Item55",
#'   "Item56"
#' ))
#' # then, load the data
#' data(simulated_test_data)
#'
#' # Create the mplusObject with MplusAutomation
#' # notice the explicit call of each item, instead of the shorthand "Item1-Item56"
#' initial.MplusAutomation.model <- MplusAutomation::mplusObject(
#'   TITLE = "Trial ACO MpluAutomation with FERA 2016 Data;",
#'   MODEL = "Ability BY Item1 Item2 Item3 Item4 Item5
#'   Item6 Item7 Item8 Item9 Item10 Item11 Item12
#'   Item13 Item14 Item15 Item16 Item17 Item18
#'   Item19 Item20 Item21 Item22 Item23 Item24
#'   Item25 Item26 Item27 Item28 Item29 Item30
#'   Item31 Item32 Item33 Item34 Item35 Item36
#'   Item37 Item38 Item39 Item40 Item41 Item42
#'   Item43 Item44 Item45 Item46 Item47 Item48
#'   Item49 Item50 Item51 Item52 Item53 Item54
#'   Item55 Item56;",
#'   ANALYSIS = "ESTIMATOR = WLSMV;",
#'   VARIABLE = "CATEGORICAL = Item1 Item2 Item3 Item4 Item5
#'   Item6 Item7 Item8 Item9 Item10 Item11 Item12
#'   Item13 Item14 Item15 Item16 Item17 Item18
#'   Item19 Item20 Item21 Item22 Item23 Item24
#'   Item25 Item26 Item27 Item28 Item29 Item30
#'   Item31 Item32 Item33 Item34 Item35 Item36
#'   Item37 Item38 Item39 Item40 Item41 Item42
#'   Item43 Item44 Item45 Item46 Item47 Item48
#'   Item49 Item50 Item51 Item52 Item53 Item54
#'   Item55 Item56;",
#'   OUTPUT = "stdyx;",
#'   rdata = simulated_test_data
#' )
#'
#' # finally, call the function with some minor changes to the default values.
#' abilityShortForm <- antcolony.mplus(
#'   ants = 3, evaporation = 0.7,
#'   mplus = initial.MplusAutomation.model, list.items = list.items, full = 56,
#'   i.per.f = 15, factors = "Ability", steps = 3, max.run = 50, resultfile = NULL,
#'   summaryfile = "C:/Users/lordmaxwell/Desktop/summary.txt",
#'   min.CFI = 0.95, min.TLI = 0.95, max.RMSEA = 0.06,
#'   feedbackfile = "C:/Users/lordmaxwell/Desktop/iteration.html", Mplus.Automation = TRUE,
#'   dataOut = "exampleModel.dat",
#'   modelOut = "exampleModel.inp"
#' )
#' }
#' @export
#' @author Walter Leite; Anthony W Raborn, \email{anthony.w.raborn@@gmail.com}
#' @references \doi{10.1080/00273170802285743}

antcolony.mplus <- function(ants = 20, evaporation = 0.95, mplus = NULL, list.items = NULL, full = NULL,
                            i.per.f = NULL, factors = NULL, steps = 50, max.run = 1000,
                            resultfile = NULL, summaryfile = "summary.txt",
                            min.CFI = 0.95, min.TLI = 0.95, max.RMSEA = 0.06,
                            feedbackfile = "iteration.html", loc.gammas,
                            loc.variances, predictors, var.predictors,
                            Mplus.Automation = FALSE, dataOut = "tempModel.dat", modelOut = "tempModel.inp") {

  # create initial, empty summaryfile to be used
  write(x = "", file = summaryfile)

  # creates the table of initial pheromone levels.
  include <- rep(2, full)
  # puts initial best solution (all items selected).
  best.so.far.solution <- include
  # starts counting the iterations
  count <- 1
  # starts counting continuous runs regardless of result.
  run <- 1
  # creates a vector with all items.
  item.vector <- unlist(list.items, use.names = F)
  # defines initial best so far (overall) pheromone
  best.so.far.pheromone <- 0
  # defines initial best pheromone for the current trial of n ants.
  best.pheromone <- 0
  # defines initial solutions.
  previous.solution <- include
  step <- 0
  # creates selected.items list
  selected.items <- list()

  # check whether to use MplusAutomation or to call Mplus directly through R
  if (Mplus.Automation == FALSE) {

    # reads mplus syntax.
    # setwd("C:/Program Files/Mplus")
    input <- scan(file = paste(mplus, ".inp", sep = ""), what = "character", sep = "\n", quote = NULL)

    # creates a list to store factors.
    selected.items <- list.items

    # starts loop through iterations.
    while (count <= steps) {
      if (run <= max.run) {

        # sends a number of ants per time.
        ant <- 1
        while (ant <= ants) {
          print(paste0("Count Number ", count, " and Ant Number", ant, " and Run Number ", run, "."))

          # selects items for all factors.
          for (factor in 1:length(list.items)) {

            # selects the items for a short form for the factor
            positions <- is.element(item.vector, list.items[[factor]])
            prob <- include[positions] / sum(include[positions])


            items <- sample(list.items[[factor]], size = i.per.f[factor], replace = F, prob)

            # stores selected items.
            selected.items[[factor]] <- items


            # replaces the mplus syntax for factor specification.
            # make sure the factor name provided is the same as the one in the MPLUS file.
            factor.position <- grep(paste(factors[factor], "BY"), input, ignore.case = T)
            input[factor.position] <- paste(c(factors[factor], "BY", items, ";"), collapse = " ")

            # finishes loop
          }

          selected.items <- lapply(selected.items, sort)
          selected.vector <- unlist(selected.items, use.names = F)
          select.indicator <- is.element(item.vector, selected.vector)
          notselect.indicator <- (select.indicator == FALSE)

          # MODIFY MPLUS SYNTAX.
          # when constructing mplus file, leave two lines blank after the usevariables command.


          # make sure there are at least three lines of space between them.
          # defines variables to be used.
          position1 <- grep("usevariables", input, ignore.case = T)
          input[position1] <- paste("usevariables are ", predictors)
          input[position1 + 1] <- paste(c(selected.vector[1:round(sum(i.per.f) / 2, 0)]),
            collapse = " "
          )
          input[position1 + 2] <- paste(c(selected.vector[(round(sum(i.per.f) / 2, 0) + 1):sum(i.per.f)], ";"),
            collapse = " "
          )
          position2 <- grep("categorical", input, ignore.case = T)
          input[position2 + 1] <- paste(c(selected.vector[1:round(sum(i.per.f) / 2, 0)]),
            collapse = " "
          )
          input[position2 + 2] <- paste(c(selected.vector[(round(sum(i.per.f) / 2, 0) + 1):sum(i.per.f)], ";"),
            collapse = " "
          )


          # writes the mplus syntax for fitting the short form.
          write.table(input,
            file = paste(mplus, ".inp", sep = ""), quote = F,
            col.names = F, row.names = F
          )


          # Run MPLUS with the syntax file specified.
          system(paste("mplus.exe ", mplus, ".inp", sep = ""), wait = TRUE)
          outfile <- c(paste(mplus, ".out", sep = ""))

          # read the output file to verify the outcome of the analysis.
          output <- scan(file = outfile, what = character(), sep = "\n")
          # check if the outcome was a non-positive definite matrix.
          outcome1 <- grep("WARNING", output, ignore.case = T)
          # check if there was non-convergence.
          outcome0 <- grep("EXCEEDED", output, ignore.case = T)
          outcome2 <- grep("Model estimation did not terminate normally.  No results were saved", output, ignore.case = T)
          # set pheromone to zero if the there was non-convergence or a non-positive definite solution.
          if ((length(outcome1) == 1) || (length(outcome0) == 1) || (length(outcome2) == 1)) {
            pheromone <- 0

            # writes summary about non-convergence and non-positive definite.
            fit.info <- matrix(c(select.indicator, run, count, ant, 999, 999, round((include), 5)), 1, )
            write.table(fit.info,
              file = summaryfile, append = T,
              quote = F, sep = " ", row.names = F, col.names = F
            )

            # provide feedback about search.
            feedback <- c(paste("<h1>", run, "-", count, "-", ant, "-", step, "- Failure", "</h1>"))
            write(feedback, file = feedbackfile)

            # finishes if for non-convergent cases.
          } else {

            # The SAVE command at the MPLUS syntax specifies that results should be
            # saved to results.txt.
            # scans all results.
            results <- scan(resultfile)
            # read the CFI,TLI and RMSEA.
            CFI.row <- output[grep("CFI[ ]{1,}", output)]
            CFI <- as.numeric(regmatches(CFI.row, regexpr("[0-9\\.]{5}", CFI.row)))

            TLI.row <- output[grep("TLI[ ]{1,}", output)]
            TLI <- as.numeric(regmatches(TLI.row, regexpr("[0-9\\.]{5}", TLI.row)))

            RMSEA.row <- output[grep("RMSEA", output) + 1]
            RMSEA <- as.numeric(regmatches(RMSEA.row, regexpr("[0-9\\.]{5}", RMSEA.row)))

            # reads the regression coefficients (gammas)
            gammas <- results[loc.gammas]

            # reads the residual variances of the latent variables.
            res.variances <- results[loc.variances]


            # obtains the variance explained by the predictors using path tracing
            var.explained <- t(t(gammas)^2 * var.predictors)
            var.explained <- apply(var.explained, 1, sum)

            # obtains the total variance of the latent variable
            # residual variance + variance explained
            variances <- res.variances + var.explained

            # standardizes the gammas.
            std.gammas <- gammas / sqrt(variances)

            # saves summary information about the selected items and the RMSEA they generated.
            fit.info <- matrix(c(
              select.indicator, run, count, ant, CFI, TLI, RMSEA, mean(std.gammas),
              round(include, 2)
            ), 1, )

            write.table(fit.info,
              file = summaryfile, append = T,
              quote = F, sep = " ", row.names = F, col.names = F
            )

            # provide feedback about search.
            feedback <- c(paste(
              "<h1>", "run:", run, "count:", count, "ant:", ant, "step:", step, "<br>",
              "CFI:", CFI, "TLI:", TLI, "RMSEA:", RMSEA, "<br>",
              "GAMMA:", mean(std.gammas), "</h1>"
            ))
            write(feedback, file = feedbackfile)

            print(feedback)
            # implements fit requirement.
            if ((CFI < min.CFI) | (TLI < min.TLI) | (RMSEA > max.RMSEA)) {
              pheromone <- 0
            } else {
              # calculate pheromone (mean of standardized gammas).
              pheromone <- mean(std.gammas)
            }
            # adjusts count based on outcomes and selects best solution.
            if (pheromone >= best.pheromone) {

              # updates solution.
              best.solution <- select.indicator
              # updates best fit indices
              best.RMSEA <- RMSEA
              best.CFI <- CFI
              best.TLI <- TLI
              # updates best pheromone
              best.pheromone <- pheromone
            }

            # Move to next ant.
            ant <- ant + 1

            # end else clause for converged solutions
          }
          # ends loop through ants.
          run <- run + 1
        }

        # adjusts pheromone only if the current pheromone is best than the previous.
        if (best.pheromone > best.so.far.pheromone) {

          # implements pheromone evaporation.
          include <- include * evaporation

          # Adjusts the pheromone levels.
          include.pheromone <- best.solution * best.pheromone * (0.1 * run)


          # updates pheromone.
          include <- include + include.pheromone

          # updates best so far solution and pheromone, with corresponding RMSEA.
          best.so.far.solution <- best.solution
          best.so.far.pheromone <- best.pheromone
          best.so.far.RMSEA <- best.RMSEA
          best.so.far.CFI <- best.CFI
          best.so.far.TLI <- best.TLI
          # re-starts count.
          count <- 1

          # end if clause for pheromone adjustment.
        } else {

          # advances count.
          count <- count + 1


          # adds more pheromone to the best so far solution.
          # include.pheromone = best.so.far.solution * best.so.far.pheromone

          # updates pheromone.
          # include = include + include.pheromone

          # finish else clause.
        }

        # ends loop.
      } else {
        stop("Maximum number of runs reached.")
      }
    }
  } else if (Mplus.Automation == TRUE) {
    if (!requireNamespace("MplusAutomation", quietly = TRUE)) {
      stop("The `MplusAutomation` package is required to use this function. Please install `MplusAutomation`, then try to use this function again.")
    }

    # starts loop through iterations.
    while (count <= steps) {
      if (run <= max.run) {

        # sends a number of ants per time.
        ant <- 1
        while (ant <= ants) {
          print(paste0("Count Number ", count, " and Ant Number ", ant, " and Run Number ", run, "."))

          # selects items for all factors.
          for (factor in 1:length(list.items)) {
            # selects the items for a short form for the factor
            positions <- is.element(item.vector, list.items[[factor]])
            prob <- include[positions] / sum(include[positions])
            items <- sample(list.items[[factor]], size = i.per.f[factor], replace = F, prob)
            # stores selected items.
            selected.items[[factor]] <- items
            # finishes loop
          }

          # replaces the mplus syntax for factor specification.
          # make sure the factor name provided is the same as the one in the mplusObject!
          measurement.pattern <- c()
          for (i in 1:length(factors)) measurement.pattern[i] <- paste0("(", factors[i], " BY)[ A-z0-9]*")
          updated.model <- gsub("\\n", " ", mplus$MODEL)
          for (i in 1:length(factors)) updated.model <- gsub(measurement.pattern[i], paste(c(factors[i], "BY", selected.items[[i]]), collapse = " "), updated.model)
          updated.model <- gsub("(;)(\\s|$)", "\\1\n", updated.model)
          updated.model <- gsub("(.{1,80})(\\s|$)", "\\1\n", updated.model)
          updated.model <- gsub(";", ";\n", updated.model)

          # writes the mplus syntax for fitting the short form.
          new.input <- mplus
          new.input$MODEL <- updated.model

          factor.patterns <- c()
          for (i in 1:length(factors)) factor.patterns[i] <- paste0("(", factors[i], " BY|", factors[i], " ON| ON ", factors[i], ")", collapse = "")

          new.usevariables <- updated.model
          for (i in 1:length(factors)) new.usevariables <- gsub(factor.patterns[i], "", new.usevariables)
          new.usevariables <- gsub(";\\n", "", new.usevariables)
          new.usevariables <- gsub("\\n", " ", new.usevariables)
          new.usevariables <- unlist(lapply(strsplit(new.usevariables, " "), unique))
          new.usevariables <- new.usevariables[new.usevariables != ""]
          new.usevariables <- new.usevariables[new.usevariables != "ON"]
          new.usevariables <- new.usevariables[!(new.usevariables %in% factors)]
          new.input$usevariables <- new.usevariables

          outcome.vars <- unlist(regmatches(updated.model, gregexpr("[A-z0-9 ]{1,} (ON) [A-z0-9 ]{1,}", updated.model)))
          for (i in 1:length(factors)) outcome.vars <- gsub(factors[i], "", outcome.vars)
          outcome.vars <- gsub("( ON){1,}", "", outcome.vars)
          outcome.vars <- unique(gsub(" ", "", outcome.vars))
          if (length(outcome.vars) == 0) outcome.vars <- "none"
          dependent.vars <- new.usevariables[new.usevariables != outcome.vars]
          if (grepl(outcome.vars, mplus$VARIABLE)) {
            new.input$VARIABLE <- gsub("(.{1,80})(\\s|$)", "\\1\n", gsub("(?<= ){1}[A-z0-9 - \\n]{1,}", paste(c(outcome.vars, dependent.vars), collapse = " "), new.input$VARIABLE, perl = TRUE))
          } else {
            new.input$VARIABLE <- gsub("(.{1,80})(\\s|$)", "\\1\n", gsub("(?<= ){1}[A-z0-9 - \\n]{1,}", paste(dependent.vars, collapse = " "), new.input$VARIABLE, perl = TRUE))
          }

          selected.items <- lapply(selected.items, sort)
          selected.vector <- unlist(selected.items, use.names = F)
          select.indicator <- is.element(item.vector, selected.vector)
          notselect.indicator <- (select.indicator == FALSE)

          # Run MPLUS with the updated model just specified.
          current.ant.model <- MplusAutomation::mplusModeler(new.input, dataout = dataOut, modelout = modelOut, run = 1, writeData = "always", hashfile = FALSE)

          # check if the outcome was a non-positive definite matrix.
          output <- paste(c(unlist(current.ant.model$results$warnings), unlist(current.ant.model$results$errors)), collapse = " ")
          outcome1 <- grep("WARNING", output, ignore.case = T)
          # check if there was non-convergence.
          outcome0 <- grep("EXCEEDED", output, ignore.case = T)
          outcome2 <- grep("Model estimation did not terminate normally.  No results were saved", output, ignore.case = T)
          # set pheromone to zero if the there was non-convergence or a non-positive definite solution.
          if ((length(outcome1) == 1) || (length(outcome0) == 1) || (length(outcome2) == 1)) {
            pheromone <- 0

            # writes feedback about non-convergence and non-positive definite.
            fit.info <- matrix(c(select.indicator, run, count, ant, 999, 999, round((include), 5)), 1, )
            write.table(fit.info,
              file = summaryfile, append = T,
              quote = F, sep = " ", row.names = F, col.names = F
            )

            # provide feedback about search.
            feedback <- c(paste("<h1> Run", run, "- Count", count, "- Ant", ant, "- Step", step, "- Failure", "</h1>"))
            write(feedback, file = feedbackfile)

            # finishes if for non-convergent cases.
          } else {

            # read the CFI,TLI and RMSEA.
            CFI <- current.ant.model$results$summaries$CFI
            TLI <- current.ant.model$results$summaries$TLI
            RMSEA <- current.ant.model$results$summaries$RMSEA_Estimate

            # extract the standardized regression coefficients
            std.gammas <- current.ant.model$results$parameters$std[grep(".ON", current.ant.model$results$parameters$unstandardized$paramHeader), "est"]
            if (length(std.gammas) == 0) {
              std.gammas <- current.ant.model$results$parameters$r2$est
            }

            # saves information about the selected items and the RMSEA they generated.
            fit.info <- matrix(c(
              select.indicator, run, count, ant, CFI, TLI, RMSEA, mean(std.gammas),
              round(include, 2)
            ), 1, )

            write.table(fit.info,
              file = summaryfile, append = T,
              quote = F, sep = " ", row.names = F, col.names = F
            )

            # provide feedback about search.
            feedback <- c(paste(
              "<h1>", "run:", run, "count:", count, "ant:", ant, "step:", step, "<br>",
              "CFI:", CFI, "TLI:", TLI, "RMSEA:", RMSEA, "<br>",
              "GAMMA:", mean(std.gammas), "</h1>"
            ))
            write(feedback, file = feedbackfile)

            print(feedback)
            # implements fit requirement.
            if ((CFI < min.CFI) | (TLI < min.TLI) | (RMSEA > max.RMSEA)) {
              pheromone <- 0
            } else {
              # calculate pheromone (mean of standardized gammas).
              pheromone <- mean(std.gammas)
            }
            # adjusts count based on outcomes and selects best solution.
            if (pheromone >= best.pheromone) {

              # updates solution.
              best.solution <- select.indicator
              # updates best RMSEA.
              best.RMSEA <- RMSEA
              best.CFI <- CFI
              best.TLI <- TLI
              # updates best pheromone
              best.pheromone <- pheromone
            }

            # Move to next ant.
            ant <- ant + 1

            # end else clause for converged solutions
          }

          # ends loop through ants.
          run <- run + 1
        }

        # adjusts pheromone only if the current pheromone is best than the previous.
        if (best.pheromone > best.so.far.pheromone) {

          # implements pheromone evaporation.
          include <- include * evaporation

          # Adjusts the pheromone levels.
          include.pheromone <- best.solution * best.pheromone * (0.1 * run)


          # updates pheromone.
          include <- include + include.pheromone

          # updates best so far solution and pheromone, with corresponding RMSEA.
          best.so.far.solution <- best.solution
          best.so.far.pheromone <- best.pheromone
          best.so.far.RMSEA <- best.RMSEA
          best.so.far.CFI <- best.CFI
          best.so.far.TLI <- best.TLI
          # re-starts count.
          count <- 1

          # end if clause for pheromone adjustment.
        } else {

          # advances count.
          count <- count + 1


          # adds more pheromone to the best so far solution.
          include.pheromone <- best.so.far.solution * best.so.far.pheromone

          # updates pheromone.
          include <- include + include.pheromone
        }
      } else {
        stop("Maximum number of runs reached.")
      }
      # ends loop.
    }
  }


  final.solution <- matrix(c(best.so.far.RMSEA, best.so.far.CFI, best.so.far.TLI, best.so.far.pheromone, best.so.far.solution), 1, dimnames = list(NULL, c("RMSEA", "CFI", "TLI", "mean_gamma", item.vector)))

  # FINISH FUNCTION.
  return(final.solution)
}

linearTemperature <- function(currentStep, maxSteps) {
  currentTemp <- (maxSteps - (currentStep)) / maxSteps
}

quadraticTemperature <- function(currentStep, maxSteps) {
  currentTemp <- ((maxSteps - (currentStep)) / maxSteps)^2
}

logisticTemperature <- function(currentStep, maxSteps) {
  x <- 1:maxSteps
  x.new <- scale(x, center = T, scale = maxSteps / 12)
  currentTemp <- 1 / (1 + exp((x.new[(currentStep + 1)])))
}

currentStep <- 500
maxSteps <- 5000

test_that(
  "each temperature function returns the correct value given currentStep and maxSteps", {

    # linear temp should equal 4500/5000 = 0.9
    expect_equal(
      linearTemperature(
        currentStep = currentStep,
        maxSteps = maxSteps
        ),
      (1-500/5000)
    )

    # quadratic temp should equal (4500/5000)^2 = 0.81
    expect_equal(
      quadraticTemperature(
        currentStep = currentStep,
        maxSteps = maxSteps
      ),
      (1-500/5000)^2
    )

    # logistic temp should equal the formula given
    expect_equal(
      logisticTemperature(
        currentStep = currentStep,
        maxSteps = maxSteps
      ),
      1/(1 + exp((scale(1:maxSteps, center=T, scale = maxSteps/12)[currentStep+1])))
    )
  }
)

#' Given a fitted lavaan model, a search table, and an objective criterion,
#' performs a Tabu model specification search. Currently only supports
#' neighbors that are 1 move away from the current model.
#'
#' @param init.model initial fitted model of class lavaan
#' @param ptab search table (e.g., created by search.prep) that lists candidate
#'  parameters that can be modified as part of the search and how the parameters
#'  can be modified (fixed to what values)
#' @param obj objective function to be MINIMIZED. Any function that takes a
#' lavaan object as the sole argument and returns a numeric value can be used.
#' @param niter number of Tabu iterations to perform
#' @param tabu.size size of Tabu list
#'
#' @return A list with three elements: best.obj, the numerical value of the best (minimal) objective function achieved; best.mod, the final lavaan model, and best.binvec, a data.frame of the lavaan-formatted parameter table for the final model.
#' @export
#'
#' @examples
#' # load simulation data and select columns used in this example
#' data(simulated_test_data)
#' tabuData <- simulated_test_data[, c(1:10)]
#'
#' # specify an improper model (improper because data is unidimensional)
#' tabuModel <- "
#' Ability =~ Item1 + Item2 + Item3 + Item4
#' FakeAbility =~ Item5 + Item6 + Item7 + Item8
#' Ability ~ Outcome
#' FakeAbility ~ 0*Outcome"
#'
#' # run the initial misspecified model for Tabu
#'
#' init.model <- lavaan::lavaan(
#'   model = tabuModel, data = tabuData,
#'   auto.var = TRUE, auto.fix.first = FALSE, std.lv = TRUE, auto.cov.lv.x = TRUE
#' )
#'
#' # Use search.prep to prepare for the Tabu search
#' ptab <- search.prep(fitted.model = init.model, loadings = TRUE, fcov = TRUE, errors = FALSE)
#'
#' # Perform Tabu Search
#' trial <- tabu.sem(init.model = init.model, ptab = ptab, obj = AIC, niter = 2, tabu.size = 5)
#' @author Carl F. Falk
#' @references \doi{10.1080/10705511.2017.1409074}

tabu.sem <- function(init.model,
                     ptab,
                     obj,
                     niter = 30,
                     tabu.size = 5) {
  start.time = Sys.time()
  # Initialize objective function and best model
  best.obj <- all.obj <- current.obj <- obj(init.model)
  best.model <- current.model <- init.model
  best.binvec <- current.binvec <- ptab

  tabu.list <- vector("numeric")

  # Do iterations
  for (it in 1:niter) {
    cat(paste0("\rRunning iteration ", it, " of ", niter, ".   "))
    # Loop through all neighbors
    tmp.obj <- vector("numeric")
    tmp.mod <- list()
    tmp.vec <- list()
    for (j in 1:nrow(current.binvec)) {
      tmp.binvec <- current.binvec
      bin <- 1 - tmp.binvec$free[j]
      tmp.binvec$free[j] <- bin
      fitmodel <- refit.model(init.model, tmp.binvec)

      if (fitmodel@Fit@converged & !any(is.na(fitmodel@Fit@se))) {
        fit.val <- obj(fitmodel)
      } else {
        fit.val <- NA
      }

      tmp.obj <- c(tmp.obj, fit.val)
      tmp.mod[[j]] <- fitmodel
      tmp.vec[[j]] <- tmp.binvec
    }

    # Check which indices result in a valid objective function
    valid <- which(!is.na(tmp.obj))

    # Get just models not on Tabu list
    valid <- valid[!(valid %in% tabu.list)]

    # Out of valid models, pick model with best objective function value
    indx <- which.min(tmp.obj[valid])

    # Move current state to next model
    current.obj <- (tmp.obj[valid])[indx]
    all.obj <- c(all.obj, current.obj)
    current.mod <- (tmp.mod[valid])[[indx]]
    current.binvec <- (tmp.vec[valid])[[indx]]

    # Update Tabu list
    tabu.list <- c(valid[indx], tabu.list)
    if (length(tabu.list) > tabu.size) {
      tabu.list <- tabu.list[1:tabu.size]
    }

    # Update if the current model is better than the best model
    if (current.obj < best.obj) {
      best.obj <- current.obj
      best.mod <- current.mod
      best.binvec <- current.binvec
      tabu.list <- vector("numeric") # Clear Tabu list
    }
  }

  ret <-
    new("TS",
      function_call = match.call(),
      all_fit = all.obj,
      best_fit = best.obj,
      best_model = best.mod,
      best_syntax = best.binvec,
      runtime = Sys.time() - start.time,
      final_tabu_list = list(tabu.list)
  )

  ret
}

#' An adaptation of the simulated annealing algorithm for psychometric models.
#'
#' @description Simulated annealing mimics the physical process of annealing metals together. [Kirkpatrick et al. (1983)](https://www.science.org/doi/10.1126/science.220.4598.671) introduces this analogy and demonstrates its use; the implementation here follows this demonstration closely, with some modifications to make it better suited for psychometric models.
#'
#' @details \strong{Outline of the Pieces of the Simulated Annealing Algorithm}
#' * initialModel -- the initial, full form
#' * currentModel -- the model of the current step
#' * maxSteps -- the maximum number of steps (iterations)
#' * currentStep -- the current step
#' * currentTemp -- the current temperature. A function of the number of steps (such that temp = 0 at maxSteps), and values that control the shape of the overall temperature. A part of the function that determines the acceptance probability of newly -- generated models
#' * randomNeighbor -- a function that determines how the form is changed at each step. Should be able to change one or more parameters, and should have a way to control how many are changed.
#' * goal -- a function that determines the "goodness" of the currentModel. Typically in SA goodness is defined as minimization! Sometimes called an energy function
#' * selectionFunction -- a function that determines if a randomNeighbor change is accepted. Uses the goal function that determines the "goodness" of the currentModel and the "goodness" of the randomNeighbor, and the currentTemp to generate a probability of acceptance, then compares this probability to a Uniform(0,1) variable to determine if accepted or not. A standard version of this is:
#' ![](SA-goal.jpg)
#' (Kirkpatrick et al., 1983)
#' * bestModel -- the model with the best value of the goal function achieved so far
#' * bestGoal -- the best value of the goal function achieved so far
#' * restartCriteria -- if utilized, this would "restart" the SA process by changing currentModel to bestModel and continuing the process. Could be based on (1) the currentStep value, (2) the difference between goal(currentModel) and goal(bestModel), (3) randomness (i.e., could randomly restart, could randomly restart based on some values, etc), (4) other criteria.
#'
#' @param initialModel The initial model as a `character` vector with lavaan model.syntax.
#' @param originalData The original `data.frame` with variable names.
#' @param maxSteps The number of iterations for which the algorithm will run.
#' @param fitStatistic Either a single model fit statistic produced by lavaan, or a user-defined fit statistic function.
#' @param temperature Either an acceptable `character` value, or a user-defined temperature function. The acceptable values are "linear", "quadratic", or "logistic".
#' @param maximize Logical indicating if the goal is to maximize (`TRUE`) the fitStatistic for model selection.
#' @param Kirkpatrick Either `TRUE` to use Kirkpatrick et al. (1983) acceptance probability, or a user-defined function for accepting proposed models.
#' @param randomNeighbor Either `TRUE` to use the included function for randomNeighbor selection, or a user-defined function for creating random models.
#' @param lavaan.model.specs A `list` which contains the specifications for the
#'  lavaan model. The default values are the defaults for lavaan to perform a
#'  CFA. See \link[lavaan]{lavaan} for more details.
#' @param maxChanges An `integer` value greater than 1 setting the maximum number of parameters to change within randomNeighbor. When creating a short form, should be no greater than the smallest reduction in items loading on one factor; e.g., when reducing a 2-factor scale from 10 items on each factor to 8 items on the first and 6 items on the second, maxChanges should be no greater than 2.
#' @param restartCriteria Either "consecutive" to restart after maxConsecutiveSelection times with the same model chosen in a row, or a user-defined function.
#' @param maximumConsecutive A positive `integer` value used with restartCriteria.
#' @param maxItems When creating a short form, a `vector` of the number of items per factor you want the short form to contain. Defaults to `NULL`.
#' @param items A `character` vector of item names. Defaults to `NULL`. Ignored if `maxItems==FALSE`.
#' @param bifactor Logical. Indicates if the latent model is a bifactor model. If `TRUE`, assumes that the last latent variable in the provided model syntax is the bifactor (i.e., all of the retained items will be set to load on the last latent variable). Ignored if `maxItems==FALSE`.
#' @param setChains Numeric. Sets the number of parallel chains to run. Default to `1`, which also sets the algorithm to run serially (e.g., on a single processor). Values greater than `1` result in the chains running on parallel processes using the `doSNOW` and `foreach` packages.
#' @param shortForm Logical. Are you creating a short form (`TRUE`) or not (`FALSE`)? Default is `TRUE`.
#' @param ... Further arguments to be passed to other functions. Not implemented for any of the included functions.
#'
#' @return A named list: the 'bestModel' found, the 'bestFit', and 'allFit' values found by the algorithm.
#'
#' @examples
#' \dontrun{
#' data(exampleAntModel)
#' data(simulated_test_data)
#' trial1 <- simulatedAnnealing(
#'   initialModel = lavaan::cfa(
#'     model = exampleAntModel,
#'     data = simulated_test_data
#'   ),
#'   originalData = simulated_test_data, maxSteps = 3,
#'   fitStatistic = "rmsea", maximize = FALSE
#' )
#' summary(trial1) # shows the resulting model
#'
#' trial2 <- simulatedAnnealing(
#'   initialModel = exampleAntModel,
#'   originalData = simulated_test_data,
#'   maxSteps = 2, maxItems = 30, items = paste0("Item", 1:56)
#' )
#' summary(trial2) # shows the resulting model
#' }
#' @import lavaan utils
#' @export
#' @md

simulatedAnnealing <-
  function(initialModel,
           originalData,
           maxSteps,
           fitStatistic = "cfi",
           temperature = "linear",
           maximize = TRUE,
           Kirkpatrick = TRUE,
           randomNeighbor = TRUE,
           lavaan.model.specs = list(
             model.type = "cfa",
             auto.var = TRUE,
             estimator = "default",
             ordered = NULL,
             int.ov.free = TRUE,
             int.lv.free = FALSE,
             std.lv = TRUE,
             auto.fix.first = FALSE,
             auto.fix.single = TRUE,
             auto.cov.lv.x = TRUE,
             auto.th = TRUE,
             auto.delta = TRUE,
             auto.cov.y = TRUE
           ),
           maxChanges = 5,
           restartCriteria = "consecutive",
           maximumConsecutive = 25,
           maxItems = NULL,
           items = NULL,
           bifactor = FALSE,
           setChains = 1,
           shortForm = T,
           ...) {
    #### initial values ####
    if(!exists('originalData')) {
      stop("Please check that you have included the original data frame!")
    }
    if(!is.numeric(setChains)|setChains<1) {
      warning("The value for setChains was not valid. Defaulting to setChains=1.")
      setChains = 1
    }
    allFit <- c()

    # creates objects in the function environment that are fed into the lavaan function in order to fine-tune the model to user specifications
    # solution from: https://stackoverflow.com/questions/6375790/r-creating-an-environment-in-the-globalenv-from-inside-a-function
    checkModelSpecs(lavaan.model.specs)
    mapply(
      assign,
      names(lavaan.model.specs),
      lavaan.model.specs,
      MoreArgs = list(envir = environment())
    )

    if (!is.null(maxItems)) {
      # if using the short form option
      cat("Initializing short form creation.")
      
      mapply(
        assign,
        c("factors", "allItems"),
        syntaxExtraction(initialModel, items = items),
        MoreArgs = list(envir = environment())
      )

      currentModel <-
        bestModel <- 
        randomInitialModel(init.model = initialModel,
                           maxItems = maxItems,
                           allItems = allItems,
                           initialData = originalData,
                           bifactorModel = bifactor,
                           lavaan.model.specs = lavaan.model.specs
      )
      cat(paste0(
        "\nThe initial short form is:\n",
        paste0(currentModel@model.syntax, collapse = "")
      ))
      bestFit <-
        tryCatch(
          lavaan::fitmeasures(object = bestModel@model.output, fit.measures = fitStatistic),
          error = function(e, checkMaximize = maximize) {
            if (length(e) > 0) {
              if (checkMaximize == TRUE) {
                return(0)
                } else {
                  return(Inf)
                }
            }
            }
          )
      if (is.null(items)) {
        stop(
          "To use this function for short forms, you need to set both the maxItems to consider as well as the names of the items."
        )
      }

      cat("\nUsing the short form randomNeighbor function.")
      cat("\nFinished initializing short form options.")
    } else {
      # if not using the short form option
      bestFit <- tryCatch(
        lavaan::fitmeasures(object = bestModel, fit.measures = fitStatistic),
        error = function(e, checkMaximize = maximize) {
          if (length(e) > 0) {
            if (checkMaximize == TRUE) {
              return(0)
            } else {
              return(Inf)
            }
          }
        }
      )
    }


    #### selecting functions for use in algorithm ####
    if (temperature == "linear") {
      temperatureFunction <- linearTemperature
    } else if (temperature == "quadratic") {
      temperatureFunction <- quadraticTemperature
    } else if (temperature == "logistic") {
      temperatureFunction <- logisticTemperature
    } else if (inherits(temperature, "function")) {
      temperatureFunction <- temperature
    } else {
      stop(
        "You need to specify an appropriate default temperature (one of \"linear\", \"quadratic\", or \"logistic\") or include a temperature in the form of a function to continue."
      )
    }

    if (inherits(restartCriteria, "function")) {
      restartCriteria <- restartCriteria
    } else if (restartCriteria == "consecutive") {
      restartCriteria <- consecutiveRestart
    } else {
      restartCriteria <- function(maxConsecutiveSelection, consecutive) {
        
      }
      warning(
        "The restart criteria should to be either \"consecutive\" (the default) or a custom function. It has been set to NULL so the algorithm will not restart at all."
      )
    }
    #### prepare parallel processing ####
    if (setChains > 1) {
      chk <- Sys.getenv("_R_CHECK_LIMIT_CORES_", "")
      progressPar <- function(n) {
        cat(paste("Chain number ", n, " complete. \n", sep = ""))
      }
      
      if (nzchar(chk) && chk == "TRUE") {
        # use 2 cores in CRAN/Travis/AppVeyor
        num_workers <- 2L
      } else {
        if (setChains <= parallel::detectCores()) {
          num_workers <- setChains
        } else { 
          # use all cores in devtools::test()
          num_workers <- parallel::detectCores()
          }
      }
      cl <- parallel::makeCluster(num_workers,type="PSOCK", outfile = "")
      doSNOW::registerDoSNOW(cl)
      
    } else {
      num_workers <- 1L
      progressSeq <- function(currentStep, maxSteps) {
        cat(paste0(
          "\r Current Step = ",
          currentStep,
          " of a maximum ",
          maxSteps,
          ".  "
        ), file = stdout())
      }
      
      
    }
    
    `%dopar%` <- foreach::`%dopar%`
    if (setChains > 1) {
      opts <- list(progress = progressPar)
    } else {
      opts <- list(progress = progressSeq)
    }
    
    chains = setChains
    currentStep <- 1
    consecutive <- 0
    
    #### perform algorithm ####
    start.time <- Sys.time()

    cat("\n Current Progress: \n")

    allModel <- paste0(currentModel@model.syntax, collapse = "\n")
    allFit <- bestFit
    
    chainResults <-
      foreach::foreach(chain = 1:chains, .inorder = F, .combine = rbind, 
                       .options.snow = opts) %dopar% {

        while (currentStep < maxSteps) {
          
          # how many changes to make?
          numChanges <- sample(1:maxChanges, size = 1)
          # generate random model
          if (shortForm == FALSE) {
            randomNeighborModel <- randomNeighborFull(
              currentModelObject = currentModel,
              numChanges = numChanges,
              data = originalData
            )
          } else if (shortForm == TRUE) {
            randomNeighborModel <- randomNeighborShort(
              currentModelObject = currentModel,
              numChanges = numChanges,
              data = originalData,
              allItems = allItems,
              bifactor = bifactor,
              init.model = initialModel,
              lavaan.model.specs = lavaan.model.specs
            )
          }
          # select between random model and current model
          currentModel <- selectionFunction(
            currentModelObject = currentModel,
            randomNeighborModel = randomNeighborModel,
            currentTemp = temperatureFunction(currentStep, maxSteps),
            maximize = maximize,
            fitStatistic = fitStatistic,
            consecutive = consecutive
          )
          # record fit
          allFit[currentStep + 1] <- tryCatch(
            lavaan::fitmeasures(object = currentModel@model.output, fit.measures = fitStatistic),
            error = function(e) {
              if (length(e) > 0) {
                bestFit
              }
            }
          )
          # check for current best model
          bestModel <- checkModels(currentModel, fitStatistic, maximize, bestFit, bestModel)
          bestFit <- tryCatch(
            lavaan::fitmeasures(object = bestModel@model.output, fit.measures = fitStatistic),
            error = function(e) {
              if (length(e) > 0) {
                bestFit
              }
            }
          )
          
          allModel <- c(allModel, currentModel@model.syntax)
    
          # restart if the same model was chosen too many times
          restartCriteria(
            maxConsecutiveSelection = maximumConsecutive,
            consecutive = consecutive
          )
          currentStep <- currentStep + 1
          
        }
        returnList <-
          list(
            'allModel' = allModel,
            'allFit' = allFit,
            'bestModel' = bestModel,
            'bestFit' = bestFit
          )
      }
    
    
    if (setChains > 1) {
      foreach::registerDoSEQ()
      parallel::stopCluster(cl)
      
      best_fit <-
        ifelse(maximize,
               max(as.numeric(chainResults[,'bestFit'])),
               min(as.numeric(chainResults[,'bestFit']))
        )
      
      which_best <-
        which(as.numeric(chainResults[,'bestFit']) == best_fit)
      
      bestModel <-
        chainResults[which_best,'bestModel'][[1]]
        
      names(best_fit) <-
        names(chainResults[1,4][[1]])
      
      all_fit <-
        chainResults[,'allFit']
    } else {
      best_fit <-
        as.numeric(chainResults$bestFit)
      
      names(best_fit) <-
        names(chainResults$bestFit)
      
      bestModel <-
        chainResults$bestModel
      
      all_fit <-
        as.numeric(chainResults$allFit)
    }
    
    
    results <-
      new(
        'SA',
        function_call = match.call(),
        chains = setChains,
        chain_results = chainResults,
        all_fit = all_fit,
        best_fit = best_fit,
        best_model = bestModel,
        best_syntax = bestModel@model.syntax,
        runtime = Sys.time() - start.time
      )

    results
  }

randomInitialModel <- 
  function(init.model,
           maxItems,
           allItems,
           initialData,
           bifactorModel,
           lavaan.model.specs) {
    mapply(
      assign,
      names(lavaan.model.specs),
      lavaan.model.specs,
      MoreArgs = list(envir = environment())
    )
  # extract the latent factor syntax
  mapply(
    assign,
    c("factors", "itemsPerFactor"),
    syntaxExtraction(initialModelSyntaxFile = init.model, items = allItems),
    MoreArgs = list(envir = environment())
  )
  
  # save the external relationships
  vectorModel <- unlist(strsplit(x = init.model, split = "\\n"))
  externalRelation <- vectorModel[grep(" ~ ", vectorModel)]
  factorRelation <- vectorModel[grep(" ~~ ", vectorModel)]
  
  # reduce the number of items for each factor according to maxItems
  newItemsPerFactor <- list()
  for (i in 1:length(itemsPerFactor)) {
    newItemsPerFactor[[i]] <- 
      sample(x = unique(unlist(itemsPerFactor[i])), 
             size = unlist(maxItems[i]))
  }
  
  if (bifactorModel == TRUE) {
    # if bifactorModel == TRUE, fix the items so the newItems all load on the bifactor
    # assumes that the bifactor latent variable is the last one
    newItemsPerFactor[[length(itemsPerFactor)]] <- unlist(newItemsPerFactor[1:(length(itemsPerFactor) - 1)])
  }
  
  # create the new model syntax
  
  newModelSyntax <- c()
  for (i in 1:length(factors)) {
    newModelSyntax[i] <- paste(
      factors[i],
      "=~",
      paste(newItemsPerFactor[[i]], collapse = " + ")
    )
  }
  newModelSyntax <- 
    paste(newModelSyntax, externalRelation, factorRelation, sep = "\n")
  newModelSyntax <-
    stringr::str_replace_all(newModelSyntax, "\n\n", "\n")
  
  # fit the new model
  newModel <- modelWarningCheck(
    lavaan::lavaan(
      model = newModelSyntax,
      data = initialData,
      model.type = model.type,
      int.ov.free = int.ov.free,
      int.lv.free = int.lv.free,
      auto.fix.first = auto.fix.first,
      std.lv = std.lv,
      auto.fix.single = auto.fix.single,
      auto.var = auto.var,
      auto.cov.lv.x = auto.cov.lv.x,
      auto.th = auto.th,
      auto.delta = auto.delta,
      auto.cov.y = auto.cov.y,
      ordered = ordered,
      estimator = estimator,
    ),
    modelSyntax = newModelSyntax
  )
  
  return(newModel)
}

randomNeighborShort <-
  function(currentModelObject,
           numChanges,
           allItems,
           data,
           bifactor = FALSE,
           init.model,
           lavaan.model.specs) {
    
    mapply(
      assign,
      c("factors", "itemsPerFactor"),
      syntaxExtraction(initialModelSyntaxFile = init.model, items = allItems),
      MoreArgs = list(envir = environment())
    )
    
    mapply(
      assign,
      names(lavaan.model.specs),
      lavaan.model.specs,
      MoreArgs = list(envir = environment())
    )
    # take the model syntax from the currentModelObject
    internalModelObject <- 
      stringr::str_split(currentModelObject@model.syntax, 
                         pattern = "\n",
                         simplify = T)
    
    # extract the latent factor syntax
    mapply(
      assign,
      c("factors", "currentItems"),
      syntaxExtraction(initialModelSyntaxFile = internalModelObject, items = allItems),
      MoreArgs = list(envir = environment())
    )
    
    # randomly select current items to replace
    
    replacePattern <- paste0("\\b",
                             paste0(allItems,
                                    collapse = "\\b|\\b"
                             ),
                             "\\b")
    
    replacementItemPool <- c()
    for (i in 1:length(factors)) {
      if (inherits(allItems, "list")) {
        replacementItemPool[[i]] <- 
          allItems[[i]][!(allItems[[i]] %in% currentItems[[i]])]
      } else {
        replacementItemPool[[i]] <-
          allItems[!(allItems %in% currentItems[[i]])]
      }
    }
    changingItems <- c()
    replacementItem <- c()
    for (i in 1:numChanges) {
      # randomly select factor to have an item changed
      if (bifactor) {
        currentFactor <- sample(1:(length(factors) - 1), 1)
      } else {
        currentFactor <- sample(1:length(factors), 1)
      }
      # randomly select the item to be changed
      changingItemTemp <- c()
      changingItemTemp <- sample(currentItems[[currentFactor]], 1)
      while (changingItemTemp %in% changingItems ||
             length(changingItemTemp %in% changingItems) == 0) {
        changingItemTemp <- sample(currentItems[[currentFactor]], 1)
      }
      changingItems <- c(changingItems, changingItemTemp)
      # Sample an item from the items in the item pool
      tempReplacementItems <- sample(replacementItemPool[[currentFactor]], 1)
      while (tempReplacementItems %in% replacementItem) {
        tempReplacementItems <- sample(replacementItemPool[[currentFactor]], 1)
      }
      replacementItem <- c(replacementItem, tempReplacementItems)
    }
    
    for (i in 1:length(factors)) {
      for (j in 1:numChanges) {
        currentItems[[i]] <-
          gsub(
            pattern = paste0(changingItems[j], "\\b"),
            replacement = replacementItem[j],
            x = currentItems[[i]]
          )
      }
    }
    
    if (bifactor == TRUE) {
      # if bifactor == TRUE, fix the items so the newItems all load on the bifactor
      # assumes that the bifactor latent variable is the last one
      currentItems[[length(itemsPerFactor)]] <- unlist(currentItems[1:(length(itemsPerFactor) - 1)])
    }
    
    # create the new model syntax
    newModelSyntax <- as.vector(
      stringr::str_split(currentModelObject@model.syntax, 
                         "\n", 
                         simplify = T)
    )
    for (i in 1:length(factors)) {
      newModelSyntax[i] <- paste(
        factors[i],
        "=~",
        paste(currentItems[[i]], collapse = " + ")
      )
    }
    
    newModelSyntax <- stringr::str_flatten(newModelSyntax,
                                           collapse = "\n")
    
    # refit the model with new items
    randomNeighborModel <- modelWarningCheck(
      lavaan::lavaan(
        model = newModelSyntax,
        data = data,
        model.type = model.type,
        auto.var = auto.var,
        ordered = ordered,
        estimator = estimator,
        int.ov.free = int.ov.free,
        int.lv.free = int.lv.free,
        auto.fix.first = auto.fix.first,
        std.lv = std.lv,
        auto.fix.single = auto.fix.single,
        auto.cov.lv.x = auto.cov.lv.x,
        auto.th = auto.th,
        auto.delta = auto.delta,
        auto.cov.y = auto.cov.y
      ),
      modelSyntax = newModelSyntax
    )
    
    return(randomNeighborModel)
  }

randomNeighborFull <-
  function(currentModelObject = currentModel,
           numChanges = numChanges,
           data) {
    # using lavaan functions, construct a full parameter table
    paramTable <- lavaan::parTable(currentModelObject)
    
    # select the rows that correspond to parameters related to the latent variables
    latentVariables <-
      row.names(lavaan::inspect(currentModelObject, "cor.lv"))
    currentModelParamsLV <-
      paramTable[paramTable$lhs %in% latentVariables, ]
    
    # randomly select rows to make changes to
    randomChangesRows <-
      sample(currentModelParamsLV$id, size = numChanges)
    changeParamTable <- paramTable[randomChangesRows, ]
    
    # make the changes. If currently free, fix to 0; if fixed to 0, set to free
    paramTable$free[randomChangesRows] <-
      1 - paramTable$free[randomChangesRows]
    
    # remove the starting value, estimates, and standard errors of the currentModel
    paramTable$est <- NULL
    paramTable$se <- NULL
    paramTable$start <- NULL
    paramTable$labels <- NULL
    
    # refit the model
    prevModel <- as.list(currentModelObject@call)
    prevModel$model <- paramTable
    
    randomNeighborModel <-
      modelWarningCheck(
      lavaan::lavaan(
        model = prevModel$model,
        data = data
      ),
      modelSyntax = prevModel$model
    )
    
    return(randomNeighborModel)
  }

goal <-
  function(x, fitStatistic = "cfi", maximize) {
    # if using lavaan and a singular fit statistic,
    if (inherits(x, "lavaan") &
        is.character(fitStatistic) & length(fitStatistic) == 1) {
      energy <- fitWarningCheck(lavaan::fitMeasures(x, fit.measures = fitStatistic),
                                maximize)
      if (is.na(energy) & maximize == TRUE) {
        energy = -Inf
      } else if (is.na(energy)) {
        energy = Inf
      }
      
      if (maximize == TRUE) {
        return(-energy)
      } else {
        return(energy)
      }
    } else {
      if (inherits(x, "NULL")) {
        if (maximize == TRUE) {
          return(-Inf)
        } else {
          return(Inf)
        }
      }
    }
  }

selectionFunction <-
  function (currentModelObject,
            randomNeighborModel,
            currentTemp,
            maximize,
            fitStatistic,
            consecutive)
  {
    # check if the randomNeighborModel is a valid model for use
    if (!inherits(randomNeighborModel, "modelCheck")) {
      return(currentModelObject)
    }
    
    # check that the current model isn't null
    if (!inherits(currentModelObject, "modelCheck")) {
      return(randomNeighborModel)
      }
    if (is.null(currentModelObject@model.output)) {
      return(randomNeighborModel)
    }
    
    randomNeighborConverge <-
      tryCatch(
        lavaan::fitmeasures(object = randomNeighborModel@model.output,
                            fit.measures = fitStatistic),
        error = function(e) {
          if (length(e) > 0) {
            NULL
          }
        }
      )
    if (is.null(randomNeighborConverge)) {
      cat("\rNew model did not converge.                                                                                          ")
      return(currentModelObject)
    }
    if ( 
      length(randomNeighborModel@warnings) > 0 |
      length(randomNeighborModel@errors) > 0 
    ) {
      return(currentModelObject)
    }
    
    # this is the Kirkpatrick et al. method of selecting between currentModel and randomNeighborModel
    if (goal(randomNeighborModel@model.output, fitStatistic, maximize) < goal(currentModelObject@model.output, fitStatistic, maximize)) {
      cat(paste0(
        "\rOld Fit: ",
        round(as.numeric(
          fitWarningCheck(
            lavaan::fitmeasures(object = currentModelObject@model.output,
                                fit.measures = fitStatistic),
            maximize
          )
        ), 3),
        " New Fit: ",
        round(as.numeric(
          fitWarningCheck(
            lavaan::fitmeasures(object = randomNeighborModel@model.output,
                                fit.measures = fitStatistic),
            maximize
          )
        ), 3),
        "                                                                    "
      ))
      consecutive <- consecutive + 1
      return(randomNeighborModel)
    } else {
      probability <- exp(-(
        goal(randomNeighborModel@model.output, fitStatistic, maximize) - goal(currentModelObject@model.output, fitStatistic, maximize)
      ) / currentTemp)
      
      if (is.nan(probability)) probability = 0
      
      if (probability > stats::runif(1)) {
        newModel <- randomNeighborModel
      } else {
        newModel <- currentModelObject
      }
      
      consecutive <- ifelse(
        identical(
          lavaan::parameterTable(newModel@model.output),
          lavaan::parameterTable(currentModelObject@model.output)
        ),
        consecutive + 1,
        0
      )
      cat(paste0(
        "\rOld Fit: ",
        round(as.numeric(
          lavaan::fitmeasures(object = currentModelObject@model.output,
                              fit.measures = fitStatistic)
        ), 3),
        " New Fit: ",
        round(as.numeric(
          lavaan::fitmeasures(object = randomNeighborModel@model.output,
                              fit.measures = fitStatistic)
        ), 3),
        " Probability: ",
        round(probability, 3)
      ))
      return(newModel)
    }
  }

consecutiveRestart <-
  function(maxConsecutiveSelection = 25, consecutive) {
    if (consecutive == maxConsecutiveSelection) {
      currentModel <- bestModel
      consecutive <- 0
    }
  }

checkModels <- function(currentModel, fitStatistic, maximize = maximize, bestFit = bestFit, bestModel) {
  if (is.null(currentModel)|!inherits(currentModel, "modelCheck")) {
    return(bestModel)
  }
  currentFit <- fitWarningCheck(
    lavaan::fitmeasures(object = currentModel@model.output, fit.measures = fitStatistic),
    maximize
  )
  if (maximize == TRUE) {
    if (currentFit > bestFit) {
      bestModel <- currentModel
    } else {
      bestModel <- bestModel
    }
  } else {
    if (currentFit < bestFit) {
      bestModel <- currentModel
    } else {
      if (currentFit < bestFit) {
        bestModel <- currentModel
      } else {
        bestModel <- bestModel
      }
    }
  }

  return(bestModel)
}

modelWarningCheck <- function(expr, modelSyntax) {
  warn <- err <- c()
  value <- withCallingHandlers(
    tryCatch(
      expr,
      error = function(e) {
        err <<-
          append(err, e$message)
        NULL
      }
    ),
    warning = function(w) {
      warn <<-
        append(warn, w$message)
      invokeRestart("muffleWarning")
    }
  )
  
  new(
    'modelCheck',
    "model.output" = value,
    "warnings" = as.character(unlist(warn)),
    "errors" = as.character(unlist(err)),
    'model.syntax' = modelSyntax
  )
}


syntaxExtraction <- function(initialModelSyntaxFile, items) {

  # extract the latent factor syntax
  factors <- unique(lavaan::lavaanify(initialModelSyntaxFile)[lavaan::lavaanify(initialModelSyntaxFile)$op ==
    "=~", "lhs"])
  vectorModelSyntax <- stringr::str_split(
    string = initialModelSyntaxFile,
    pattern = "\\n",
    simplify = TRUE
  )
  factorSyntax <- c()
  itemSyntax <- c()
  for (i in 1:length(factors)) {
    chosenFactorLocation <- c(1:length(vectorModelSyntax))[grepl(x = vectorModelSyntax, pattern = paste0(factors[i], "[ ]{0,}=~ "))]
    factorSyntax[i] <- vectorModelSyntax[chosenFactorLocation]
    # remove the factors from the syntax
    itemSyntax[i] <-
      gsub(
        pattern = paste(factors[i], "=~ "),
        replacement = "",
        x = factorSyntax[i]
      )
  }

  # extract the items for each factor
  itemsPerFactor <- stringr::str_extract_all(
    string = itemSyntax,
    pattern = paste0("(\\b", paste0(
      paste0(unlist(items), collapse = "\\b)|(\\b"), "\\b)"
    ))
  )
  return(list("factors" = factors, "itemsPerFactor" = itemsPerFactor))
}

fitWarningCheck <-
  function(expr, maximize) {
    value <- withCallingHandlers(tryCatch(
      expr,
      error = function(e) {
        return(NA)
        invokeRestart("muffleWarning")
      }
    ))
    return(value)
  }

checkModelSpecs <- 
  function(
    x
  ) {
    
    requiredElements <-
      c('model.type',
        'auto.var',
        'estimator',
        'ordered',
        'int.ov.free',
        'int.lv.free',
        'auto.fix.first',
        'auto.fix.single',
        'auto.cov.lv.x',
        'auto.th',
        'auto.delta',
        'auto.cov.y',
        'std.lv')
    
    missingSpecs <-
      requiredElements[
        which(
          !requiredElements %in% names(x)
        )
        ]
    
    if (length(missingSpecs) > 0) {
      errorMessage <-
        paste0("The following elements of lavaan.model.specs have not been specified:\n\n",
               paste(missingSpecs, collapse = "\n"),
               "\n\nPlease include the proper specifications for these elements, or use the default values provided.")
      stop(errorMessage)
    }
  }

fitmeasuresCheck <-
  function(
    x
  ) {
    validMeasures <-
      c(
        "npar", "fmin",
        "chisq", "df", "pvalue", 
        "baseline.chisq", "baseline.df", "baseline.pvalue",    
        "cfi", "tli", "nnfi",
        "rfi", "nfi", "pnfi",
        "ifi", "rni",                
        "logl", "unrestricted.logl",
        "aic", "bic", "ntotal", "bic2",
        "rmsea", "rmsea.ci.lower", "rmsea.ci.upper", "rmsea.pvalue",
        "rmr", "rmr_nomean", 
        "srmr", "srmr_bentler", "srmr_bentler_nomean", 
        "crmr", "crmr_nomean", "srmr_mplus", "srmr_mplus_nomean",
        "cn_05", "cn_01", 
        "gfi", "agfi", "pgfi", "mfi", "ecvi",
        "chisq.scaled", "df.scaled", "pvalue.scaled", "chisq.scaling.factor",
        "baseline.chisq.scaled", "baseline.df.scaled", "baseline.pvalue.scaled",
        "baseline.chisq.scaling.factor",
        "cfi.scaled", "tli.scaled",                   
        "cfi.robust", "tli.robust",                   
        "nnfi.scaled", "nnfi.robust",                  
        "rfi.scaled",  "nfi.scaled", "ifi.scaled", "rni.scaled", "rni.robust",
        "rmsea.scaled", "rmsea.ci.lower.scaled", "rmsea.ci.upper.scaled",
        "rmsea.pvalue.scaled", "rmsea.robust", "rmsea.ci.lower.robust", 
        "rmsea.ci.upper.robust", "rmsea.pvalue.robust"
      )
    
    invalidMeasures <-
      x[
        which(
          !x %in% validMeasures
        )
        ]
    
    if (length(invalidMeasures) > 0) {
      errorMessage <-
        paste0("The following elements of fit.indices or fitStatistics are not valid fit measures provided by the lavaan::fitmeasures function:\n\n",
               paste(invalidMeasures, collapse = "\n"),
               "\n\nPlease check the output of this function for proper spelling and capitalization of the fit measure(s) you are interested in.")
      stop(errorMessage)
    }

  }

fitStatTestCheck <-
  function(measures, test) {
    tempEnv <-
      new.env()
    mapply(assign,
           measures,
           0,
           MoreArgs = list(envir = tempEnv))
    
    checkIfEval <-
      tryCatch(
        expr = eval(parse(text = test),
                    envir = tempEnv),
        error = function(e) {
          stop(
            "There was a problem with the fit.statistics.test provided.
            It cannot be evaluated properly.
            Please read the function documentation to see how to properly specify a test."
          )
        }
      )
    
    if (!is.character(test)&is.logical(test)) {
      stop(
        "There is a problem with the fit.statistics.test provided.
        The fit.statistics.test was given as a logical, not a character.
        Please read the function documentation to see how to properly specify a test."
      )
    } else if (!is.character(test)) {
      stop(
        "There is a problem with the fit.statistics.test provided.
        The fit.statistics.test was not given as a character object.
        Please read the function documentation to see how to properly specify a test."
      )
    }
    checkIfEval
  }

# allArgs ####
# allArgs <- function(orig_values = FALSE) {
#   # source: https://stackoverflow.com/a/47955845
#   # get formals for parent function
#   parent_formals <- formals(sys.function(sys.parent(n = 1)))
#   
#   # Get names of implied arguments
#   fnames <- names(parent_formals)
#   
#   # Remove '...' from list of parameter names if it exists
#   fnames <- fnames[-which(fnames == '...')]
#   
#   # Get currently set values for named variables in the parent frame
#   args <- evalq(as.list(environment()), envir = parent.frame())
#   
#   # Get the list of variables defined in '...'
#   args <- c(args[fnames], evalq(list(...), envir = parent.frame()))
#   
#   
#   if(orig_values) {
#     # get default values
#     defargs <- as.list(parent_formals)
#     defargs <- defargs[unlist(lapply(defargs, FUN = function(x) inherits(x, "name")))]
#     args[names(defargs)] <- defargs
#     setargs <- evalq(as.list(match.call())[-1], envir = parent.frame())
#     args[names(setargs)] <- setargs
#   }
#   return(args)
# }

# randomInitialModel ####
test_that(
  "randomInitialModel produces a modelCheck object", {
    defaultModel <-
    ' visual  =~ x1 + x2 + x3
      textual =~ x4 + x5 + x6
      speed   =~ x7 + x8 + x9'
    defaultBifactor <-
    ' visual   =~ x1 + x2 + x3
      textual  =~ x4 + x5 + x6
      speed    =~ x7 + x8 + x9
      bifactor =~ x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9'
    maxItems <- c(2,2,2)
    allItems <- paste0("x", 1:9)
    data <- lavaan::HolzingerSwineford1939
    bifactor <- FALSE
    lavaanSpecs <-
      list(model.type = "cfa",
           estimator = "ML",
           ordered = NULL,
           int.ov.free = TRUE,
           int.lv.free = FALSE,
           auto.fix.first = FALSE,
           std.lv = TRUE,
           auto.fix.single = TRUE,
           auto.var = TRUE,
           auto.cov.lv.x = TRUE,
           auto.efa = TRUE,
           auto.th = TRUE,
           auto.delta = TRUE,
           auto.cov.y = TRUE,
           missing = "listwise",
           group = NULL,
           group.label = NULL,
           group.equal = 'loadings',
           group.partial = NULL,
           group.w.free = FALSE)

    ## bifactorModel == FALSE ####
    # should output a modelCheck object
    expect_s4_class(
      randomInitialModel(
        init.model = defaultModel,
        maxItems = maxItems,
        allItems = allItems,
        initialData = data,
        bifactorModel = bifactor,
        lavaan.model.specs = lavaanSpecs
      ),
      'modelCheck'
    )

    # bifactor == TRUE ####
    bifactor = TRUE
    # should output a list
    expect_s4_class(
      randomInitialModel(
        init.model = defaultModel,
        maxItems = maxItems,
        allItems = allItems,
        initialData = data,
        bifactorModel = bifactor,
        lavaan.model.specs = lavaanSpecs
      ),
      'modelCheck'
    )
  }
)

# randomNeighborShort ####
test_that(
  "randomNeighborShort produces a list of appropriate length and with appropriate names", {
    defaultModel <-
    ' visual  =~ x1 + x2
      textual =~ x4 + x5
      speed   =~ x7 + x8'
    defaultBifactor <-
    ' visual   =~ x1 + x2
      textual  =~ x4 + x5
      speed    =~ x7 + x8
      bifactor =~ x1 + x2 + x4 + x5 + x7 + x8'
    currentModel <-
      modelWarningCheck(
        lavaan::cfa(
          model = defaultModel,
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax = defaultModel
      )
    currentBifactor <-
      modelWarningCheck(
        lavaan::cfa(
          model = defaultBifactor,
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax = defaultModel
      )
    lavaanSpecs <-
      list(model.type = "cfa",
           estimator = "ML",
           ordered = NULL,
           int.ov.free = TRUE,
           int.lv.free = FALSE,
           auto.fix.first = FALSE,
           std.lv = TRUE,
           auto.fix.single = TRUE,
           auto.var = TRUE,
           auto.cov.lv.x = TRUE,
           auto.efa = TRUE,
           auto.th = TRUE,
           auto.delta = TRUE,
           auto.cov.y = TRUE,
           missing = "listwise",
           group = NULL,
           group.label = NULL,
           group.equal = 'loadings',
           group.partial = NULL,
           group.w.free = FALSE)

    ## bifactorModel == FALSE ####
    # should output a modelCheck object
    expect_s4_class(
      randomNeighborShort(
        currentModelObject = currentModel,
        numChanges = 1,
        allItems = paste0("x", 1:9),
        data = lavaan::HolzingerSwineford1939,
        bifactor = FALSE,
        init.model = defaultModel,
        lavaan.model.specs = lavaanSpecs
      ),
      'modelCheck'
    )


    # bifactorModel == FALSE ####
    # should output a modelCheck object
    expect_s4_class(
      randomNeighborShort(
        currentModelObject = currentBifactor,
        numChanges = 1,
        allItems = paste0("x", 1:9),
        data = lavaan::HolzingerSwineford1939,
        bifactor = TRUE,
        init.model = defaultBifactor,
        lavaan.model.specs = lavaanSpecs
      ),
      'modelCheck'
    )
  }
)

# # randomNeighborFull ####
# test_that(
#   "randomNeighborFull produces a list of appropriate length and with appropriate names", {
#     currentModel <-
#       modelWarningCheck(
#         lavaan::cfa(
#           model =
#           ' visual  =~ x1 + x2 + x3
#             textual =~ x4 + x5 + x6
#             speed   =~ x7 + x8 + x9',
#           data = lavaan::HolzingerSwineford1939
#         ),
#         modelSyntax =
#           ' visual  =~ x1 + x2 + x3
#             textual =~ x4 + x5 + x6
#             speed   =~ x7 + x8 + x9'
#       )
#     # should produce a message
#     expect_message(
#       randomNeighborFull(
#         currentModelObject = currentModel@model.output,
#         numChanges = 1,
#         data = lavaan::HolzingerSwineford1939
#       )
#     )
#
#     # should output a list
#     expect_type(
#       randomNeighborFull(
#         currentModelObject = currentModel$model.output,
#         numChanges = 2,
#         data = lavaan::HolzingerSwineford1939
#       ),
#       'list'
#     )
#
#     # the list should be of length 4
#     expect_length(
#       randomNeighborFull(
#         currentModelObject = currentModel$model.output,
#         numChanges = 2,
#         data = lavaan::HolzingerSwineford1939
#       ),
#       4
#     )
#
#     # the list should have specified names in specified order
#     expect_named(
#       randomNeighborFull(
#         currentModelObject = currentModel$model.output,
#         numChanges = 2,
#         data = lavaan::HolzingerSwineford1939
#       ),
#       expected =
#         c("model.output", "warnings", "errors", "model.syntax")
#     )
#
#
#   }
# )
# goal ####
test_that(
  "goal returns the 'energy' value of a fit statistic test, including in cases of NA fit", {
    currentModel <-
      modelWarningCheck(
        lavaan::cfa(
          model =
            ' visual  =~ x1 + x2 + x3
            textual =~ x4 + x5 + x6
            speed   =~ x7 + x8 + x9',
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax =
          ' visual  =~ x1 + x2 + x3
            textual =~ x4 + x5 + x6
            speed   =~ x7 + x8 + x9'
      )
    currentBadModel <-
      modelWarningCheck(
        lavaan::cfa(
          model =
          ' visual  =~ x1 + x2 + x3 + x4 + x5 + x6
            textual =~ x4 + x5 + x6 + x7 + x8 + x9
            speed   =~ x7 + x8 + x9 + x1 + x2 + x3',
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax =
          ' visual  =~ x1 + x2 + x3 + x4 + x5 + x6
            textual =~ x4 + x5 + x6 + x7 + x8 + x9
            speed   =~ x7 + x8 + x9 + x1 + x2 + x3'
      )

    # lavaan input with single, maximized fit statistic should equal the negative value of the fit statistic
    expect_equal(
      goal(
        x = currentModel@model.output,
        fitStatistic = 'cfi',
        maximize = T
        ),
      -fitmeasures(
        object = currentModel@model.output,
        fit.measures = 'cfi'
      )
    )

    # lavaan input with single, minimized fit statistic should equal the value of that fit statistic
    expect_equal(
      goal(
        x = currentModel@model.output,
        fitStatistic = 'rmsea',
        maximize = F
      ),
      fitmeasures(
        object = currentModel@model.output,
        fit.measures = 'rmsea'
      )
    )

    # lavaan input with no fit statistc due to poor model, goal maximized
    expect_equal(
      goal(
        x = currentBadModel@model.output,
        fitStatistic = 'cfi',
        maximize = T
      ),
      Inf
    )

    # lavaan input with no fit statistc due to poor model, goal minimized
    expect_equal(
      goal(
        x = currentBadModel@model.output,
        fitStatistic = 'rmsea',
        maximize = F
      ),
      Inf
    )

  }
)
# selectionFunction ####
test_that(
  "selectionFunction accurately chooses between the best of two models", {
    currentModel <-
      modelWarningCheck(
        lavaan::cfa(
          model =
            ' visual  =~ x1 + x2 + x3
            textual =~ x4 + x5 + x6
            speed   =~ x7 + x8 + x9',
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax =
          ' visual  =~ x1 + x2 + x3
            textual =~ x4 + x5 + x6
            speed   =~ x7 + x8 + x9'
      )
    intermediateModel <-
      modelWarningCheck(
        lavaan::cfa(
          model =
          ' visual  =~ x1 + x2 + x3
            textual =~ x4 + x5 + x6
            speed   =~ x7 + x8 + x9
            bifactor=~ x2 + x6 + x9',
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax =
          ' visual  =~ x1 + x2 + x3
            textual =~ x4 + x5 + x6
            speed   =~ x7 + x8 + x9
            bifactor=~ x2 + x6 + x9'
        )
    currentBadModel <-
      modelWarningCheck(
        lavaan::cfa(
          model =
            ' visual  =~ x1 + x2 + x3 + x4 + x5 + x6
            textual =~ x4 + x5 + x6 + x7 + x8 + x9
            speed   =~ x7 + x8 + x9 + x1 + x2 + x3',
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax =
          ' visual  =~ x1 + x2 + x3 + x4 + x5 + x6
            textual =~ x4 + x5 + x6 + x7 + x8 + x9
            speed   =~ x7 + x8 + x9 + x1 + x2 + x3'
      )
    currentTemp <-
      0.5

    # new model does not converge, so return currentModelObject with message
    expect_equal(
      selectionFunction(
        currentModelObject = currentModel,
        randomNeighborModel = currentBadModel,
        currentTemp = currentTemp,
        maximize = TRUE,
        fitStatistic = "cfi",
        consecutive = 1
      ),
      currentModel
    )
    expect_output(
      selectionFunction(
        currentModelObject = currentModel,
        randomNeighborModel = currentBadModel,
        currentTemp = currentTemp,
        maximize = TRUE,
        fitStatistic = "cfi",
        consecutive = 1
      ),
      "New model did not converge."
    )

    # old model is null, so return the new model
    expect_equal(
      selectionFunction(
        currentModelObject = NULL,
        randomNeighborModel = currentBadModel,
        currentTemp = currentTemp,
        maximize = TRUE,
        fitStatistic = "cfi",
        consecutive = 1
      ),
      currentBadModel
    )

    # new model is null, so return the old
    expect_equal(
      selectionFunction(
        currentModelObject = currentModel,
        randomNeighborModel = NULL,
        currentTemp = currentTemp,
        maximize = TRUE,
        fitStatistic = "cfi",
        consecutive = 1
      ),
      currentModel
    )

    # choosing between current model with better fit than new model, prints message with probability
    expect_output(
      selectionFunction(
        currentModelObject = intermediateModel,
        randomNeighborModel = currentModel,
        currentTemp = currentTemp,
        maximize = TRUE,
        fitStatistic = "cfi",
        consecutive = 1
      ),
      "Probability:"
    )

    # choosing between current model with worse fit than new model, prints message WITHOUT probability
    expect_output(
      selectionFunction(
        currentModelObject = currentModel,
        randomNeighborModel = intermediateModel,
        currentTemp = currentTemp,
        maximize = TRUE,
        fitStatistic = "cfi",
        consecutive = 1
      ),
      "^((?!Probability:))",
      perl = TRUE
    )
  }
)
# consecutiveRestart ####
test_that(
  "consecutiveRestart produces the correct model depending on the value of consecutive", {
    # currently not really testing due to how consecutiveRestart is working
    bestModel <-
      modelWarningCheck(
        lavaan::cfa(
          model =
            ' visual  =~ x1 + x2 + x3
            textual =~ x4 + x5 + x6
            speed   =~ x7 + x8 + x9',
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax =
          ' visual  =~ x1 + x2 + x3
            textual =~ x4 + x5 + x6
            speed   =~ x7 + x8 + x9'
      )
    currentModel <-
      testedModel <-
      modelWarningCheck(
        lavaan::cfa(
          model =
          ' visual  =~ x1 + x2 + x4
            textual =~ x3 + x5 + x7
            speed   =~ x6 + x8 + x9',
          data = lavaan::HolzingerSwineford1939
        ),
        modelSyntax =
          ' visual  =~ x1 + x2 + x4
            textual =~ x3 + x5 + x7
            speed   =~ x6 + x8 + x9'
      )
    consecutive = 1

    # consecutive < maxConsecutiveSelection, do not replace currentModel
    expect_false(
      c(
        consecutiveRestart(
          maxConsecutiveSelection = 25,
          consecutive = 1
          ), identical(currentModel, bestModel)
      )
    )
  }
)
# checkModels ####
test_that(
  "checkModels returns the appropriate bestModel depending on the conditions", {
    # example models
    bestHolzinger <-
      modelWarningCheck(
        lavaan::lavaan(model =
        ' visual  =~ x1 + x2 + x3
          textual =~ x4 + x5 + x6
          speed   =~ x7 + x8 + x9 ',
        data = lavaan::HolzingerSwineford1939,
        auto.var=TRUE,
        auto.fix.first=TRUE,
        auto.cov.lv.x=TRUE),
        ' visual  =~ x1 + x2 + x3
          textual =~ x4 + x5 + x6
          speed   =~ x7 + x8 + x9 '
      )
    badHolzinger <-
      modelWarningCheck(
        lavaan::lavaan(model =
        ' visual  =~ x1 + x2 + x3 + x7
          textual =~ x4 + x5 + x6 + x8 + x9',
        data = lavaan::HolzingerSwineford1939,
        auto.var=TRUE,
        auto.fix.first=TRUE,
        auto.cov.lv.x=TRUE),
        ' visual  =~ x1 + x2 + x3 + x7
          textual =~ x4 + x5 + x6 + x8 + x9'
      )

    # model is null, return bestModel
    expect_equal(
      checkModels(
        currentModel = NULL,
        fitStatistic = 'cfi',
        maximize = TRUE,
        bestFit = fitmeasures(bestHolzinger@model.output, 'cfi'),
        bestModel = bestHolzinger
      ),
      bestHolzinger
    )

    ## maximize == TRUE ####
    # models are same, return one of them
    expect_equal(
      checkModels(
        currentModel = bestHolzinger,
        fitStatistic = 'cfi',
        maximize = TRUE,
        bestFit = fitmeasures(bestHolzinger@model.output, 'cfi'),
        bestModel = bestHolzinger
      ),
      bestHolzinger
    )

    # comparing best model to (current) bad model, return best
    expect_equal(
      checkModels(
        currentModel = badHolzinger,
        fitStatistic = 'cfi',
        maximize = TRUE,
        bestFit = fitmeasures(bestHolzinger@model.output, 'cfi'),
        bestModel = bestHolzinger
      ),
      bestHolzinger
    )

    # comparing (current bad) best model to (new) best model, return best
    expect_equal(
      checkModels(
        currentModel = bestHolzinger,
        fitStatistic = 'cfi',
        maximize = TRUE,
        bestFit = fitmeasures(badHolzinger@model.output, 'cfi'),
        bestModel = badHolzinger
      ),
      bestHolzinger
    )

    # expect the bestModel if the currentModel is improperly specified
    expect_equal(
      checkModels(
        currentModel = "bestHolzinger",
        fitStatistic = 'cfi',
        maximize = TRUE,
        bestFit = fitmeasures(bestHolzinger@model.output, 'cfi'),
        bestModel = bestHolzinger
      ),
      bestHolzinger
    )

    # compare currentModel (as best so far) to true bestModel, return best

    expect_equal(
      checkModels(
        currentModel = bestHolzinger,
        fitStatistic = 'cfi',
        maximize = TRUE,
        bestFit = fitmeasures(badHolzinger@model.output, 'cfi'),
        bestModel = badHolzinger
      ),
      bestHolzinger
    )

    # maximize == FALSE ####
    # models are same, return one of them
    expect_equal(
      checkModels(
        currentModel = bestHolzinger,
        fitStatistic = 'rmsea',
        maximize = FALSE,
        bestFit = fitmeasures(bestHolzinger@model.output, 'rmsea'),
        bestModel = bestHolzinger
      ),
      bestHolzinger
    )

    # comparing best model to (current) bad model, return best
    expect_equal(
      checkModels(
        currentModel = badHolzinger,
        fitStatistic = 'rmsea',
        maximize = FALSE,
        bestFit = fitmeasures(bestHolzinger@model.output, 'rmsea'),
        bestModel = bestHolzinger
      ),
      bestHolzinger
    )

    # comparing (current bad) best model to (new) best model, return best
    expect_equal(
      checkModels(
        currentModel = bestHolzinger,
        fitStatistic = 'rmsea',
        maximize = FALSE,
        bestFit = fitmeasures(badHolzinger@model.output, 'rmsea'),
        bestModel = badHolzinger
      ),
      bestHolzinger
    )

    # expect the bestModel if the currentModel is improperly specified
    expect_equal(
      checkModels(
        currentModel = "bestHolzinger",
        fitStatistic = 'rmsea',
        maximize = FALSE,
        bestFit = fitmeasures(bestHolzinger@model.output, 'rmsea'),
        bestModel = bestHolzinger
      ),
      bestHolzinger
    )

    # compare currentModel (as best so far) to FALSE bestModel, return best

    expect_equal(
      checkModels(
        currentModel = bestHolzinger,
        fitStatistic = 'rmsea',
        maximize = FALSE,
        bestFit = fitmeasures(badHolzinger@model.output, 'rmsea'),
        bestModel = badHolzinger
      ),
      bestHolzinger
    )
  }
)

# modelWarningCheck ####
test_that(
  "modelWarningCheck identifies any warning/error in a model while returning model object", {
    defaultModel <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '
    wrongItemModel <- ' visual  =~ y1 + y2 + y3
              textual =~ y4 + y5 + y6
              speed   =~ y7 + y8 + y9 '
    poorModel <- ' visual  =~ x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9'

    # model fits without error
    goodModel <-
      modelWarningCheck(
        expr = lavaan::lavaan(defaultModel,
                              data = lavaan::HolzingerSwineford1939,
                              auto.var=TRUE,
                              auto.fix.first=TRUE,
                              auto.cov.lv.x=TRUE),
        modelSyntax = defaultModel
      )

    expect_length(goodModel@warnings, 0)
    expect_length(goodModel@errors, 0)
    expect_s4_class(goodModel@model.output, 'lavaan')
    # model doesn't fit and throws an error
    noModel <-
      modelWarningCheck(
        expr = lavaan::lavaan(wrongItemModel,
                              data = lavaan::HolzingerSwineford1939,
                              auto.var=TRUE,
                              auto.fix.first=TRUE,
                              auto.cov.lv.x=F),
        modelSyntax = wrongItemModel
      )

    expect_length(noModel@warnings, 0)
    expect_length(noModel@errors, 1)
    expect_null(noModel@model.output)

    # model fits but throws a warning
    warningModel <-
      modelWarningCheck(
        expr = lavaan::lavaan(poorModel,
                              data = lavaan::HolzingerSwineford1939,
                              auto.var=TRUE,
                              auto.fix.first=TRUE,
                              auto.cov.lv.x=F,
                              ordered = T),
        modelSyntax = poorModel
      )

    expect_length(warningModel@warnings, 1)
    expect_length(warningModel@errors, 0)
    expect_s4_class(warningModel@model.output, 'lavaan')

  }
)

# syntaxExtraction ####
test_that(
  "syntaxExtraction creates a valid list", {
    initialModelSyntaxFile <-"
f1 =~ Item1 + Item2 + Item3 + Item4 + Item5
f2 =~ Item6 + Item7 + Item8 + Item9 + Item10
f3 =~ Item11 + Item12 + Item13 + Item14 + Item15
f4 =~ Item16 + Item17 + Item18 + Item19 + Item20
"
    items <-
      paste0("Item", c(1:20))

    expectedOutput <-
      list(
        'factors' = paste0("f", c(1:4)),
        'itemsPerFactor' =
          list(
            paste0("Item", c(1:5)),
            paste0("Item", c(6:10)),
            paste0("Item", c(11:15)),
            paste0("Item", c(16:20))
          )
      )

    expect_equal(
      object =
        syntaxExtraction(
          initialModelSyntaxFile = initialModelSyntaxFile,
          items = items
        ),
      expected =
        expectedOutput
    )

  }
)

# fitWarningCheck ####
test_that(
  "fitWarningCheck returns value when working and NA when expr is not valid", {
    exampleCFI <-
      0.95
    exampleChisq <-
      120

    # normal input/output
    expect_equal(
      object =
        fitWarningCheck(expr = exampleCFI,
                        maximize = TRUE),
      expected = exampleCFI
    )

    expect_equal(
      object =
        fitWarningCheck(expr = exampleChisq,
                        maximize = FALSE),
      expected = exampleChisq
    )

    # error input, maximize
    expect_equal(
      object =
        fitWarningCheck(expr = 'exampleNull'/2,
                        maximize = TRUE),
      expected = NA
    )

    # error input, minimize
    expect_equal(
      object =
        fitWarningCheck(expr = 'exampleNull'/2,
                        maximize = FALSE),
      expected = NA
    )
  }
)

# checkModelSpecs ####
test_that(
  "checkModelSpecs returns silent when working, error when not", {
    # full, correct model specs list
    lavaanDesc <-
      list(model.type = "sem",
           estimator = "ML",
           ordered = NULL,
           int.ov.free = TRUE,
           int.lv.free = FALSE,
           auto.fix.first = FALSE,
           std.lv = TRUE,
           auto.fix.single = TRUE,
           auto.var = TRUE,
           auto.cov.lv.x = TRUE,
           auto.efa = TRUE,
           auto.th = TRUE,
           auto.delta = TRUE,
           auto.cov.y = TRUE)

    # correct model specs list with missing values
    lavaanErr <-
      list(model.type = "sem",
           estimator = "ML",
           ordered = NULL,
           int.ov.free = TRUE,
           std.lv = TRUE,
           auto.fix.single = TRUE,
           auto.var = TRUE,
           auto.cov.lv.x = TRUE,
           auto.efa = TRUE,
           auto.th = TRUE,
           auto.delta = TRUE,
           auto.cov.y = TRUE)

    expect_silent(
      checkModelSpecs(lavaanDesc)
    )

    expect_error(
      checkModelSpecs(lavaanErr)
    )
  }
)

# fitmeasuresCheck ####
test_that(
  "fitmeasuresCheck returns silent when working, error when not", {
    #when provided correct fitmeasure
    expect_silent(
      fitmeasuresCheck('cfi')
      )

    #when provided incorrect fitmeasure as string
    expect_error(
      fitmeasuresCheck('something')
    )

    #when provided incorrect fitmeasure as not-a-string
    expect_error(
      fitmeasuresCheck(2)
    )
    }
  )

# fitStatTestCheck ####
test_that(
  "user-defined fit.statistics.test passes validity checks for evaluation", {
    expect_error(
      fitStatTestCheck(measures = "(cfi > 0.95)&(tli > 0.95)&(rmsea < 0.06)",
                       test = TRUE),
      regexp = "logical"
    )

    expect_error(
      fitStatTestCheck(measures = "(cfi > 0.95)&(tli > 0.95)&(rmsea < 0.06)",
                       test = 1),
      regexp = "not given as a character object"
    )

    expect_false(
      fitStatTestCheck(measures = c('cfi','tli','rmsea'),
                       test = "(cfi > 0.95)&(tli > 0.95)&(rmsea < 0.06)"),
    )

    expect_error(
      fitStatTestCheck(measures = c('cfi','tli','rmsea'),
                       test = "(cfi >> 0.95)&(tli > 0.95)&(rmsea < 0.06)"),
    )
  }
)

# # allArgs ####
# test_that(
#   "allArgs can provide the default values for a function", {
#     temp <-
#       function(x = 2, y = 10, default = TRUE, ...) {
#         z <-
#           x^y
# 
#         return(
#           list(
#             "result" = z,
#             "args" = allArgs(orig_values = default)
#           )
#         )
#       }
#     defaultOutput <-
#       list(
#         "result" = 1024,
#         "args" =
#           list(
#             "x" = 2,
#             "y" = 10,
#             "default" = TRUE
#             )
#       )
#     expect_equal(
#       temp(),
#       defaultOutput
#     )
#   }
# )

#' A simulated data set based on a standardized test.
#'
#' Simulated response patterns, abilities, and outcomes based on a uni-dimensional state-issued standardized test.
#'
#'  @format A data frame of 1000 rows (observations) and 58 columns (variables):
#'  \describe{
#'  \item{Outcome}{a binary external criterion variable correlated with TrueAbility}
#'  \item{TrueAbility}{the simulated true ability parameter used to generate response patterns}
#'  \item{Item1-Item56}{binary responses to items generated using the TrueAbility parameters and simulated 3PL item parameters generated from the distribution of parameters estimated from a state-issued standardized test}
#'  }
"simulated_test_data"

#' Model syntax for the example in the \code{\link{antcolony.lavaan}} function.
#'
#' A character vector containing the model syntax used for the one factor, 56-item example in the \code{\link{antcolony.lavaan}}.
#'
#' @format A character vector.
"exampleAntModel"

#' Model syntax for the short example in the \code{\link{antcolony.lavaan}} function.
#'
#' A character vector containing the model syntax used for the one factor, 15-item, example in the \code{\link{antcolony.lavaan}}.
#'
#' @format A character vector.
"shortExampleAntModel"

#' A function to implement the ant colony optimization algorithm for short form
#' specification searches with the package \link[lavaan]{lavaan}.
#'
#' @description The Ant Colony Optimization (ACO) algorithm (Dorigo & Stutzle,
#'  2004) can produce short forms of scales that are optimized with respect to
#'  characteristics selected by the developer, such as model fit and predictive
#'  relationships with other variables. The algorithm is based on the foraging
#'  behavior of a group of ants, which start searching for food in a variety of
#'  directions and then eventually all ants converge to the shortest distance to
#'  the food source. This behavior occurs because ants leave a pheronome trail
#'  behind as they search for food and ants in shorter paths leave stronger
#'  pheronome trails, which are detected by other ants and that will lead them
#'  to follow the shortest trail.
#'
#' @details This function sends a specified number of ants per iteration, which
#'  randomly select items to build a model, then evaluates the model based on
#'  pheromone levels. The pheromone levels are updated after each iteration
#'  according to the best-fitting model of that iteration. The algorithm's
#'  stopping rule is to end the search when a certain solution is the same for a
#'  given number of ants in a row.
#'
#'  PREPARATORY STEPS: For the ACO algorithm implementation for short for
#'  selection, the following decisions are needed:
#'
#'  1. Determine the target size for the short form.
#'
#'  2. Determine which characteristics should be optimized.
#'
#'  3. Define how the pheronome level will be computed: This is a function of
#'  the characteristics of the short form that will be optimized. In Leite,
#'  Huang and Marcoulides (2008), the pheromone level was zero if model fit
#'  indices did not meet Hu and Bentler's (1999) suggested thresholds, and equal
#'  to the sum of path coefficients of a predictor variable if model fit indices
#'  met thresholds. Currently, the package only implements pheromone calculation
#'  based on regression coefficients or variance explained, with user-selected
#'  model fit index thresholds.
#'
#'  4. Define how many short forms should be evaluated before the best-so-far
#'  pheronome level is examined. Leite, Huang and Marcoulides (2008) used 10
#'  short forms.
#'
#'  5. Define the percentage of pheronome evaporation, if any. Leite, Huang and
#'  Marcoulides (2008) used 5\%.
#'
#'  6. Define convergence criterion. Leite, Huang and Marcoulides (2008) set the
#'  algorithm to converge if the short form did not improve in 100 x number of
#'  short forms in step 4.
#'
#'  IMPLEMENTATION: Once these decisions are made, the ACO algorithm selects
#'  short forms with the following steps:
#'
#'  Step 1. All items are assigned an initial weight of 1.
#'
#'  Step 2. A set of n short forms is selected by sampling with probability
#'  proportional to the item weights.
#'
#'  Step 3. Fit the latent variable model to the n short forms.
#'
#'  Step 4. Calculate the pheromone levels for the n short forms. Define the
#'  best-so-far pheronome level (if iteration 1) or compare the current best
#'  pheronome from the set of n short forms to the best-so-far pheronome.
#'
#'  Step 5. If the pheromone level of the best short form from step 4 exceeds
#'  the best-so-far pheronome level, update the best-so-far pheromone level and
#'  add it to the current weight of the items of the best short form.
#'
#'  Step 6. Return to step 2 until convergence criterion is reached.
#'
#' @param data The data being used in data frame format. Default value is
#'  \code{null}. Only one of \code{data} or \code{sample.cov} should be used.
#' @param sample.cov The sample covariance matrix. See \link[lavaan]{lavaan} for
#'  the specific format needed. Default value is \code{null}. Only one of
#'  \code{data} or \code{sample.cov} should be used.
#' @param sample.nobs A numeric value indicating the number of observations in
#'  the sample covariance matrix. If \code{sample.cov} is used, this must be
#'  filled in. Default value is \code{null}.
#' @param ants A numeric value indicating the number of ants to send (e.g.,
#'  number of short forms to evaluate) per iteration. Default value is 20.
#' @param evaporation A numeric value which sets the percentage of the pheromone
#'  that is retained after evaporation between steps of the algorithm. Default
#'  value is 0.9, indicating 10% evaporation. Should be within the range of
#'  (0,1), exclusive.
#' @param antModel The lavaan formatted model. See \link[lavaan]{lavaan} for more
#'  details. Defaults to the default \link[lavaan]{lavaan} values. NOTE: Each factor
#'  and/or regression needs to be specified on a single line. Newline breaks and
#'  carriage returns WILL break the function.
#' @param list.items A list containing one or more character vectors of item
#'  names for each factor, where each factor is a separate element of the list.
#'  The items should be input in the order in which the factors are input in
#'  \code{i.per.f} and \code{factors}.
#' @param full A numeric value indicating the total number of unique items in the
#'  test or scale.
#' @param i.per.f Vector with number of items per factor (e.g. target number), in
#'  the same order of \code{list.items} and \code{factors}.
#' @param factors Character vector with names of factors in the same order of
#'  \code{list.items} and \code{i.per.f}.
#' @param bifactor Either the name of the factor that all of the chosen items
#' will load on (as character), or `NULL` if the model is not a bifactor model.
#' @param steps A numeric value that sets the stopping rule, which is the number
#'  of ants in a row for which the model does not change.
#' @param lavaan.model.specs A list which contains the specifications for the
#'  lavaan model. The default values are the defaults for lavaan to perform a
#'  CFA. These are automatically set internally, then updated by the user-provided
#'  values. Note that this drastically affects the algorithm, and care must be 
#'  taken to ensure that the algorithm can fit valid models as it searches for
#'  the best model. See the default arguments for examples of what you can change
#'  and \link[lavaan]{lavaan} for more details on what arguments are available 
#'  to change.
#' @param pheromone.calculation A character string specifying the method for
#'  calculating the pheromone strength. Must be one of "\code{gamma}"
#'  (standardized latent regression coefficients), "\code{beta}"
#'  (standardized observed regression coefficients), "\code{regression}"
#'  (both latent and observed regression coefficients, if they exist)
#'   or "\code{variance}" (proportion of
#'  variance explained by model). You must specify the entire string. Default is
#'  \code{gamma}.
#' @param fit.indices The fit indices (in lavaan format) extracted for model
#'  optimization. See \link[lavaan]{lavaan} for more details.
#' @param fit.statistics.test A character vector of the logical test being used
#'  for model optimization. The default is \code{"(cfi > 0.95)&(tli >
#'  0.95)&(rmsea < 0.06)"}. The format for the logical test should match 1) the
#'  names of the indices being used in \link[lavaan]{lavaan} and 2) the default
#'  provided above. At least one fit index must be included.
#' @param summaryfile The name of the summary file generated. A .txt file is
#'  suggested. Default is "summary.txt" and writes into the current working
#'  directory. This file writes a line for each ant within each step and
#'  includes (a) a vector of a 0/1 value for each item indicating whether the
#'  item was selected by that ant, (b) the run number, (c) the count number, (d)
#'  the ant number, and (e) the current pheromone level.
#' @param feedbackfile The name of the feedback file generated. An .html file is
#'  suggested. Default is "iteration.html" and writes into the current working
#'  directory. This file saves the result of each run, which includes (a) the
#'  run number, (b) the count number, (c) the ant number, (d) the step number
#'  (if the current run is successful) or "Failure" (if the current run is
#'  unsuccessful), and for successful runs (f) the chosen fit statistics (from
#'  \code{fit.indices}), the average of the gammas and betas (standardized regression
#'  coefficients), and the overall variance explained of the current run.
#' @param max.run The maximum number of ants to run before the algorithm stops.
#'  This includes failed iterations as well. Default is 1000.
#' @param parallel An option for using parallel processing. If \code{TRUE}, the
#'  function will utilize all available cores (up to the number of ants). Default
#'  is \code{TRUE}.
#' @return A list with four elements: the first containing a named matrix with
#'  final model's best fit indices, the final pheromone level (either the mean
#'  of the standardized regression coefficients (gammas, betas, or both), or the mean variance
#'  explained), and a series of 0/1 values indicating the items selected in the
#'  final solution,  the second element containing tbe summary matrix of the
#'  best fit statistic value(s) for each run, the items chosen for said best fit,
#'  the mean gamma, beta, and variance explained for the best fit, and the item pheromone
#'  levels after each run, the third containing the best-fitting lavaan model
#'  object, and the fourth containing the best-fitting model syntax.
#'
#' @family Ant Colony Algorithms
#' @seealso \code{\link{antcolony.mplus}}
#' @examples
#' # a 3-factor example using the HolzingerSwineford1939 data from `lavaan`
#'
#' # some changes to the default values
#' # notice that in this example we are recreating the original model
#' abilityShortForm <- antcolony.lavaan(
#'   data = lavaan::HolzingerSwineford1939,
#'   ants = 2, evaporation = 0.7,
#'   antModel = " visual  =~ x1 + x2 + x3
#'              textual =~ x4 + x5 + x6
#'              speed   =~ x7 + x8 + x9 ",
#'   list.items = list(c(
#'     "x1",
#'     "x2", "x3"
#'   ), c("x4", "x5", "x6"), c("x7", "x8", "x9")), full = 9, i.per.f =
#'     c(3, 3, 3), factors = c("visual", "textual", "speed"), steps = 2, fit.indices =
#'     c("cfi"), fit.statistics.test = "(cfi > 0.6)", summaryfile =
#'     NULL, feedbackfile = NULL, max.run = 2, parallel = FALSE
#' )
#' \dontrun{
#' # using simulated test data and the default values for lavaan.model.specs
#' # first, read in the original or "full" model
#' data(exampleAntModel) # a character vector for a lavaan model
#'
#' # then, create the list of the items by the factors
#' # in this case, all items load onto the general 'Ability' factor
#' list.items <- list(c(
#'   "Item1", "Item2", "Item3", "Item4", "Item5",
#'   "Item6", "Item7", "Item8", "Item9", "Item10",
#'   "Item11", "Item12", "Item13", "Item14", "Item15",
#'   "Item16", "Item17", "Item18", "Item19", "Item20",
#'   "Item21", "Item22", "Item23", "Item24", "Item25",
#'   "Item26", "Item27", "Item28", "Item29", "Item30",
#'   "Item31", "Item32", "Item33", "Item34", "Item35",
#'   "Item36", "Item37", "Item38", "Item39", "Item40",
#'   "Item41", "Item42", "Item43", "Item44", "Item45",
#'   "Item46", "Item47", "Item48", "Item49", "Item50",
#'   "Item51", "Item52", "Item53", "Item54", "Item55", "Item56"
#' ))
#'
#' # load the data
#' data(simulated_test_data)
#'
#' # finally, call the function with some minor changes to the default values.
#' abilityShortForm <- antcolony.lavaan(
#'   data = simulated_test_data,
#'   ants = 5, evaporation = 0.7, antModel = exampleAntModel,
#'   list.items = list.items, full = 56, i.per.f = 20,
#'   factors = "Ability", steps = 3, fit.indices = c("cfi", "rmsea"),
#'   fit.statistics.test = "(cfi > 0.95)&(rmsea < 0.05)",
#'   summaryfile = "summary.txt",
#'   feedbackfile = "iteration.html",
#'   max.run = 500
#' )
#'
#' abilityShortForm # print the results of the final short form
#' 
#' # an example using binary (ordered) data
#' # create the simulated full model and model data
#' sim_model <- "
#' f1 =~ x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9 + x10
#' f2 =~ x11 + x12 + x13 + x14 + x15 + x16 + x17 + x18 + x19 + x20
#' f3 =~ x21 + x22 + x23 + x24 + x25 + x26 + x27 + x28 + x29 + x30"
#' 
#' sim_data <-
#'   cbind(
#'     psych::sim.rasch(nvar = 10)$items,
#'     psych::sim.rasch(nvar = 10)$items,
#'     psych::sim.rasch(nvar = 10)$items
#'   )
#' 
#' colnames(sim_data) = paste0("x", 1:30)
#' # fit with antcolony.lavaan
#' # note that ONLY the estimator and ordered args
#' # of lavaan.model.specs are changed. This retains
#' # the default args, fitting a CFA but with ordered data.
#' example <-
#' antcolony.lavaan(
#'   data = sim_data,
#'   ants = 5, evaporation = 0.7,
#'   antModel = sim_model,
#'   lavaan.model.specs = 
#'     list(estimator = "wlsmv", ordered = T),
#'   list.items = list(paste0('x', 1:10),
#'                     paste0('x', 11:20),
#'                     paste0('x', 21:30)), 
#'   full = 30, 
#'   i.per.f = c(5, 5, 5), 
#'   factors = c("f1", "f2", "f3"),
#'   steps = 20, 
#'   fit.indices = c("cfi.scaled"), 
#'   fit.statistics.test = "(cfi.scaled > 0.90)", 
#'   summaryfile = NULL, 
#'   feedbackfile = NULL, 
#'   max.run = 500, 
#'   parallel = T
#' )
#' # note that this example will take a bit of time to run
#' # as ordered data factor analysis is computationally expensive.
#' }
#' @import lavaan utils
#' @export
#' @author Anthony W Raborn, \email{anthony.w.raborn@@gmail.com}

antcolony.lavaan <- function(data = NULL, sample.cov = NULL, sample.nobs = NULL,
                             ants = 20, evaporation = 0.9, antModel, list.items = NULL,
                             full = NULL, i.per.f = NULL, factors = NULL, bifactor = NULL, steps = 50,
                             lavaan.model.specs = list(
                               model.type = "cfa", auto.var = T, estimator = "default",
                               ordered = NULL, int.ov.free = TRUE, int.lv.free = FALSE,
                               auto.fix.first = TRUE, auto.fix.single = TRUE, auto.var = TRUE,
                               auto.cov.lv.x = TRUE, auto.th = TRUE, auto.delta = TRUE,
                               auto.cov.y = TRUE, std.lv = F,
                               group = NULL, group.label = NULL, group.equal = 'loadings',
                               group.partial = NULL, group.w.free = FALSE
                             ),
                             pheromone.calculation = "gamma", fit.indices = c("cfi", "tli", "rmsea"),
                             fit.statistics.test = "(cfi > 0.95)&(tli > 0.95)&(rmsea < 0.06)",
                             summaryfile = NULL,
                             feedbackfile = NULL, max.run = 1000, parallel = T) {
  if (!requireNamespace("lavaan", quietly = TRUE)) {
    stop("The `lavaan` package is required to use this function. Please install `lavaan`, then try to use this function again.")
  }
  fitmeasuresCheck(fit.indices)
  antcolony.lavaan.env <- new.env(parent = baseenv())

  if (pheromone.calculation %in% c("gamma", "beta", "regression", "variance") == FALSE) {
    stop("Pheromone calculation not recognized! Enter one of \'gamma\', \'beta\', \'regression\' or \'variance\'.")
  }
  # create initial, empty files to be used
  if (length(summaryfile) > 0) {
    write(x = "", file = summaryfile)
  }

  summaryObject <- matrix(
    nrow = 1,
    ncol = (full + 3 + 3 + length(fit.indices) + full)
  )

  if (length(feedbackfile) > 0) {
    write(x = "", file = feedbackfile)
  }
  # creates the table of initial pheromone levels.
  include <- rep(2, full)
  # puts initial best solution (all items selected).
  best.so.far.solution <- include

  # creates a vector with all items. UNIQUE USED FOR CASES WHEN ITEMS CROSS-LOAD
  item.vector <- unique(unlist(list.items, use.names = F))
  if (!is.null(bifactor)) {
    item.vector <- item.vector[which(item.vector != bifactor)]
  }

  # reads the Lavaan model syntax input into the function
  input <- unlist(strsplit(antModel, "\n"))

  # creates a list to store factors.
  selected.items <- list.items

  # starts counting the iterations
  count <- 1

  # starts counting continuous runs regardless of result.
  run <- 1

  # defines initial best so far (overall) pheromone
  best.so.far.pheromone <- 0
  # defines initial best pheromone for the current trial of n ants.
  best.pheromone <- 0
  # defines initial solutions.
  previous.solution <- include
  step <- 1

  # creates objects in the function environment that are fed into the lavaan function in order to fine-tune the model to user specifications
  # use a default set of specifications that fits a CFA
  default.lavaan.model.specs = list(
    model.type = "cfa", auto.var = T, estimator = "default",
    ordered = T, int.ov.free = TRUE, int.lv.free = FALSE, auto.fix.first = TRUE,
    auto.fix.single = TRUE, auto.var = TRUE, auto.cov.lv.x = TRUE, auto.th = TRUE,
    auto.delta = TRUE, auto.cov.y = TRUE, std.lv = F, group = NULL, group.label = NULL,
    group.equal = "loadings", group.partial = NULL, group.w.free = FALSE
  )
  mapply(assign, names(default.lavaan.model.specs), default.lavaan.model.specs, MoreArgs = list(envir = antcolony.lavaan.env))
  # overwrite with user-provided definitions
  mapply(assign, names(lavaan.model.specs), lavaan.model.specs, MoreArgs = list(envir = antcolony.lavaan.env))

  # create values of "bad warnings" and "bad errors" that result in uninterpretable models
  bad.warnings <- c(
    "could not compute standard errors",
    "could not compute scaled test statistic",
    "covariance matrix of latent variables is not positive definite",
    "model has NOT converged",
    "could not invert information matrix",
    "the optimizer warns that a solution has NOT been found",
    "some estimated ov variances are negative"
  )
  bad.errors <- c(
    "initial model-implied matrix (Sigma) is not positive definite",
    "missing observed variables in dataset"
  )

  chk <- Sys.getenv("_R_CHECK_LIMIT_CORES_", "")

  if (parallel) {
    if (nzchar(chk) && chk == "TRUE") {
      # use 2 cores in CRAN/Travis/AppVeyor
      num_workers <- 2L
    } else {
      # use all cores in devtools::test()
      num_workers <- parallel::detectCores()
    }

    `%dopar%` <- foreach::`%dopar%`
    cl <- parallel::makeCluster(num_workers,type="PSOCK", outfile = "")
    doSNOW::registerDoSNOW(cl)

  } else {
    `%dopar%` <- foreach::`%do%`
    num_workers = 1

  }

  ant = 0L
  progress <- function(n) {
    cat(paste("\r Run number ", run, " and ant number ", n, ".           ", sep = ""))
  }
  opts <- list(progress = progress)

  start.time <- Sys.time()


  # starts loop through iterations.
    while (run <= max.run) {
      antResults <-
        foreach::foreach(ant = 1:ants, .inorder = F, .combine = rbind, .options.snow = opts, .errorhandling = 'remove') %dopar% {

        # selects items for all factors.
        newModelList <- antcolonyNewModel(
          itemList = list.items,
          itemVector = item.vector,
          includedItems = include,
          model = input,
          itemCount = i.per.f,
          factorNames = factors,
          bifactor
        )

        mapply(assign, names(newModelList), newModelList, MoreArgs = list(envir = antcolony.lavaan.env))


        selected.items <- lapply(antcolony.lavaan.env$selected.items, sort)
        selected.vector <- unlist(antcolony.lavaan.env$selected.items, use.names = F)
        select.indicator <- is.element(item.vector, selected.vector)

        # MODIFY LAVAAN SYNTAX
        new_ant_model <- 
          input
        for (factor in 1:length(factors)) {
          temp_factor_definition <-
            new_ant_model[grepl(factors[factor], new_ant_model)]
          temp_factor_definition <-
            sub(
              "=~[[:alnum:][:space:] +]{1,}", 
              paste0("=~ ", paste0(selected.items[[factor]], collapse = " + ")), 
              temp_factor_definition
              )
          new_ant_model[grepl(factors[factor], new_ant_model)] <-
            temp_factor_definition
        }

        # Run the model check function
        # checks for and saves error/warning messages within the lavaan output,
        # as well as the fit indices
        modelCheck <- modelWarningCheck(
          lavaan::lavaan(
            model = new_ant_model, data = data, sample.cov = sample.cov,
            sample.nobs = sample.nobs,
            model.type = antcolony.lavaan.env$model.type,
            ordered = antcolony.lavaan.env$ordered,
            estimator = antcolony.lavaan.env$estimator,
            int.ov.free = antcolony.lavaan.env$int.ov.free,
            int.lv.free = antcolony.lavaan.env$int.lv.free,
            auto.fix.first = antcolony.lavaan.env$auto.fix.first,
            std.lv = antcolony.lavaan.env$std.lv,
            auto.fix.single = antcolony.lavaan.env$auto.fix.single,
            auto.var = antcolony.lavaan.env$auto.var,
            auto.cov.lv.x = antcolony.lavaan.env$auto.cov.lv.x,
            auto.th = antcolony.lavaan.env$auto.th,
            auto.delta = antcolony.lavaan.env$auto.delta,
            auto.cov.y = antcolony.lavaan.env$auto.cov.y
          ),
          modelSyntax = new_ant_model
        )

        # Save the error and warning messages
        warnings <- modelCheck@warnings
        errors <- modelCheck@errors
        # Check the above messages and set pheromone to zero under 'bad' circumstances
        if (length(warnings) > 0 | length(errors) > 0) {
          if (grepl(paste0(bad.errors, collapse = "|"), errors, ignore.case = T) ||
              (grepl(paste0(bad.warnings, collapse = "|", warnings, ignore.case = T)))) {
            pheromone <- 0
            
            # writes feedback about non-convergence and non-positive definite.
            if (length(summaryfile) > 0) {
              fit.info <- matrix(c(select.indicator, run, count, ant, 999, 999, round((include), 5)), 1, )
              write.table(fit.info,
                          file = summaryfile, append = T,
                          quote = F, sep = " ", row.names = F, col.names = F
              )
            }
            
            # provide feedback about search.
            if (length(feedbackfile) > 0) {
              feedback <- c(paste("<h1>", run, "-", count, "-", ant, "-", step, "- Failure", "</h1>"))
              write(feedback, file = feedbackfile, append = T)
            }
            # finishes if for non-convergent cases.
          }
        }
         else {
          modelInfo <- modelInfoExtract(
            modelCheckObj = modelCheck,
            fitIndices = fit.indices
          )

          mapply(assign, names(modelInfo), modelInfo, MoreArgs = list(envir = antcolony.lavaan.env))
          mapply(assign, names(antcolony.lavaan.env$model.fit), antcolony.lavaan.env$model.fit, MoreArgs = list(envir = antcolony.lavaan.env))

          # provide feedback about search.
          if (length(feedbackfile) > 0) {
            feedback <- c(paste(
              "<h1>", "run:", run, "count:", count, "ant:", ant, "step:", step, "<br>",
              "Fit Statistics:", antcolony.lavaan.env$model.fit, "<br>",
              "GAMMA:", mean(antcolony.lavaan.env$std.gammas),
              "BETA:", mean(antcolony.lavaan.env$std.betas),
              "VAR.EXP:", mean(antcolony.lavaan.env$variance.explained), "</h1>"
            ))
            write(feedback, file = feedbackfile, append = T)
          }

          # implements fit requirement.
          if (eval(parse(text = fit.statistics.test),
                   envir = antcolony.lavaan.env
          ) == FALSE) {
            # Model didn't fit well enough, so set pheromone to 0.
            pheromone <- 0
          } else {
            # Model fit well enough, so calculate pheromone by either gamma or variance.
            if (pheromone.calculation == "gamma") { # mean of standardized gammas
              pheromone <- round(mean(antcolony.lavaan.env$std.gammas, na.rm = T), 3)
            } else {
              if (pheromone.calculation == "beta") { # mean of standardized betas
                pheromone <- round(mean(antcolony.lavaan.env$std.betas, na.rm = T), 3)
              } else {
                if (pheromone.calculation == "regression") { # mean of all regression coefs
                  pheromone <- round(mean(antcolony.lavaan.env$std.reg.coef, na.rm = T), 3)
                }
                if (pheromone.calculation == "variance") { # mean of r^2 values
                  pheromone <- round(mean(antcolony.lavaan.env$variance.explained, na.rm = T), 3)
                }
              }
            }
          }

          # end else clause for converged solutions
        }

        returnMatrix = list(
          'solution' = select.indicator,
          'run' = run,
          'count' = count,
          'ant' = ant,
          'model.fit' = antcolony.lavaan.env$model.fit,
          'pheromone' = pheromone,
          'mean.std.gammas' = mean(antcolony.lavaan.env$std.gammas),
          'mean.std.betas' = mean(antcolony.lavaan.env$std.betas),
          'mean.var.exp' = mean(antcolony.lavaan.env$variance.explained),
          'model.output' = modelCheck@model.output,
          'model.syntax' = new_ant_model
        )

        returnMatrix
      }

      # implements pheromone evaporation.
      include <- include * evaporation

      bestAnt <-
        which(unlist(antResults[,'pheromone'])==
                max(unlist(antResults[,'pheromone'])))[[1]]
      best.pheromone <-
        antResults[[bestAnt,'pheromone']]


      # adjusts pheromone and best.so.far values only if the current pheromone is as good or better than the previous.
      if (best.pheromone >= best.so.far.pheromone) {
        include.pheromone <- antResults[[bestAnt, 'solution']] * best.pheromone
        include <- include + include.pheromone

        best.so.far.solution <- as.numeric(antResults[[bestAnt, 'solution']])
        best.so.far.pheromone <- best.pheromone
        best.so.far.fit.indices <- antResults[[bestAnt, 'model.fit']]
        best.so.far.model <- antResults[[bestAnt, 'model.output']]
        best.so.far.syntax <- antResults[[bestAnt, 'model.syntax']]

        if (!all(sapply(antResults[-1,'solution'], FUN = identical, antResults[1, 'solution']))) {
          # re-starts count.
          count <- 1
        }
      } else {

        # advances count.
        count <- count + ants

        # adds more pheromone to the best so far solution.
        include.pheromone <- best.so.far.solution * best.so.far.pheromone

        # updates pheromone.
        include <- include + include.pheromone
      }


      # ends loop.
      run <- run + 1
      summaryObject <-
        rbind(
          summaryObject,
          matrix(
            c(
              as.numeric(antResults[[bestAnt, 'solution']]),
              run,
              bestAnt,
              count,
              if (length(antResults[[bestAnt, 'model.fit']]) < length(fit.indices)) {
                rep(NA, times = length(fit.indices))
              } else {
                antResults[[bestAnt, 'model.fit']]
              },
              antResults[[bestAnt, 'mean.std.gammas']],
              antResults[[bestAnt, 'mean.std.betas']],
              antResults[[bestAnt, 'mean.var.exp']],
              include
            ),
           nrow = 1
           )
        )
        if (count >= steps) {
          break
        }

    }

  if (parallel) {
    foreach::registerDoSEQ()
    parallel::stopCluster(cl)
  }

  print("Compiling results.")

  summaryObject <- data.frame(summaryObject)[-1, ]
  colnames(summaryObject) <-
    c(item.vector,
      "run",
      "ant",
      "count",
      fit.indices,
      "mean.gamma",
      "mean.beta",
      "mean.var.exp",
      paste0(item.vector, ".Pheromone")
      )

  final.solution <-
    matrix(
      c(best.so.far.fit.indices, best.so.far.pheromone, best.so.far.solution),
      1,
      dimnames = list(
        NULL,
        c(names(best.so.far.fit.indices), paste0("mean_", pheromone.calculation), item.vector)
        )
  )

  results <-
    new(
      'ACO',
      function_call = match.call(),
      summary = summaryObject,
      final_solution = final.solution,
      best_model = best.so.far.model,
      best_syntax = best.so.far.syntax,
      runtime = Sys.time() - start.time
    )

  results
}

setClassUnion("charactorORdata.frame", c("character","data.frame"))

#' An S4 class for the Tabu Search Algorithm
#'
#' @slot function_call The original function call.
#' @slot all_fit A summary `vector` indicating the model fit results for
#' each iteration.
#' @slot best_fit The best model fit result using the selected `fitStatistic`. A numeric value or vector, possibly named.
#' @slot best_model A `lavaan` object of the final solution.
#' @slot best_syntax A `character` vector of the final solution model syntax.
#' @slot runtime A `difftime` object of the total run time of the function.
#' @slot final_tabu_list The final list of Tabu models. Each element of the list is a `lavaan` object. 
#'
#' @importFrom methods new show
#'
#' @return An S4 object of class `TS`.
#' @export
#'
setClass('TS',
         slots =
           list(
             function_call = 'call',
             all_fit = 'vector',
             best_fit = 'numeric',
             best_model = 'lavaan',
             best_syntax = 'charactorORdata.frame',
             runtime = 'ANY',
             final_tabu_list = 'list'
           )
)

#' Print method for class `TS`
#' 
#' @param object An S4 object of class `TS`.
#' 
#' @export
setMethod('show',
          signature = 'TS',
          definition = function(object) {
            # check if the lavaan_partable was returned (a data.frame)
            if (is.data.frame(object@best_syntax)) {
              # convert to model.syntax for printing
              factors = unique(object@best_syntax$lhs)
              model.syntax = c()
              for (i in factors) {
                factorRows = which(object@best_syntax$lhs==i&
                                     object@best_syntax$op=="=~"&
                                     object@best_syntax$free==1)
                model.syntax = c(model.syntax,
                                 paste0(i, " =~ ", 
                                        paste0(object@best_syntax$rhs[factorRows],
                                               collapse = " + "
                                               )
                                        )
                                 )
              }
              for (i in factors) {
                factorRows = which(object@best_syntax$lhs==i&
                                     object@best_syntax$op=="~~"&
                                     object@best_syntax$free==1)
                if (length(factorRows) > 0) {
                  model.syntax = c(model.syntax,
                                   paste0(i, " ~~ ", 
                                          paste0(object@best_syntax$rhs[factorRows],
                                                 collapse = " + "
                                          )
                                   )
                  )
                }
              }
              model.syntax = paste0(model.syntax, collapse = "\n")
              } else (model.syntax = object@best_syntax)
            line0 = c("Algorithm: Tabu Search")
            line1 = paste0(
              "Total Run Time: ",
              round(object@runtime[[1]], 3), 
              " ", 
              attr(object@runtime, "units"),
              "\n"
            )
            line2 = suppressWarnings(
              stringr::str_wrap(
                as.vector(c("Function call:\n", object@function_call, "\n")), 
                exdent = 2
              )
            )
            line3 = paste0(
              stringr::str_wrap(
                c("Final Model Syntax:", 
                  unlist(strsplit(model.syntax, '\n'))), 
                exdent = 2), 
              collapse = "\n"
            )
            to_console = paste0(c(line0, line1, line2, line3), collapse = "\n")
            cat(to_console)
          }
)

#' Plot method for class `TS`
#' 
#' @param x,y An S4 object of class `TS`.
#' @param ... Not used.
#' 
#' @export
setMethod("plot",
          signature = "TS",
          definition = function(x, y, ...) {

            fit_values <- as.numeric(x@all_fit)
            iterations <- seq_along(fit_values) - 1

            if (length(fit_values) == 0 || all(is.na(fit_values))) {
              plot.new()
              title(main = "Changes in Model Fit Value per Iteration")
              text(
                x = 0.5,
                y = 0.5,
                labels = "No fit values available",
                cex = 1
              )
              box()

              return(invisible(x))
            }

            plot(
              x = iterations,
              y = fit_values,
              type = "l",
              xlab = "Iteration",
              ylab = paste("Model Fit Value(s):", names(x@best_fit)),
              main = "Changes in Model Fit Value per Iteration",
              ...
            )

            box()

            invisible(x)
          }
        )

#' Summary method for class `TS`
#' 
#' @param object An S4 object of class `TS`.
#' 
#' @export
setMethod('summary',
          signature = 'TS',
          definition = function(object) {
            # check if the lavaan_partable was returned (a data.frame)
            if (is.data.frame(object@best_syntax)) {
              # convert to model.syntax for printing
              factors = unique(object@best_syntax$lhs)
              model.syntax = c()
              for (i in factors) {
                factorRows = which(object@best_syntax$lhs==i&
                                     object@best_syntax$op=="=~"&
                                     object@best_syntax$free==1)
                model.syntax = c(model.syntax,
                                 paste0(i, " =~ ", 
                                        paste0(object@best_syntax$rhs[factorRows],
                                               collapse = " + "
                                        )
                                 )
                )
              }
              for (i in factors) {
                factorRows = which(object@best_syntax$lhs==i&
                                     object@best_syntax$op=="~~"&
                                     object@best_syntax$free==1)
                if (length(factorRows) > 0) {
                  model.syntax = c(model.syntax,
                                   paste0(i, " ~~ ", 
                                          paste0(object@best_syntax$rhs[factorRows],
                                                 collapse = " + "
                                          )
                                   )
                  )
                }
              }
              model.syntax = paste0(model.syntax, collapse = "\n")
            } else (model.syntax = object@best_syntax)
            
            line0 = c("Algorithm: Tabu Search")
            line1 = paste0(
              "Total Run Time: ",
              round(object@runtime[[1]], 3), 
              " ", 
              attr(object@runtime, "units"),
              "\n"
            )
            line2 = c(capture.output(print(object@best_model)), "\n")
            line3 = paste0(
              stringr::str_wrap(
                c("\nFinal Model Syntax:", 
                  unlist(strsplit(model.syntax, "\n"))), 
                exdent = 2), 
              collapse = "\n"
            )
            to_console = paste0(c(line0, line1, line2, line3), collapse = "\n")
            cat(to_console)
          }
)

# A copy of the internal lavaan functions
# it is copied here to help pass the automated CRAN checks
# all code here is owned by the various authors of the lavaan package, with minor modifications

lav_partable_full <- function(partable = NULL,
                              strict.exo = FALSE, free = FALSE,
                              start = FALSE) {
  stopifnot(
    !is.null(partable$lhs), !is.null(partable$op),
    !is.null(partable$rhs)
  )
  meanstructure <- any(partable$op == "~1")
  nblocks <- lav_partable_nblocks(partable)
  lv.names <- vnames(partable, type = "lv")
  ov.names <- vnames(partable, type = "ov")
  ov.names.x <- vnames(partable, type = "ov.x")
  ov.names.nox <- vnames(partable, type = "ov.nox")
  lv.names.x <- vnames(partable, type = "lv.x")
  ov.names.y <- vnames(partable, type = "ov.y")
  lv.names.y <- vnames(partable, type = "lv.y")
  lvov.names.y <- c(ov.names.y, lv.names.y)
  ov.names.ord <- vnames(partable, type = "ov.ord")
  ov.names.ind <- vnames(partable, type = "ov.ind")
  l.lhs <- r.rhs <- op <- character(0)
  l.lhs <- rep(lv.names, each = length(ov.names.nox))
  l.rhs <- rep(ov.names.nox, times = length(lv.names))
  l.op <- rep("=~", length(l.lhs))
  ov.lhs <- ov.rhs <- ov.op <- character(0)
  OV <- ov.names.nox
  nx <- length(OV)
  idx <- lower.tri(matrix(0, nx, nx), diag = TRUE)
  ov.lhs <- rep(OV, each = nx)[idx]
  ov.rhs <- rep(OV, times = nx)[idx]
  ov.op <- rep("~~", length(ov.lhs))
  if (!strict.exo && length(ov.names.x) > 0L) {
    OV <- ov.names.x
    nx <- length(OV)
    idx <- lower.tri(matrix(0, nx, nx), diag = TRUE)
    more.lhs <- rep(OV, each = nx)[idx]
    more.rhs <- rep(OV, times = nx)[idx]
    ov.lhs <- c(ov.lhs, more.lhs)
    ov.rhs <- c(ov.rhs, more.rhs)
    ov.op <- c(ov.op, rep("~~", length(more.lhs)))
  }
  lv.lhs <- lv.rhs <- lv.op <- character(0)
  nx <- length(lv.names)
  idx <- lower.tri(matrix(0, nx, nx), diag = TRUE)
  lv.lhs <- rep(lv.names, each = nx)[idx]
  lv.rhs <- rep(lv.names, times = nx)[idx]
  lv.op <- rep("~~", length(lv.lhs))
  r.lhs <- r.rhs <- r.op <- character(0)
  if (any(partable$op == "~")) {
    eqs.names <- unique(c(
      partable$lhs[partable$op == "~"],
      partable$rhs[partable$op == "~"]
    ))
    eqs.y <- eqs.names
    if (strict.exo) {
      x.idx <- which(eqs.names %in% ov.names.x)
      if (length(x.idx) > 0L) {
        eqs.y <- eqs.names[-x.idx]
      }
    }
    eqs.x <- eqs.names
    r.lhs <- rep(eqs.y, each = length(eqs.x))
    r.rhs <- rep(eqs.x, times = length(eqs.y))
    idx <- which(r.lhs == r.rhs)
    r.lhs <- r.lhs[-idx]
    r.rhs <- r.rhs[-idx]
    bad.idx <- which(r.lhs %in% ov.names.ind & r.rhs %in%
      lv.names)
    if (length(bad.idx) > 0L) {
      r.lhs <- r.lhs[-bad.idx]
      r.rhs <- r.rhs[-bad.idx]
    }
    r.op <- rep("~", length(r.rhs))
  }
  int.lhs <- int.rhs <- int.op <- character(0)
  if (meanstructure) {
    if (strict.exo) {
      int.lhs <- c(ov.names.nox, lv.names)
    }
    else {
      int.lhs <- c(ov.names, lv.names)
    }
    int.rhs <- rep("", length(int.lhs))
    int.op <- rep("~1", length(int.lhs))
  }
  th.lhs <- th.rhs <- th.op <- character(0)
  if (length(ov.names.ord) > 0L) {
    tmp <- strsplit(
      vnames(partable, "th"),
      "\\|"
    )
    th.lhs <- sapply(tmp, function(x) x[1])
    th.rhs <- sapply(tmp, function(x) x[2])
    th.op <- rep("|", length(th.lhs))
  }
  delta.lhs <- delta.rhs <- delta.op <- character(0)
  if (nblocks > 1L && length(ov.names.ord) > 0L) {
    delta.lhs <- ov.names.ord
    delta.rhs <- ov.names.ord
    delta.op <- rep("~*~", length(delta.lhs))
  }
  lhs <- c(l.lhs, ov.lhs, lv.lhs, r.lhs, int.lhs, th.lhs, delta.lhs)
  rhs <- c(l.rhs, ov.rhs, lv.rhs, r.rhs, int.rhs, th.rhs, delta.rhs)
  op <- c(l.op, ov.op, lv.op, r.op, int.op, th.op, delta.op)
  block <- 1L
  if (nblocks > 1) {
    block <- rep(1:nblocks, each = length(lhs))
    lhs <- rep(lhs, times = nblocks)
    op <- rep(op, times = nblocks)
    rhs <- rep(rhs, times = nblocks)
  }
  LIST <- data.frame(
    lhs = lhs, op = op, rhs = rhs, block = block,
    stringsAsFactors = FALSE
  )
  if (free) {
    LIST$free <- rep(0L, nrow(LIST))
  }
  if (start) {
    LIST$start <- rep(0, nrow(LIST))
  }
  LIST
}

vnames <- function(partable, type = NULL, ..., warn = FALSE, ov.x.fatal = FALSE) {
  if (length(partable$lhs) == 0) {
    return(character(0L))
  }
  dotdotdot <- list(...)
  type.list <- c(
    "ov", "ov.x", "ov.nox", "ov.model", "ov.y",
    "ov.num", "ov.ord", "ov.ind", "ov.orphan", "ov.interaction",
    "th", "th.mean", "lv", "lv.regular", "lv.formative",
    "lv.x", "lv.y", "lv.nox", "lv.nonnormal", "lv.interaction",
    "eqs.y", "eqs.x"
  )
  stopifnot(is.list(partable), !missing(type), type %in% c(
    type.list,
    "all"
  ))
  if (length(type) == 1L && type == "all") {
    type <- type.list
  }
  if (is.null(partable$block)) {
    partable$block <- rep(1L, length(partable$lhs))
  }
  nblocks <- lav_partable_nblocks(partable)
  block.select <- lav_partable_block_values(partable)
  ndotdotdot <- length(dotdotdot)
  if (ndotdotdot > 0L) {
    dot.names <- names(dotdotdot)
    block.select <- rep(TRUE, length(partable$lhs))
    for (dot in seq_len(ndotdotdot)) {
      block.var <- dot.names[dot]
      block.val <- dotdotdot[[block.var]]
      if (is.null(partable[[block.var]])) {
        stop(
          "lavaan ERROR: selection variable `", block.var,
          " not found in the parameter table."
        )
      }
      else {
        if (!all(block.val %in% partable[[block.var]])) {
          stop(
            "lavaan ERROR: ", block.var, " column does not contain value `",
            block.val, "'"
          )
        }
        block.select <- (block.select & partable[[block.var]] %in%
          block.val)
      }
    }
    block.select <- unique(partable$block[block.select])
    if (length(block.select) == 0L) {
      warnings("lavaan WARNING: no blocks selected.")
    }
  }
  OUT <- vector("list", length = nblocks)
  OUT$ov <- vector("list", length = nblocks)
  OUT$ov.x <- vector("list", length = nblocks)
  OUT$ov.nox <- vector("list", length = nblocks)
  OUT$ov.model <- vector("list", length = nblocks)
  OUT$ov.y <- vector("list", length = nblocks)
  OUT$ov.num <- vector("list", length = nblocks)
  OUT$ov.ord <- vector("list", length = nblocks)
  OUT$ov.ind <- vector("list", length = nblocks)
  OUT$ov.orphan <- vector("list", length = nblocks)
  OUT$ov.interaction <- vector("list", length = nblocks)
  OUT$th <- vector("list", length = nblocks)
  OUT$th.mean <- vector("list", length = nblocks)
  OUT$lv <- vector("list", length = nblocks)
  OUT$lv.regular <- vector("list", length = nblocks)
  OUT$lv.formative <- vector("list", length = nblocks)
  OUT$lv.x <- vector("list", length = nblocks)
  OUT$lv.y <- vector("list", length = nblocks)
  OUT$lv.nox <- vector("list", length = nblocks)
  OUT$lv.nonnormal <- vector("list", length = nblocks)
  OUT$lv.interaction <- vector("list", length = nblocks)
  OUT$eqs.y <- vector("list", length = nblocks)
  OUT$eqs.x <- vector("list", length = nblocks)
  for (b in block.select) {
    lv.names <- unique(partable$lhs[partable$block == b &
      (partable$op == "=~" | partable$op == "<~")])
    int.names <- unique(partable$rhs[partable$block == b &
      grepl(":", partable$rhs)])
    n.int <- length(int.names)
    if (n.int > 0L) {
      ok.idx <- logical(n.int)
      for (iv in seq_len(n.int)) {
        NAMES <- strsplit(int.names[iv], ":", fixed = TRUE)[[1L]]
        if (sum(NAMES %in% lv.names) > 0L) {
          ok.idx[iv] <- TRUE
        }
      }
      lv.interaction <- int.names[ok.idx]
      lv.names <- c(lv.names, lv.interaction)
    }
    else {
      lv.interaction <- character(0L)
    }
    if ("lv" %in% type) {
      OUT$lv[[b]] <- lv.names
    }
    if ("lv.regular" %in% type) {
      out <- unique(partable$lhs[partable$block == b &
        partable$op == "=~"])
      OUT$lv.regular[[b]] <- out
    }
    if ("lv.interaction" %in% type) {
      OUT$lv.interaction[[b]] <- lv.interaction
    }
    if ("lv.formative" %in% type) {
      out <- unique(partable$lhs[partable$block == b &
        partable$op == "<~"])
      OUT$lv.formative[[b]] <- out
    }
    if (!(length(type) == 1L && type %in% c(
      "lv", "lv.regular",
      "lv.nonnormal"
    ))) {
      eqs.y <- unique(partable$lhs[partable$block == b &
        partable$op == "~"])
    }
    if ("eqs.y" %in% type) {
      OUT$eqs.y[[b]] <- eqs.y
    }
    if (!(length(type) == 1L && type %in% c(
      "lv", "lv.regular",
      "lv.nonnormal", "lv.x"
    ))) {
      eqs.x <- unique(partable$rhs[partable$block == b &
        (partable$op == "~" | partable$op == "<~")])
    }
    if ("eqs.x" %in% type) {
      OUT$eqs.x[[b]] <- eqs.x
    }
    if (!(length(type) == 1L && type %in% c(
      "lv", "lv.regular",
      "lv.nonnormal"
    ))) {
      v.ind <- unique(partable$rhs[partable$block == b &
        partable$op == "=~"])
    }
    if (!(length(type) == 1L && type %in% c(
      "lv", "lv.regular",
      "lv.nonnormal", "lv.x", "lv.y"
    ))) {
      ov.ind <- v.ind[!v.ind %in% lv.names]
      ov.y <- eqs.y[!eqs.y %in% c(lv.names, ov.ind)]
      ov.x <- eqs.x[!eqs.x %in% c(lv.names, ov.ind, ov.y)]
    }
    if (!(length(type) == 1L && type %in% c(
      "lv", "lv.regular",
      "lv.nonnormal", "lv.x", "lv.y"
    ))) {
      ov.cov <- c(partable$lhs[partable$block == b & partable$op ==
        "~~" & !partable$lhs %in% lv.names], partable$rhs[partable$block ==
        b & partable$op == "~~" & !partable$rhs %in%
        lv.names])
      ov.int <- partable$lhs[partable$block == b & (partable$op ==
        "~1" | partable$op == "|") & !partable$lhs %in%
        lv.names]
      ov.tmp <- c(ov.ind, ov.y, ov.x)
      ov.extra <- unique(c(ov.cov, ov.int))
      ov.names <- c(ov.tmp, ov.extra[!ov.extra %in% ov.tmp])
    }
    if ("ov" %in% type) {
      OUT$ov[[b]] <- ov.names
    }
    if ("ov.ind" %in% type) {
      OUT$ov.ind[[b]] <- ov.ind
    }
    if ("ov.interaction" %in% type) {
      ov.int.names <- ov.names[grepl(":", ov.names)]
      n.int <- length(ov.int.names)
      if (n.int > 0L) {
        ov.names.noint <- ov.names[!ov.names %in% ov.int.names]
        ok.idx <- logical(n.int)
        for (iv in seq_len(n.int)) {
          NAMES <- strsplit(ov.int.names[iv], ":", fixed = TRUE)[[1L]]
          if (all(NAMES %in% ov.names.noint)) {
            ok.idx[iv] <- TRUE
          }
        }
        ov.interaction <- ov.int.names[ok.idx]
      }
      else {
        ov.interaction <- character(0L)
      }
      OUT$ov.interaction[[b]] <- ov.interaction
    }
    if (any(type %in% c(
      "ov.x", "ov.nox", "ov.num", "ov.model",
      "th.mean", "lv.nonnormal"
    ))) {
      if (is.null(partable$user)) {
        partable$user <- rep(1L, length(partable$lhs))
      }
      vars <- c(
        partable$lhs[partable$block == b & partable$op ==
          "~1" & partable$user == 1], partable$lhs[partable$block ==
          b & partable$op == "~~" & partable$user == 1],
        partable$rhs[partable$block == b & partable$op ==
          "~~" & partable$user == 1]
      )
      idx.no.x <- which(ov.x %in% vars)
      if (length(idx.no.x)) {
        if (ov.x.fatal) {
          stop(
            "lavaan ERROR: model syntax contains variance/covariance/intercept formulas\n  involving (an) exogenous variable(s): [",
            paste(ov.x[idx.no.x], collapse = " "), "];\n  Please remove them and try again."
          )
        }
        if (warn) {
          warning(
            "lavaan WARNING: model syntax contains variance/covariance/intercept formulas\n  involving (an) exogenous variable(s): [",
            paste(ov.x[idx.no.x], collapse = " "), "];\n  Please use fixed.x=FALSE or leave them alone"
          )
        }
        ov.x <- ov.x[-idx.no.x]
      }
      ov.tmp.x <- ov.x
      if (!is.null(partable$exo)) {
        ov.cov <- c(
          partable$lhs[partable$block == b &
            partable$op == "~~" & partable$exo == 1L],
          partable$rhs[partable$block == b & partable$op ==
            "~~" & partable$exo == 1L]
        )
        ov.int <- partable$lhs[partable$block == b &
          partable$op == "~1" & partable$exo == 1L]
        ov.extra <- unique(c(ov.cov, ov.int))
        ov.tmp.x <- c(ov.tmp.x, ov.extra[!ov.extra %in%
          ov.tmp.x])
      }
      ov.names.x <- ov.tmp.x
    }
    if ("ov.x" %in% type) {
      OUT$ov.x[[b]] <- ov.names.x
    }
    if ("ov.orphan" %in% type) {
      OUT$ov.orphan[[b]] <- ov.extra
    }
    if (any(type %in% c(
      "ov.nox", "ov.num", "ov.model",
      "th.mean", "lv.nonnormal"
    ))) {
      ov.names.nox <- ov.names[!ov.names %in% ov.names.x]
    }
    if ("ov.nox" %in% type) {
      OUT$ov.nox[[b]] <- ov.names.nox
    }
    if ("ov.model" %in% type) {
      if (any(partable$block == b & partable$op == "~" &
        partable$exo == 1L)) {
        OUT$ov.model[[b]] <- ov.names.nox
      }
      else {
        OUT$ov.model[[b]] <- ov.names
      }
    }
    if (any(type %in% c(
      "ov.ord", "th", "th.mean", "ov.num",
      "lv.nonnormal"
    ))) {
      tmp <- unique(partable$lhs[partable$block == b &
        partable$op == "|"])
      ord.names <- ov.names[ov.names %in% tmp]
    }
    if ("ov.ord" %in% type) {
      OUT$ov.ord[[b]] <- ord.names
    }
    if (any(type %in% c("ov.num", "lv.nonnormal"))) {
      ov.num <- ov.names.nox[!ov.names.nox %in% ord.names]
    }
    if ("ov.num" %in% type) {
      OUT$ov.num[[b]] <- ov.num
    }
    if ("lv.nonnormal" %in% type) {
      lv.reg <- unique(partable$lhs[partable$block ==
        b & partable$op == "=~"])
      if (length(lv.reg) > 0L) {
        out <- unlist(lapply(lv.reg, function(x) {
          tmp.ind <- unique(partable$rhs[partable$block ==
            b & partable$op == "=~" & partable$lhs ==
            x])
          if (!all(tmp.ind %in% ov.num)) {
            return(x)
          }
          else {
            return(character(0))
          }
        }))
        OUT$lv.nonnormal[[b]] <- out
      }
      else {
        OUT$lv.nonnormal[[b]] <- character(0)
      }
    }
    if (any(c("th", "th.mean") %in% type)) {
      TH.lhs <- partable$lhs[partable$block == b & partable$op ==
        "|"]
      TH.rhs <- partable$rhs[partable$block == b & partable$op ==
        "|"]
    }
    if ("th" %in% type) {
      if (length(ord.names) > 0L) {
        out <- unlist(lapply(ord.names, function(x) {
          idx <- which(x == TH.lhs)
          TH <- unique(paste(TH.lhs[idx], "|", TH.rhs[idx],
            sep = ""
          ))
          sort(TH)
        }))
      }
      else {
        out <- character(0L)
      }
      OUT$th[[b]] <- out
    }
    if ("th.mean" %in% type) {
      if (length(ov.names.nox) > 0L) {
        out <- unlist(lapply(ov.names.nox, function(x) {
          if (x %in% ord.names) {
            idx <- which(x == TH.lhs)
            TH <- unique(paste(TH.lhs[idx], "|", TH.rhs[idx],
              sep = ""
            ))
            sort(TH)
          }
          else {
            x
          }
        }))
      }
      else {
        out <- character(0L)
      }
      OUT$th.mean[[b]] <- out
    }
    if (any(c("lv.x", "lv.nox") %in% type)) {
      tmp <- lv.names[!lv.names %in% c(v.ind, eqs.y)]
      lv.names.x <- lv.names[lv.names %in% tmp]
    }
    if ("lv.x" %in% type) {
      OUT$lv.x[[b]] <- lv.names.x
    }
    if ("ov.y" %in% type) {
      tmp <- eqs.y[!eqs.y %in% c(v.ind, eqs.x, lv.names)]
      OUT$ov.y[[b]] <- ov.names[ov.names %in% tmp]
    }
    if ("lv.y" %in% type) {
      tmp <- eqs.y[!eqs.y %in% c(v.ind, eqs.x) & eqs.y %in%
        lv.names]
      OUT$lv.y[[b]] <- lv.names[lv.names %in% tmp]
    }
    if ("lv.nox" %in% type) {
      OUT$lv.nox[[b]] <- lv.names[!lv.names %in% lv.names.x]
    }
  }
  if (length(type) == 1L) {
    OUT <- OUT[[type]]
    if (ndotdotdot == 0L) {
      OUT <- unique(unlist(OUT))
    }
    else if (length(block.select) == 1L) {
      OUT <- OUT[[block.select]]
    }
    else {
      OUT <- OUT[block.select]
    }
  }
  else {
    OUT <- OUT[type]
  }
  OUT
}

lav_partable_nblocks <- function(partable) {
  if (is.null(partable$block)) {
    nblocks <- 1L
  }
  else {
    tmp <- partable$block[partable$block > 0L]
    nblocks <- length(unique(stats::na.omit(tmp)))
  }
  nblocks
}

lav_partable_block_values <- function(partable) {
  if (is.null(partable$block)) {
    block.values <- 1L
  }
  else {
    tmp <- partable$block[partable$block > 0L]
    block.values <- unique(stats::na.omit(tmp))
  }
  block.values
}

#' \code{ShortForm} package
#'
#' Automated Item Selection Algorithms for Short Forms
#'
#' See the README on \href{https://github.com/AnthonyRaborn/ShortForm#readme}{GitHub} for more information.
#'
#' @name ShortForm
#' @aliases ShortForm-package
#' @importFrom stats runif
NULL

## quiets concerns of R CMD check about internal variables
if (getRversion() >= "2.15.1") {
  utils::globalVariables(c(
    "allItems", "auto.cov.lv.x", "auto.cov.y", "auto.delta", "auto.fix.first",
    "auto.fix.single", "auto.th", "auto.var", "bestModel", "currentModel", "estimator",
    "factors", "int.lv.free", "int.ov.free", "itemsPerFactor", "model.type", "numItems",
    "std.lv"
  ))
}

#' Create Package Startup Message
#'
#' Makes package startup message.
#'
#' Idea taken from https://github.com/ntguardian/MCHT/blob/master/R/StartupMessage.R
#'
#' @import utils
#' @examples
#' ShortForm:::ShortFormStartup()
ShortFormStartup <- function() {
  ShortForm <- c("  #####                             #######                      \n #     # #    #  ####  #####  ##### #        ####  #####  #    # \n #       #    # #    # #    #   #   #       #    # #    # ##  ## \n  #####  ###### #    # #    #   #   #####   #    # #    # # ## # \n       # #    # #    # #####    #   #       #    # #####  #    # \n #     # #    # #    # #   #    #   #       #    # #   #  #    # \n  #####  #    #  ####  #    #   #   #        ####  #    # #    # \n ")
  version <- paste("\t\t Version", as.character(utils::packageVersion("ShortForm")))
  penguin <- c("\t\t\t (o<", "\t\t\t //\\", "\t\t\t V_/_ ")

  message <- c(ShortForm, version, penguin)

  paste0(message, collapse = "\n")
}

#' Package Attach Hook Function
#'
#' Hook triggered when package attached.
#'
#' Idea taken from https://github.com/ntguardian/MCHT/blob/master/R/StartupMessage.R
#'
#' @param lib a character string giving the library directory where the package
#'            defining the namespace was found
#' @param pkg a character string giving the name of the package
#' @examples
#' ShortForm:::.onAttach(.libPaths()[1], "ShortForm")
.onAttach <- function(lib, pkg) {
  msg <- ShortFormStartup()
  if (!interactive()) {
    msg[1] <- paste("Package 'ShortForm' version", packageVersion("ShortForm"))
  }
  packageStartupMessage(msg)
  invisible()
}

#' An S4 class for the Ant Colony Optimization Algorithm
#'
#' @slot function_call The original function call.
#' @slot summary A summary `data.frame` indicating the algorithm results for
#' each iteration.
#' @slot final_solution A `matrix` with the final solution information, including
#' fit indices, selected items, and pheromone level.
#' @slot best_model A `lavaan` object of the final solution.
#' @slot best_syntax A `character` vector of the final solution model syntax.
#' @slot runtime A `difftime` object of the total run time of the function.
#'
#' @importFrom methods new show
#'
#' @return An S4 object of class `ACO`.
#' @export
#'
setClass('ACO',
         slots =
           list(
             function_call = 'call',
             summary = 'data.frame',
             final_solution = 'matrix',
             best_model = 'lavaan',
             best_syntax = 'character',
             runtime = 'ANY'
           )
)

#' Print method for class `ACO`
#'
#' @param object An S4 object of class `ACO`
#'
#' @export
setMethod('show',
          signature = 'ACO',
          definition = function(object) {
            line0 = c("Algorithm: Ant Colony Optimization")
            line1 = paste0(
              "Total Run Time: ",
              round(object@runtime[[1]], 3),
              " ",
              attr(object@runtime, "units"),
              "\n"
            )
            line2 = suppressWarnings(
              stringr::str_wrap(
                as.vector(c("Function call:\n", object@function_call, "\n")),
                exdent = 2
                )
              )
            line3 = paste0(
              stringr::str_wrap(
                c("Final Model Syntax:",
                  unlist(strsplit(object@best_syntax, '\n'))),
                exdent = 2),
              collapse = "\n"
              )
            to_console = paste0(c(line0, line1, line2, line3), collapse = "\n")
            cat(to_console)
          }
          )

#' Plot method for class `ACO`
#'
#' @param x,y An S4 object of class `ACO`
#' @param type A `character` value specifying the plot type. One of `'all'`
#' (for all plots), `'pheromone'`, `'gamma'`, `'beta'`, or `'variance'`.
#' @param ... Not used.
#' 
#' @importFrom grDevices hcl.colors
#' @importFrom graphics box polygon text plot.new title
#' @importFrom rlang .data
#' 
#' @export
setMethod("plot",
          signature = "ACO",
          definition = function(x, y, type = c("all", "pheromone", "gamma", "beta", "variance"), ...) {

            type <- match.arg(type)

            summary_results <- x@summary

            if (!"run" %in% names(summary_results)) {
              stop("x@summary must contain a 'run' column.")
            }

            run <- summary_results$run

            item_pheromone_names <- grep("Pheromone", names(summary_results), value = TRUE)

            old_par <- par(no.readonly = TRUE)
            on.exit(par(old_par), add = TRUE)

            plot_empty_panel <- function(main_title, message) {
              plot.new()
              title(main = main_title)
              text(
                x = 0.5,
                y = 0.5,
                labels = message,
                cex = 1
              )
              box()
              invisible(FALSE)
            }

            plot_pheromone <- function() {

              if (length(item_pheromone_names) == 0) {
                plot_empty_panel(
                  main_title = "Changes in Pheromone",
                  message = "No pheromone columns found"
                )
                return(invisible(FALSE))
              }

              pheromone_mat <- as.matrix(summary_results[, item_pheromone_names, drop = FALSE])
              pheromone_mat[is.na(pheromone_mat)] <- 0

              if (nrow(pheromone_mat) == 0 || ncol(pheromone_mat) == 0) {
                plot_empty_panel(
                  main_title = "Changes in Pheromone",
                  message = "No pheromone values available"
                )
                return(invisible(FALSE))
              }

              cumulative_pheromone <- t(apply(pheromone_mat, 1, cumsum))

              y_max <- max(cumulative_pheromone, na.rm = TRUE)

              if (!is.finite(y_max)) {
                plot_empty_panel(
                  main_title = "Changes in Pheromone",
                  message = "No finite pheromone values available"
                )
                return(invisible(FALSE))
              }

              plot(
                run,
                cumulative_pheromone[, ncol(cumulative_pheromone)],
                type = "n",
                ylim = c(0, y_max),
                xlab = "Run",
                ylab = "Total Pheromone",
                main = "Changes in Pheromone",
                ...
              )

              cols <- hcl.colors(ncol(pheromone_mat), palette = "Set 3")

              lower <- rep(0, length(run))

              for (i in seq_len(ncol(pheromone_mat))) {
                upper <- cumulative_pheromone[, i]

                polygon(
                  x = c(run, rev(run)),
                  y = c(upper, rev(lower)),
                  col = cols[i],
                  border = "black",
                  lwd = 0.1
                )

                lower <- upper
              }

              box()
              invisible(TRUE)
            }

            plot_metric <- function(column, ylab, main_title) {

              if (!column %in% names(summary_results)) {
                plot_empty_panel(
                  main_title = main_title,
                  message = paste0("Column '", column, "' not found")
                )
                return(invisible(FALSE))
              }

              y_values <- summary_results[[column]]

              if (
                length(y_values) == 0 ||
                length(y_values) != length(run) ||
                all(is.na(y_values))
              ) {
                plot_empty_panel(
                  main_title = main_title,
                  message = paste0("Column '", column, "' is empty")
                )
                return(invisible(FALSE))
              }

              finite_values <- is.finite(y_values) & is.finite(run)

              if (!any(finite_values)) {
                plot_empty_panel(
                  main_title = main_title,
                  message = paste0("Column '", column, "' has no finite values")
                )
                return(invisible(FALSE))
              }

              plot(
                run,
                y_values,
                type = "l",
                xlab = "Run",
                ylab = ylab,
                main = main_title,
                ...
              )

              endpoint_runs <- range(run[finite_values], na.rm = TRUE)

              endpoint_index <- which(
                run %in% endpoint_runs &
                  is.finite(y_values)
              )

              if (length(endpoint_index) > 0) {
                text(
                  x = run[endpoint_index],
                  y = y_values[endpoint_index],
                  labels = round(y_values[endpoint_index], 3),
                  pos = rep(c(4, 2), length.out = length(endpoint_index)),
                  cex = 0.8
                )
              }

              box()
              invisible(TRUE)
            }

            if (type == "all") {

              par(mfrow = c(2, 2))

              plot_pheromone()

              plot_metric(
                column = "mean.gamma",
                ylab = expression("Mean " * gamma),
                main_title = expression("Changes in Mean " * gamma)
              )

              plot_metric(
                column = "mean.beta",
                ylab = expression("Mean " * beta),
                main_title = expression("Changes in Mean " * beta)
              )

              plot_metric(
                column = "mean.var.exp",
                ylab = "Mean Variance Explained",
                main_title = "Changes in Mean Variance Explained"
              )

            } else if (type == "pheromone") {

              plot_pheromone()

            } else if (type == "gamma") {

              plot_metric(
                column = "mean.gamma",
                ylab = expression("Mean " * gamma),
                main_title = expression("Changes in Mean " * gamma)
              )

            } else if (type == "beta") {

              plot_metric(
                column = "mean.beta",
                ylab = expression("Mean " * beta),
                main_title = expression("Changes in Mean " * beta)
              )

            } else if (type == "variance") {

              plot_metric(
                column = "mean.var.exp",
                ylab = "Mean Variance Explained",
                main_title = "Changes in Mean Variance Explained"
              )
            }

            invisible(x)
          }
        )

#' Summary method for class `ACO`
#'
#' @param object An S4 object of class `ACO`
#'
#' @export
setMethod('summary',
          signature = 'ACO',
          definition = function(object) {
            line0 = c("Algorithm: Ant Colony Optimization")
            line1 = paste0(
              "Total Run Time: ",
              round(object@runtime[[1]], 3),
              " ",
              attr(object@runtime, "units"),
              "\n"
            )
            line2 = c(capture.output(print(object@best_model)), "\n")
            line3 = paste0(
              stringr::str_wrap(
                c("\nFinal Model Syntax:",
                  unlist(strsplit(object@best_syntax, "\n"))),
                exdent = 2),
              collapse = "\n"
            )
            to_console = paste0(c(line0, line1, line2, line3), collapse = "\n")
            cat(to_console)
          }
)

antcolonyNewModel <- 
  function(itemList, 
           itemVector,
           includedItems, 
           model, 
           itemCount, 
           factorNames, 
           bifactor = NULL
           ) {
  all.items <- c()

  selected.items <- vector("list", length(itemList))

  if (!is.character(bifactor)) {
    for (factor in 1:length(itemList)) {

      # selects the items for a short form for the factor
      positions <- is.element(itemVector, itemList[[factor]])
      prob <- includedItems[positions] / sum(includedItems[positions])

      items <- sample(itemList[[factor]], size = itemCount[factor], replace = F, prob)

      # stores selected items.
      selected.items[[factor]] <- items


      # replaces the lavaan syntax for factor specification.
      factor.position <- grep(paste(factorNames[factor], "[ ]{0,}=~"), model, ignore.case = T)
      model[factor.position] <- paste(factorNames[factor], "=~", paste(items, collapse = " + "))
      all.items <- c(all.items, items)
    }
  } else {
    bifactor.items <- c()
    for (factor in 1:(length(itemList) - 1)) {
      # selects the items for a short form for the factor
      positions <- is.element(itemVector, itemList[[factor]])
      prob <- includedItems[positions] / sum(includedItems[positions])

      items <- sample(itemList[[factor]], size = itemCount[factor], replace = F, prob)

      # stores selected items.
      selected.items[[factor]] <- items
      bifactor.items <- c(bifactor.items, items)

      # replaces the lavaan syntax for factor specification.
      factor.position <- grep(paste(factorNames[factor], "=~"), model, ignore.case = T)
      model[factor.position] <- paste(factorNames[factor], "=~", paste(items, collapse = " + "))
      all.items <- c(all.items, items)
    }

    # add bifactor items
    model[grep(paste(bifactor, "=~"), model)] <- paste(bifactor, "=~", paste(all.items, collapse = " + "))
    selected.items[[length(itemList)]] <- bifactor.items
  }

  return(list(
    "input" = model,
    "selected.items" = selected.items,
    "all.items" = all.items
  ))
}

modelInfoExtract <- 
  function(modelCheckObj, 
           fitIndices
           ) {

  # first, fit indices
  model.fit <-
    lavaan::fitMeasures(modelCheckObj@model.output, 
                        fitIndices
                        )

  # next, gamma/beta/variances
  # estimate the standardized coefficients of the variables
  standard.coefs <- 
    lavaan::standardizedSolution(modelCheckObj@model.output, 
                                 se = FALSE, 
                                 zstat = FALSE, 
                                 pvalue = FALSE, 
                                 remove.def = TRUE
                                 )
  # extract the regression coefficients
  std.gammas <-
    standard.coefs[which(standard.coefs[, 2] == "=~"), ]$est.std
  std.betas <-
    standard.coefs[which(standard.coefs[, 2] == "~"), ]$est.std
  std.reg.coef <-
    standard.coefs[which(standard.coefs[, 2] == "~" | standard.coefs[, 2] == "=~"), ]$est.std

  # obtains the variance explained ("rsquare") from lavaan
  variance.explained <-
    lavaan::lavInspect(modelCheckObj@model.output, "rsquare")

  return(list(
    "model.fit" = model.fit,
    "std.gammas" = std.gammas,
    "std.betas" = std.betas,
    "std.reg.coef" = std.reg.coef,
    "variance.explained" = variance.explained
  ))
}

#' Given a fitted lavaan model (e.g., CFA), prepares a table that contains
#' parameters that can be fixed/freed as part of a model specification search.
#'
#' @param fitted.model - an object of class "lavaan" that contains the initially fitted model for the search
#' @param loadings - a logical value that indicates whether cross-loadings will be part of the search
#' @param fcov - a logical value indicating whether factor covariances will be part of the search
#' @param errors - a logical value indicating whether error covariances will be part of the search
#'
#' @return A \code{data.frame} with lavaan-formatted parameter values.
#' @family Tabu Search
#' @import lavaan
#' @export
#'
#' @examples
#' \dontrun{
#' # load simulation data and select columns used in this example
#' data(simulated_test_data)
#' tabuData <- simulated_test_data[, c(1:10)]
#'
#' # specify an improper model (improper because data is unidimensional)
#' tabuModel <- "
#' Ability =~ Item1 + Item2 + Item3 + Item4
#' FakeAbility =~ Item5 + Item6 + Item7 + Item8
#' Ability ~ Outcome
#' FakeAbility ~ 0*Outcome"
#'
#' # run the initial misspecified model for Tabu
#'
#' init.model <- lavaan::lavaan(
#'   model = tabuModel, data = tabuData,
#'   auto.var = TRUE, auto.fix.first = FALSE, std.lv = TRUE, auto.cov.lv.x = TRUE
#' )
#'
#' # Use search.prep to prepare for the Tabu search
#' ptab <- search.prep(fitted.model = init.model, loadings = TRUE, fcov = TRUE, errors = FALSE)
#'
#' # add an additional (mispecified) parameter
#' additional.param <- "Item1 ~~ 0.5*Item3"
#' ptab <- add.param(fitted.model = init.model, ptab = ptab, syntax = additional.param)
#'
#' # Perform Tabu Search
#' trial <- tabu.sem(init.model = init.model, ptab = ptab, obj = AIC, niter = 2, tabu.size = 5)
#' }
#' @author Carl F. Falk
#' @references \doi{10.1080/10705511.2017.1409074}

search.prep <- function(fitted.model, loadings = TRUE, fcov = TRUE, errors = FALSE) {

  # Parameter table from fitted model
  ptab <- lavaan::parTable(fitted.model)

  # Full parameter table
  fulltab <- lav_partable_full(ptab)

  # Merge together so we have what's free
  mergetab <- lavaan::lav_partable_merge(ptab, fulltab, remove.duplicated = TRUE, warn = FALSE)

  idx <- vector("numeric")
  null.val <- vector("numeric")

  lv.names <- vnames(ptab, type = "lv") # extract names of latent vars
  ov.names <- vnames(ptab, type = "ov.nox") # indicators

  # Loadings
  if (loadings) {
    # Index for which rows of table are loadings
    indx <- mergetab$op == "=~"

    # Omit any loadings that are fixed to a non-zero value
    # These may be fixed for identification
    indx[mergetab$free[indx] == 0 & mergetab$ustart[indx] != 0] <- FALSE

    idx <- c(idx, which(indx))
    null.val <- c(null.val, rep(0, sum(indx)))
  }

  # Factor cov
  if (fcov) {
    # Index for which rows are cov among latent vars
    indx <- mergetab$op == "~~" & mergetab$lhs %in% lv.names & mergetab$lhs != mergetab$rhs
    idx <- c(idx, which(indx))
    null.val <- c(null.val, rep(0, sum(indx)))
  }

  # Error covariances among indicators
  if (errors) {
    # Index for which rows are cov among indicators
    indx <- mergetab$op == "~~" & mergetab$lhs %in% ov.names & mergetab$lhs != mergetab$rhs
    idx <- c(idx, which(indx))
    null.val <- c(null.val, rep(0, sum(indx)))
  }

  # Check whether any parameters in the table are involved in any constraints
  flag <- check.const(fitted.model, mergetab[idx, ])

  if (flag) {
    warning("Some parameters in this table may be involved in (in)equality constraints")
  }

  # Clean up
  mergetab <- mergetab[idx, c("lhs", "op", "rhs", "block", "free", "label")]
  mergetab$nullval <- null.val
  mergetab$free <- ifelse(mergetab$free == 0, 0, 1)

  # Sort table
  lvnames <- lavaan::lavNames(mergetab, "lv")
  ovnames <- lavaan::lavNames(mergetab, "ov")
  mergetab <- mergetab[order(
    mergetab$block, # by group (or block)
    is.na(match(mergetab$lhs, c(lvnames))), # put latent vars first
    mergetab$op, # by op w/in lv and ov sections
    match(mergetab$lhs, c(lvnames, ovnames)), # then by lv and ov names on lhs
    match(mergetab$rhs, c(lvnames, ovnames))
  ), ] # then by same on rhs
  rownames(mergetab) <- 1:nrow(mergetab)

  ret <- mergetab

  return(ret)
}

#' Short Form Tabu Search
#'
#' Given an initial (full) lavaan model string, the original data, a criterion
#' function to minimize, and some additional specifications,
#' performs a Tabu model specification search. Currently only supports
#' neighbors that are 1 move away from the current model.
#'
#'
#' @param initialModel The initial model (typically the full form) as a character vector with lavaan model.syntax.
#' @param originalData The original data frame with variable names.
#' @param numItems A numeric vector indicating the number of items to retain for each factor.
#' @param criterion A function calculating the objective criterion to minimize. Default is to use the built-in `rmsea` value from `lavaan::fitmeasures()`.
#' @param niter A numeric value indicating the number of iterations (model specification selections)
#' to perform. Default is 50.
#' @param tabu.size A numeric value indicating the size of Tabu list. Default is 5.
#' @param lavaan.model.specs A list which contains the specifications for the
#'  lavaan model. The default values are the defaults for lavaan to perform a
#'  CFA. See \link[lavaan]{lavaan} for more details.
#' @param bifactor Logical. Indicates if the latent model is a bifactor model. If `TRUE`, assumes that the last latent variable in the provided model syntax is the bifactor (i.e., all of the retained items will be set to load on the last latent variable).
#' @param verbose Logical. If `TRUE`, prints out the initial short form and the selected short form at the end of each iteration.
#' @param parallel An option for using parallel processing. If \code{TRUE}, the 
#'  function will utilize all available cores. Default is \code{TRUE}.
#'
#' @return A named list with the best value of the objective function (`best.obj`) and the best lavaan model object (`best.mod`).
#' @export
#'
#' @examples
#' shortAntModel <- "
#' Ability =~ Item1 + Item2 + Item3 + Item4 + Item5 + Item6 + Item7 + Item8
#' Ability ~ Outcome
#' "
#'
#' data(simulated_test_data)
#' tabuResult <- tabuShortForm(
#'   initialModel = shortAntModel,
#'   originalData = simulated_test_data, numItems = 7,
#'   niter = 1, tabu.size = 3, parallel = FALSE
#' )
#' summary(tabuResult) # shows the resulting model
#' \dontrun{
#' # create simulation data from the `psych` package
#' # four factors, 12 items each, 48 total items
#' # factor loading matrix - not quite simple structure
#' fxMatrix <-
#'   matrix(
#'     data = c(
#'       rep(x = c(.9, .7, .5, .3), times = 3),
#'       rep(0.2, times = 3 * 4 * 3), # first factor loadings
#'
#'       rep(0.2, times = 3 * 4),
#'       rep(x = c(.9, .7, .5, .3), times = 3),
#'       rep(0.2, times = 3 * 4 * 2), # second factor loadings
#'
#'       rep(0.2, times = 3 * 4 * 2),
#'       rep(x = c(.9, .7, .5, .3), times = 3),
#'       rep(0.2, times = 3 * 4), # third factor loadings
#'
#'       rep(0.2, times = 3 * 4 * 3),
#'       rep(x = c(.9, .7, .5, .3), times = 3) # fourth factor loadings
#'     ),
#'     ncol = 4
#'   )
#' # factor correlation matrix - all factors uncorrelated
#' PhiMatrix <-
#'   matrix(data = c(
#'     1, 0, 0, 0,
#'     0, 1, 0, 0,
#'     0, 0, 1, 0,
#'     0, 0, 0, 1
#'   ), ncol = 4)
#' tabuData <-
#'   psych::sim(
#'     fx = fxMatrix,
#'     Phi = PhiMatrix,
#'     n = 1000,
#'     raw = TRUE
#'   )$observed # observed is the simulated observed data
#'
#' # NOTE: you must specify the model such that each factor is on a single line!
#' # otherwise, the algorithm will not work correctly!
#' tabuModel <- "
#' Trait1 =~ Item1 + Item2 + Item3 + Item4 + Item5 + Item6 +
#' Item7 + Item8 + Item9 + Item10 + Item11 + Item12
#' Trait2 =~ Item13 + Item14 + Item15 + Item16 + Item17 +
#' Item18 + Item19 + Item20 + Item21 + Item22 + Item23 + Item24
#' Trait3 =~ Item25 + Item26 + Item27 + Item28 + Item29 + Item30 +
#' Item31 + Item32 + Item33 + Item34 + Item35 + Item36
#' Trait4 =~ Item37 + Item38 + Item39 + Item40 + Item41 +
#' Item42 + Item43 + Item44 + Item45 + Item46 + Item47 + Item48
#' "
#'
#' colnames(tabuData) <- paste0("Item", 1:48)
#' # specify the criterion function that the Tabu Search minimizes
#' # wrap this in a tryCatch in case a model does not converge!
#' # specify an appropriate error value: if minimizing, error value must be large
#' tabuCriterion <- function(x) {
#'   tryCatch(lavaan::fitmeasures(object = x, fit.measures = "chisq"),
#'     error = function(e) Inf
#'   )
#' }
#'
#' # use the tabuShortForm function
#' # reduce form to the best 12 items
#' tabuShort <- tabuShortForm(
#'   initialModel = tabuModel, originalData = tabuData,
#'   numItems = c(3, 3, 3, 3),
#'   criterion = tabuCriterion,
#'   niter = 20, tabu.size = 10
#' )
#' }
#'
tabuShortForm <- function(originalData,
                           initialModel,
                           numItems,
                           criterion = function(x) {
                             tryCatch(-lavaan::fitmeasures(object = x, fit.measures = "cfi"),
                               error = function(e) Inf
                             )
                           },
                           niter = 20,
                           tabu.size = 5,
                           lavaan.model.specs = list(
                             int.ov.free = TRUE,
                             int.lv.free = FALSE,
                             std.lv = TRUE,
                             auto.fix.first = FALSE,
                             auto.fix.single = TRUE,
                             auto.var = TRUE,
                             auto.cov.lv.x = TRUE,
                             auto.th = TRUE,
                             auto.delta = TRUE,
                             auto.cov.y = TRUE,
                             ordered = NULL,
                             model.type = "cfa",
                             estimator = "default"
                           ),
                           bifactor = FALSE,
                           verbose = FALSE,
                           parallel = T) {
  start.time = Sys.time()
  mapply(
    assign,
    names(lavaan.model.specs),
    lavaan.model.specs,
    MoreArgs = list(envir = environment())
  )

  allItems <-
    colnames(originalData)

  # extract the latent factor syntax
  mapply(
    assign,
    c("factors", "itemsPerFactor"),
    syntaxExtraction(initialModelSyntaxFile = initialModel, items = allItems),
    MoreArgs = list(envir = environment())
  )

  # save the external relationships
  vectorModel <- unlist(strsplit(x = initialModel, split = "\\n"))
  externalRelation <- vectorModel[grep("[ ]{0,}(?<!=)~ ", vectorModel, perl = T)]
  factorRelation <- vectorModel[grep("[ ]{0,}~~ ", vectorModel)]
  
  init.model <-
    randomInitialModel(init.model = initialModel,
                       maxItems = numItems,
                       allItems = allItems,
                       initialData = originalData,
                       bifactorModel = bifactor,
                       lavaan.model.specs = lavaan.model.specs)

  best.obj <- all.obj <- current.obj <- criterion(init.model@model.output)
  best.mod <- current.model <- init.model@model.output
  best.syntax <- current.syntax <- init.model@model.syntax
  names(itemsPerFactor) <- factors
  
  convergence <-
    function(x) {
      converge <-
        x@Fit@converged
      se <-
        !any(is.na(x@Fit@se))
      
      ifelse (converge==TRUE & se == TRUE, criterion(x), NA)
    }
  
  if (verbose == TRUE) {
    print("Initial short form selected:")
    cat(current.syntax)
  }

  tabu.list <- vector("list")

  all.syntax <-
    vector(length = niter + 1)
  all.syntax[1] <-
    paste0(current.syntax, collapse = "")

  chk <- Sys.getenv("_R_CHECK_LIMIT_CORES_", "")
  
  if (parallel) {
    if (nzchar(chk) && chk == "TRUE") {
      # use 2 cores in CRAN/Travis/AppVeyor
      num_workers <- 2L
    } else {
      # use all cores in devtools::test()
      num_workers <- parallel::detectCores()
      cl <- parallel::makeCluster(num_workers,type="PSOCK", outfile = "")
      doSNOW::registerDoSNOW(cl)
      `%dopar%` <- foreach::`%dopar%`
    }
  } else {
    num_workers = 1
    `%dopar%` <- foreach::`%do%`
  }
  
  
  
  # Do iterations
  for (it in 1:niter) {
    cat(paste0("\rRunning iteration ", it, " of ", niter, ".   "))
    # Loop through all neighbors
    tmp.obj <- vector("numeric")
    tmp.mod <- vector("list", length(factors))
    tmp.syntax <- vector("list", length(factors))
    currentModelSyntax <-
        unlist(strsplit(current.syntax, split = "\n"))
    for (j in 1:length(factors)) {
      

      currentFactor <-
        factors[j]
      otherFactors <-
        factors[-j]

      currentItems <-
        itemsPerFactor[[ factors[j] ]][
          itemsPerFactor[[ factors[j] ]] %in%
            unlist(
              strsplit(
                grep(paste0(factors[j], ".*=~"), currentModelSyntax, value = T),
                split = " "
              )
            )
        ]
      removedItems <-
        itemsPerFactor[[ factors[j] ]][
          !itemsPerFactor[[ factors[j] ]] %in%
            unlist(
              strsplit(
                grep(paste0(factors[j], ".*=~"), currentModelSyntax, value = T),
                split = " "
              )
            )
        ]

      itemChange <-
        expand.grid(currentItems, removedItems)

      tmp.mod[[j]] <- vector("list", length = nrow(itemChange))
      tmp.syntax[[j]] <- vector("list", length = nrow(itemChange))

      for (k in 1:nrow(itemChange)) {
        newItems <-
          gsub(pattern = paste0(itemChange[k, 1], "\\b"), replacement = itemChange[k, 2], currentItems)
        newCurrentFactor <-
          paste(
            factors[j], "=~",
            paste(newItems, collapse = " + ")
          )

        currentModelSyntax[grepl(paste0(factors[j], ".*=~"), currentModelSyntax)] <-
          newCurrentFactor

        newModelSyntax <-
          paste(stringr::str_trim(currentModelSyntax),
            collapse = " \n"
          )
        tmp.syntax[[j]][[k]] <- newModelSyntax
      }
    }
    
    tmp.syntax <-
      unlist(tmp.syntax)
    model = 0L
    tmp.mod <-
      foreach::foreach(model = 1:length(tmp.syntax), .inorder = T) %dopar% {
        fitmodel <-
          modelWarningCheck(
            lavaan::lavaan(
              model = tmp.syntax[model],
              data = originalData,
              model.type = model.type,
              int.ov.free = int.ov.free,
              int.lv.free = int.lv.free,
              auto.fix.first = auto.fix.first,
              std.lv = std.lv,
              auto.fix.single = auto.fix.single,
              auto.var = auto.var,
              auto.cov.lv.x = auto.cov.lv.x,
              auto.th = auto.th,
              auto.delta = auto.delta,
              auto.cov.y = auto.cov.y,
              ordered = ordered,
              estimator = estimator,
              warn = FALSE
            ),
            newModelSyntax
          )@model.output

      }

    tmp.obj <- 
      sapply(tmp.mod, convergence)
    tmp.mod <- 
      unlist(tmp.mod)
    tmp.tables <-
      lapply(tmp.mod, 
             lavaan::partable)
    # Check which indices result in a valid objective function
    valid <- which(!is.na(tmp.obj))

    # Get just models not on Tabu list
    if (length(tabu.list) > 0) {
      notTabu <- rep(TRUE, length = length(tmp.tables))
      for (i in valid) {
        for (j in 1:length(tabu.list)) {
          notTabu[i] <-
            ifelse(notTabu[i] == FALSE, 
                   FALSE, 
                   !identical(tmp.tables[[i]], 
                         lavaan::partable(tabu.list[[j]])
                         )
            )
        }
      }
      newValid <- notTabu
    } else {
      newValid <- valid
    }

    # Out of valid models, pick model with best objective function value
    indx <- which.min(tmp.obj[newValid])

    # Move current state to next model
    current.obj <- (tmp.obj[newValid])[indx]
    all.obj <- c(all.obj, current.obj)
    current.mod <- (tmp.mod[newValid])[[indx]]
    current.syntax <- (tmp.syntax[newValid])[[indx]]
    all.syntax[it + 1] <- current.syntax

    if (verbose == TRUE) {
      print(current.mod)
      print(current.syntax)
    }

    # Update Tabu list
    tabu.list <- c(current.mod, tabu.list)
    if (length(tabu.list) > tabu.size) {
      tabu.list <- tabu.list[1:tabu.size]
    }

    # Update if the current model is better than the best model
    if (current.obj < best.obj) {
      best.obj <- current.obj
      best.mod <- current.mod
      best.syntax <- current.syntax
      tabu.list <- vector("list") # Clear Tabu list
    }
  }
  
  if (parallel) {
    foreach::registerDoSEQ()
    parallel::stopCluster(cl)
  }

  if (class(best.obj)[1] == "lavaan.vector") {
    temp = as.numeric(best.obj)
    names(temp) = names(best.obj)
    best.obj = temp
  }

  ret <-
    new("TS",
        function_call = match.call(),
        all_fit = all.obj,
        best_fit = best.obj,
        best_model = best.mod,
        best_syntax = best.syntax,
        runtime = Sys.time() - start.time,
        final_tabu_list = tabu.list
    )
  
  ret
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulated_annealing.R
\name{simulatedAnnealing}
\alias{simulatedAnnealing}
\title{An adaptation of the simulated annealing algorithm for psychometric models.}
\usage{
simulatedAnnealing(
  initialModel,
  originalData,
  maxSteps,
  fitStatistic = "cfi",
  temperature = "linear",
  maximize = TRUE,
  Kirkpatrick = TRUE,
  randomNeighbor = TRUE,
  lavaan.model.specs = list(model.type = "cfa", auto.var = TRUE, estimator = "default",
    ordered = NULL, int.ov.free = TRUE, int.lv.free = FALSE, std.lv = TRUE,
    auto.fix.first = FALSE, auto.fix.single = TRUE, auto.cov.lv.x = TRUE, auto.th = TRUE,
    auto.delta = TRUE, auto.cov.y = TRUE),
  maxChanges = 5,
  restartCriteria = "consecutive",
  maximumConsecutive = 25,
  maxItems = NULL,
  items = NULL,
  bifactor = FALSE,
  setChains = 1,
  shortForm = T,
  ...
)
}
\arguments{
\item{initialModel}{The initial model as a \code{character} vector with lavaan model.syntax.}

\item{originalData}{The original \code{data.frame} with variable names.}

\item{maxSteps}{The number of iterations for which the algorithm will run.}

\item{fitStatistic}{Either a single model fit statistic produced by lavaan, or a user-defined fit statistic function.}

\item{temperature}{Either an acceptable \code{character} value, or a user-defined temperature function. The acceptable values are "linear", "quadratic", or "logistic".}

\item{maximize}{Logical indicating if the goal is to maximize (\code{TRUE}) the fitStatistic for model selection.}

\item{Kirkpatrick}{Either \code{TRUE} to use Kirkpatrick et al. (1983) acceptance probability, or a user-defined function for accepting proposed models.}

\item{randomNeighbor}{Either \code{TRUE} to use the included function for randomNeighbor selection, or a user-defined function for creating random models.}

\item{lavaan.model.specs}{A \code{list} which contains the specifications for the
lavaan model. The default values are the defaults for lavaan to perform a
CFA. See \link[lavaan]{lavaan} for more details.}

\item{maxChanges}{An \code{integer} value greater than 1 setting the maximum number of parameters to change within randomNeighbor. When creating a short form, should be no greater than the smallest reduction in items loading on one factor; e.g., when reducing a 2-factor scale from 10 items on each factor to 8 items on the first and 6 items on the second, maxChanges should be no greater than 2.}

\item{restartCriteria}{Either "consecutive" to restart after maxConsecutiveSelection times with the same model chosen in a row, or a user-defined function.}

\item{maximumConsecutive}{A positive \code{integer} value used with restartCriteria.}

\item{maxItems}{When creating a short form, a \code{vector} of the number of items per factor you want the short form to contain. Defaults to \code{NULL}.}

\item{items}{A \code{character} vector of item names. Defaults to \code{NULL}. Ignored if \code{maxItems==FALSE}.}

\item{bifactor}{Logical. Indicates if the latent model is a bifactor model. If \code{TRUE}, assumes that the last latent variable in the provided model syntax is the bifactor (i.e., all of the retained items will be set to load on the last latent variable). Ignored if \code{maxItems==FALSE}.}

\item{setChains}{Numeric. Sets the number of parallel chains to run. Default to \code{1}, which also sets the algorithm to run serially (e.g., on a single processor). Values greater than \code{1} result in the chains running on parallel processes using the \code{doSNOW} and \code{foreach} packages.}

\item{shortForm}{Logical. Are you creating a short form (\code{TRUE}) or not (\code{FALSE})? Default is \code{TRUE}.}

\item{...}{Further arguments to be passed to other functions. Not implemented for any of the included functions.}
}
\value{
A named list: the 'bestModel' found, the 'bestFit', and 'allFit' values found by the algorithm.
}
\description{
Simulated annealing mimics the physical process of annealing metals together. \href{https://www.science.org/doi/10.1126/science.220.4598.671}{Kirkpatrick et al. (1983)} introduces this analogy and demonstrates its use; the implementation here follows this demonstration closely, with some modifications to make it better suited for psychometric models.
}
\details{
\strong{Outline of the Pieces of the Simulated Annealing Algorithm}
\itemize{
\item initialModel -- the initial, full form
\item currentModel -- the model of the current step
\item maxSteps -- the maximum number of steps (iterations)
\item currentStep -- the current step
\item currentTemp -- the current temperature. A function of the number of steps (such that temp = 0 at maxSteps), and values that control the shape of the overall temperature. A part of the function that determines the acceptance probability of newly -- generated models
\item randomNeighbor -- a function that determines how the form is changed at each step. Should be able to change one or more parameters, and should have a way to control how many are changed.
\item goal -- a function that determines the "goodness" of the currentModel. Typically in SA goodness is defined as minimization! Sometimes called an energy function
\item selectionFunction -- a function that determines if a randomNeighbor change is accepted. Uses the goal function that determines the "goodness" of the currentModel and the "goodness" of the randomNeighbor, and the currentTemp to generate a probability of acceptance, then compares this probability to a Uniform(0,1) variable to determine if accepted or not. A standard version of this is:
\figure{SA-goal.jpg}
(Kirkpatrick et al., 1983)
\item bestModel -- the model with the best value of the goal function achieved so far
\item bestGoal -- the best value of the goal function achieved so far
\item restartCriteria -- if utilized, this would "restart" the SA process by changing currentModel to bestModel and continuing the process. Could be based on (1) the currentStep value, (2) the difference between goal(currentModel) and goal(bestModel), (3) randomness (i.e., could randomly restart, could randomly restart based on some values, etc), (4) other criteria.
}
}
\examples{
\dontrun{
data(exampleAntModel)
data(simulated_test_data)
trial1 <- simulatedAnnealing(
  initialModel = lavaan::cfa(
    model = exampleAntModel,
    data = simulated_test_data
  ),
  originalData = simulated_test_data, maxSteps = 3,
  fitStatistic = "rmsea", maximize = FALSE
)
summary(trial1) # shows the resulting model

trial2 <- simulatedAnnealing(
  initialModel = exampleAntModel,
  originalData = simulated_test_data,
  maxSteps = 2, maxItems = 30, items = paste0("Item", 1:56)
)
summary(trial2) # shows the resulting model
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ACO.R
\name{plot,ACO,ANY-method}
\alias{plot,ACO,ANY-method}
\title{Plot method for class `ACO`}
\usage{
\S4method{plot}{ACO,ANY}(x, y, type = c("all", "pheromone", "gamma", "beta", "variance"), ...)
}
\arguments{
\item{x, y}{An S4 object of class `ACO`}

\item{type}{A `character` value specifying the plot type. One of `'all'`
(for all plots), `'pheromone'`, `'gamma'`, `'beta'`, or `'variance'`.}

\item{...}{Not used.}
}
\description{
Plot method for class `ACO`
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Tabu_refit.R
\name{refit.model}
\alias{refit.model}
\title{Given a fitted lavaan model and a search table, refits the model using the search
table as specifying what changes should be done (parameters fixed/freed).}
\usage{
refit.model(fitted.model, ptab)
}
\arguments{
\item{fitted.model}{fitted model of class lavaan}

\item{ptab}{search table}
}
\value{
An object of class \code{lavaan} if the new model fits, or an object of class \code{try-error} if the model update fails.
}
\description{
This is not meant to be called explicitly as \link{tabu.sem} uses this
internally for model refitting.
}
\references{
\doi{10.1080/10705511.2017.1409074}
}
\seealso{
Other Tabu Search:
\code{\link[=add.param]{add.param()}},
\code{\link[=search.prep]{search.prep()}}
}
\author{
Carl F. Falk
}
\concept{Tabu Search}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ShortForm.R
\name{.onAttach}
\alias{.onAttach}
\title{Package Attach Hook Function}
\usage{
.onAttach(lib, pkg)
}
\arguments{
\item{lib}{a character string giving the library directory where the package
defining the namespace was found}

\item{pkg}{a character string giving the name of the package}
}
\description{
Hook triggered when package attached.
}
\details{
Idea taken from https://github.com/ntguardian/MCHT/blob/master/R/StartupMessage.R
}
\examples{
ShortForm:::.onAttach(.libPaths()[1], "ShortForm")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ACO_MPlus.R
\name{antcolony.mplus}
\alias{antcolony.mplus}
\title{A function to implement the ant colony optimization algorithm for short form
specification searches, either using MPlus directly via
\code{\link[base]{system}} calls or using Mplus indirectly with the package
\link[MplusAutomation]{MplusAutomation}.}
\usage{
antcolony.mplus(
  ants = 20,
  evaporation = 0.95,
  mplus = NULL,
  list.items = NULL,
  full = NULL,
  i.per.f = NULL,
  factors = NULL,
  steps = 50,
  max.run = 1000,
  resultfile = NULL,
  summaryfile = "summary.txt",
  min.CFI = 0.95,
  min.TLI = 0.95,
  max.RMSEA = 0.06,
  feedbackfile = "iteration.html",
  loc.gammas,
  loc.variances,
  predictors,
  var.predictors,
  Mplus.Automation = FALSE,
  dataOut = "tempModel.dat",
  modelOut = "tempModel.inp"
)
}
\arguments{
\item{ants}{A numeric value indicating the number of ants to send send (short
forms to evaluate) per iteration. Default value is 20.}

\item{evaporation}{A numeric value which sets the percentage of the pheremone
that is retained after evaporation between steps of the algorithm. Default
value is 0.9, indicating 10% evaporation. Should be within the range of
(0,1), exclusive.}

\item{mplus}{When \code{Mplus.Automation=FALSE}, this is a character value
indicating the name of the MPlus input file without the file extension
".inp". If not in the current working directory, include the full file path
where it is located. This file will be changed during the ant colony search,
so it's suggested to make a backup of the original file before running the
function. When \code{Mplus.Automation=TRUE}, this is an object of class
\link[MplusAutomation]{mplusObject} created by
\link[MplusAutomation]{MplusAutomation} and containing the initial model.}

\item{list.items}{A list containing one or more character vectors of item
names for each factor, where each factor is a separate element of the list.
The items should be input in the order in which the factors are input in
\code{i.per.f} and \code{factors}.}

\item{full}{A numeric value indicating the total number of unique items in the
test or scale.}

\item{i.per.f}{A vector with number of items per factor (e.g. target number),
in the same order of \code{list.items} and \code{factors}.}

\item{factors}{A character vector with the names of the factors in the same
order of \code{list.items} and \code{i.per.f}.}

\item{steps}{A numeric value that sets the stopping rule, which is the number
of ants in a row for which the model does not change.}

\item{max.run}{The maximum number of ants to run before the algorithm stops.
This includes failed iterations as well. Default is 1000.}

\item{resultfile}{A character vector containing the file path where the MPlus
results for the current ant model is saved. If the file is not in the
current working directery, the full path must be specified. Not used when
\code{Mplus.Automation=FALSE}.}

\item{summaryfile}{A character vector containing the name of the summary file
generated. A .txt file is suggested. Default is "summary.txt" and writes
into the current working directory. This file writes a line for each ant
within each step and includes (a) a vector of a 0/1 value for each item
indicating whether the item was selected by that ant, (b) the run number,
(c) the count number, (d) the ant number, and (e) the current pheromone
level.}

\item{min.CFI}{A numeric value indicating the minimum CFI for "acceptable"
model fit. Models with CFI less than this value are automatically rejected.
Default is 0.95.}

\item{min.TLI}{A numeric value indicating the minimum TLI for "acceptable"
model fit. Models with TLI less than this value are automatically rejected.
Default is 0.95.}

\item{max.RMSEA}{A numeric value indicating the maximum RMSEA for "acceptable"
model fit. Models with RMSEA greater than this value are automatically
rejected. Default is 0.06}

\item{feedbackfile}{A character vector containing the name of the feedback
file generated. An .html file is suggested. Default is "iteration.html" and
writes into the current working directory. This file saves the result of
each run, which includes (a) the run number, (b) the count number, (c) the
ant number, (d) the step number (if the current run is successful) or
"Failure" (if the current run is unsuccessful), and for successful runs (f)
the value of CFI, TLI, and RMSEA fit indices, the average of the gammas
(standardized regression coefficients), and the overall variance explained
of the current run.}

\item{loc.gammas}{A numeric vector with the line numbers where the regression
coefficients of the MIMIC model start and end (locations). Not used with
\code{Mplus.Automation=TRUE}}

\item{loc.variances}{A numeric vector with the line numbers of the residual
variances of the latent factors. Not used with \code{Mplus.Automation=TRUE}}

\item{predictors}{Character vector with names of predictor variables, if any.}

\item{var.predictors}{A numeric vector with variances of the predictor(s), if
any. Not used with \code{Mplus.Automation=TRUE}}

\item{Mplus.Automation}{Logical. If \code{TRUE}, uses the
\code{MplusAutomation} package to modify the model as the algorithm
procedes. The "mplus" option will then be used as Defaults to \code{FALSE}
as it is in the process of being built.}

\item{dataOut}{A character vector specifying the location and name of the data
file generated by \code{MplusAutomation} for each iteration of the
algorithm. Default is "tempData.dat" and saves to the current working
directory. When specifying the name, be sure to use a data format that Mplus
can read. You must change the working directory to the location in which
this file will be saved. Only used when \code{Mplus.Automation=TRUE}.}

\item{modelOut}{A character vector specifying the location and name of the
Mplus model file generated by \code{MplusAutomation} for each iteration of
the algorithm. Default is "tempModel.inp" and saves to the current working
directory. When specifying the name of the model file, it must be a ".inp"
extension. You must change the working directory to the location in which
this file will be saved. Only used when \code{Mplus.Automation=TRUE}.}
}
\value{
A named matrix containing final model's best RMSEA, CFI, and TLI
 values, the final pheromone level (the mean of the standardized regression
 coefficients (gammas)), and a series of 0/1 values indicating the items
 selected in the final solution.
}
\description{
The Ant Colony Optimization (ACO) algorithm (Dorigo & Stutzle,
2004) can produce short forms of scales that are optimized with respect to
characteristics selected by the developer, such as model fit and predictive
relationships with other variables. The algorithm is based on the foraging
behavior of a group of ants, which start searching for food in a variety of
directions and then eventually all ants converge to the shortest distance to
the food source. This behavior occurs because ants leave a pheromone trail
behind as they search for food and ants in shorter paths leave stronger
pheromone trails, which are detected by other ants and that will lead them to
follow the shortest trail.
}
\details{
This function sends a specified number of ants
per iteration, which randomly select items to build a model, then evaluates
the model based on pheromone levels. The pheromone levels are updated after
each iteration according to the best-fitting model of that iteration. The
algorithm's stopping rule is to end the search when a certain solution is the
same for a given number of ants in a row. When constructing the mplus dataset
and when \code{Mplus.Automation=FALSE}, make sure that items in 'categorical
are' and 'usevariables' are specifications that take the same number of lines
per short form.

PREPARATORY STEPS: For the ACO algorithm implementation for short
for selection, the following decisions are needed:

1. Determine the target size for the short form.

2. Determine which characteristics should be optimized.

3. Define how the pheromone level will be computed: This is a function of the
characteristics of the short form that will be optimized. In Leite, Huang and
Marcoulides (2008), the pheromone level was zero if model fit indices did
not meet Hu and Bentler's (1999) suggested thresholds, and equal to the sum of
path coefficients of a predictor variable if model fit indices met thresholds.
Currently, the package only implements pheromone calculation based on
regression coefficients or variance explained, with user-selected model fit
index thresholds.

4. Define how many short forms should be evaluated before the best-so-far
pheromone level is examined. Leite, Huang and Marcoulides (2008) used 10 short
forms.

5. Define the percentage of pheromone evaporation, if any. Leite, Huang and
Marcoulides (2008) used 5\%.

6. Define convergence criterion. Leite, Huang and Marcoulides (2008) set the
algorithm to converge if the short form did not improve in 100 x number of
short forms in step 4.

IMPLEMENTATION: Once these decisions are made, the ACO algorithm
selects short forms with the following steps:

Step 1. All items are assigned an initial weight of 1.

Step 2. A set of n short forms is selected by sampling with probability
proportional to the items' weights.

Step 3. Fit latent variable model to the n short forms.

Step 4. Calculate the pheromone levels for the n short forms. Define the
best-so-far pheromone level (if iteration 1) or compare the current best
pheromone from the set of n short forms to the best-so-far pheromone.

Step 5. If the pheromone level of the best short form from step 4 exceeds the
best-so-far pheromone level, update the best-so-far pheromone level and add it
to the current weight of the items of the best short form.

Step 6. Return to step 2 until convergence criterion is reached.
}
\examples{
\dontrun{
# use MplusAutomation to find a 15-item short form of a simulated 56-item unidimensional test
# first, create the list of the items by the factors
# in this case, all items load onto the general 'Ability' factor
list.items <- list(c(
  "Item1", "Item2", "Item3", "Item4", "Item5",
  "Item6", "Item7", "Item8", "Item9", "Item10",
  "Item11", "Item12", "Item13", "Item14", "Item15",
  "Item16", "Item17", "Item18", "Item19", "Item20",
  "Item21", "Item22", "Item23", "Item24", "Item25",
  "Item26", "Item27", "Item28", "Item29", "Item30",
  "Item31", "Item32", "Item33", "Item34", "Item35",
  "Item36", "Item37", "Item38", "Item39", "Item40",
  "Item41", "Item42", "Item43", "Item44", "Item45",
  "Item46", "Item47", "Item48", "Item49", "Item50",
  "Item51", "Item52", "Item53", "Item54", "Item55",
  "Item56"
))
# then, load the data
data(simulated_test_data)

# Create the mplusObject with MplusAutomation
# notice the explicit call of each item, instead of the shorthand "Item1-Item56"
initial.MplusAutomation.model <- MplusAutomation::mplusObject(
  TITLE = "Trial ACO MpluAutomation with FERA 2016 Data;",
  MODEL = "Ability BY Item1 Item2 Item3 Item4 Item5
  Item6 Item7 Item8 Item9 Item10 Item11 Item12
  Item13 Item14 Item15 Item16 Item17 Item18
  Item19 Item20 Item21 Item22 Item23 Item24
  Item25 Item26 Item27 Item28 Item29 Item30
  Item31 Item32 Item33 Item34 Item35 Item36
  Item37 Item38 Item39 Item40 Item41 Item42
  Item43 Item44 Item45 Item46 Item47 Item48
  Item49 Item50 Item51 Item52 Item53 Item54
  Item55 Item56;",
  ANALYSIS = "ESTIMATOR = WLSMV;",
  VARIABLE = "CATEGORICAL = Item1 Item2 Item3 Item4 Item5
  Item6 Item7 Item8 Item9 Item10 Item11 Item12
  Item13 Item14 Item15 Item16 Item17 Item18
  Item19 Item20 Item21 Item22 Item23 Item24
  Item25 Item26 Item27 Item28 Item29 Item30
  Item31 Item32 Item33 Item34 Item35 Item36
  Item37 Item38 Item39 Item40 Item41 Item42
  Item43 Item44 Item45 Item46 Item47 Item48
  Item49 Item50 Item51 Item52 Item53 Item54
  Item55 Item56;",
  OUTPUT = "stdyx;",
  rdata = simulated_test_data
)

# finally, call the function with some minor changes to the default values.
abilityShortForm <- antcolony.mplus(
  ants = 3, evaporation = 0.7,
  mplus = initial.MplusAutomation.model, list.items = list.items, full = 56,
  i.per.f = 15, factors = "Ability", steps = 3, max.run = 50, resultfile = NULL,
  summaryfile = "C:/Users/lordmaxwell/Desktop/summary.txt",
  min.CFI = 0.95, min.TLI = 0.95, max.RMSEA = 0.06,
  feedbackfile = "C:/Users/lordmaxwell/Desktop/iteration.html", Mplus.Automation = TRUE,
  dataOut = "exampleModel.dat",
  modelOut = "exampleModel.inp"
)
}
}
\references{
\doi{10.1080/00273170802285743}
}
\seealso{
\code{\link{antcolony.lavaan}}

Other Ant Colony Algorithms:
\code{\link[=antcolony.lavaan]{antcolony.lavaan()}}
}
\author{
Walter Leite; Anthony W Raborn, \email{anthony.w.raborn@gmail.com}
}
\concept{Ant Colony Algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SA.R
\name{plot,SA,ANY-method}
\alias{plot,SA,ANY-method}
\title{Plot method for class `SA`}
\usage{
\S4method{plot}{SA,ANY}(x, y, ...)
}
\arguments{
\item{x, y}{An S4 object of class `SA`.}

\item{...}{Not used.}
}
\description{
Plot method for class `SA`
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ShortForm.R
\name{ShortFormStartup}
\alias{ShortFormStartup}
\title{Create Package Startup Message}
\usage{
ShortFormStartup()
}
\description{
Makes package startup message.
}
\details{
Idea taken from https://github.com/ntguardian/MCHT/blob/master/R/StartupMessage.R
}
\examples{
ShortForm:::ShortFormStartup()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{exampleAntModel}
\alias{exampleAntModel}
\title{Model syntax for the example in the \code{\link{antcolony.lavaan}} function.}
\format{
A character vector.
}
\usage{
data(exampleAntModel)
}
\description{
A character vector containing the model syntax used for the one factor, 56-item example in the \code{\link{antcolony.lavaan}}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TS.R
\name{summary,TS-method}
\alias{summary,TS-method}
\title{Summary method for class `TS`}
\usage{
\S4method{summary}{TS}(object)
}
\arguments{
\item{object}{An S4 object of class `TS`.}
}
\description{
Summary method for class `TS`
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SA.R
\docType{class}
\name{modelCheck-class}
\alias{modelCheck-class}
\title{An S4 class for the modelCheck object}
\value{
An S4 object of class `ACO`.
}
\description{
An S4 class for the modelCheck object
}
\section{Slots}{

\describe{
\item{\code{model.output}}{A `lavaan` object.}

\item{\code{warnings}}{A `character` vector of any warnings.}

\item{\code{errors}}{A `character` vector of any errors.}

\item{\code{model.syntax}}{A `character` vector of the modelCheck model syntax.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SA.R
\name{show,SA-method}
\alias{show,SA-method}
\title{Print method for class `SA`}
\usage{
\S4method{show}{SA}(object)
}
\arguments{
\item{object}{An S4 object of class `SA`.}
}
\description{
Print method for class `SA`
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Tabu_internal.R
\name{par.matches}
\alias{par.matches}
\title{A series of internal functions used by the Tabu search.
Directly accessing these functions is not advised unless you know what you are doing.}
\usage{
par.matches(ptab, label, block = NULL)
}
\description{
A series of internal functions used by the Tabu search.
Directly accessing these functions is not advised unless you know what you are doing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ACO.R
\docType{class}
\name{ACO-class}
\alias{ACO-class}
\title{An S4 class for the Ant Colony Optimization Algorithm}
\value{
An S4 object of class `ACO`.
}
\description{
An S4 class for the Ant Colony Optimization Algorithm
}
\section{Slots}{

\describe{
\item{\code{function_call}}{The original function call.}

\item{\code{summary}}{A summary `data.frame` indicating the algorithm results for
each iteration.}

\item{\code{final_solution}}{A `matrix` with the final solution information, including
fit indices, selected items, and pheromone level.}

\item{\code{best_model}}{A `lavaan` object of the final solution.}

\item{\code{best_syntax}}{A `character` vector of the final solution model syntax.}

\item{\code{runtime}}{A `difftime` object of the total run time of the function.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ShortForm.R
\name{ShortForm}
\alias{ShortForm}
\alias{ShortForm-package}
\title{\code{ShortForm} package}
\description{
Automated Item Selection Algorithms for Short Forms
}
\details{
See the README on \href{https://github.com/AnthonyRaborn/ShortForm#readme}{GitHub} for more information.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ACO_lavaan.R
\name{antcolony.lavaan}
\alias{antcolony.lavaan}
\title{A function to implement the ant colony optimization algorithm for short form
specification searches with the package \link[lavaan]{lavaan}.}
\usage{
antcolony.lavaan(
  data = NULL,
  sample.cov = NULL,
  sample.nobs = NULL,
  ants = 20,
  evaporation = 0.9,
  antModel,
  list.items = NULL,
  full = NULL,
  i.per.f = NULL,
  factors = NULL,
  bifactor = NULL,
  steps = 50,
  lavaan.model.specs = list(model.type = "cfa", auto.var = T, estimator = "default",
    ordered = NULL, int.ov.free = TRUE, int.lv.free = FALSE, auto.fix.first = TRUE,
    auto.fix.single = TRUE, auto.var = TRUE, auto.cov.lv.x = TRUE, auto.th = TRUE,
    auto.delta = TRUE, auto.cov.y = TRUE, std.lv = F, group = NULL, group.label = NULL,
    group.equal = "loadings", group.partial = NULL, group.w.free = FALSE),
  pheromone.calculation = "gamma",
  fit.indices = c("cfi", "tli", "rmsea"),
  fit.statistics.test = "(cfi > 0.95)&(tli > 0.95)&(rmsea < 0.06)",
  summaryfile = NULL,
  feedbackfile = NULL,
  max.run = 1000,
  parallel = T
)
}
\arguments{
\item{data}{The data being used in data frame format. Default value is
\code{null}. Only one of \code{data} or \code{sample.cov} should be used.}

\item{sample.cov}{The sample covariance matrix. See \link[lavaan]{lavaan} for
the specific format needed. Default value is \code{null}. Only one of
\code{data} or \code{sample.cov} should be used.}

\item{sample.nobs}{A numeric value indicating the number of observations in
the sample covariance matrix. If \code{sample.cov} is used, this must be
filled in. Default value is \code{null}.}

\item{ants}{A numeric value indicating the number of ants to send (e.g.,
number of short forms to evaluate) per iteration. Default value is 20.}

\item{evaporation}{A numeric value which sets the percentage of the pheromone
that is retained after evaporation between steps of the algorithm. Default
value is 0.9, indicating 10% evaporation. Should be within the range of
(0,1), exclusive.}

\item{antModel}{The lavaan formatted model. See \link[lavaan]{lavaan} for more
details. Defaults to the default \link[lavaan]{lavaan} values. NOTE: Each factor
and/or regression needs to be specified on a single line. Newline breaks and
carriage returns WILL break the function.}

\item{list.items}{A list containing one or more character vectors of item
names for each factor, where each factor is a separate element of the list.
The items should be input in the order in which the factors are input in
\code{i.per.f} and \code{factors}.}

\item{full}{A numeric value indicating the total number of unique items in the
test or scale.}

\item{i.per.f}{Vector with number of items per factor (e.g. target number), in
the same order of \code{list.items} and \code{factors}.}

\item{factors}{Character vector with names of factors in the same order of
\code{list.items} and \code{i.per.f}.}

\item{bifactor}{Either the name of the factor that all of the chosen items
will load on (as character), or `NULL` if the model is not a bifactor model.}

\item{steps}{A numeric value that sets the stopping rule, which is the number
of ants in a row for which the model does not change.}

\item{lavaan.model.specs}{A list which contains the specifications for the
lavaan model. The default values are the defaults for lavaan to perform a
CFA. These are automatically set internally, then updated by the user-provided
values. Note that this drastically affects the algorithm, and care must be 
taken to ensure that the algorithm can fit valid models as it searches for
the best model. See the default arguments for examples of what you can change
and \link[lavaan]{lavaan} for more details on what arguments are available 
to change.}

\item{pheromone.calculation}{A character string specifying the method for
calculating the pheromone strength. Must be one of "\code{gamma}"
(standardized latent regression coefficients), "\code{beta}"
(standardized observed regression coefficients), "\code{regression}"
(both latent and observed regression coefficients, if they exist)
 or "\code{variance}" (proportion of
variance explained by model). You must specify the entire string. Default is
\code{gamma}.}

\item{fit.indices}{The fit indices (in lavaan format) extracted for model
optimization. See \link[lavaan]{lavaan} for more details.}

\item{fit.statistics.test}{A character vector of the logical test being used
for model optimization. The default is \code{"(cfi > 0.95)&(tli >
0.95)&(rmsea < 0.06)"}. The format for the logical test should match 1) the
names of the indices being used in \link[lavaan]{lavaan} and 2) the default
provided above. At least one fit index must be included.}

\item{summaryfile}{The name of the summary file generated. A .txt file is
suggested. Default is "summary.txt" and writes into the current working
directory. This file writes a line for each ant within each step and
includes (a) a vector of a 0/1 value for each item indicating whether the
item was selected by that ant, (b) the run number, (c) the count number, (d)
the ant number, and (e) the current pheromone level.}

\item{feedbackfile}{The name of the feedback file generated. An .html file is
suggested. Default is "iteration.html" and writes into the current working
directory. This file saves the result of each run, which includes (a) the
run number, (b) the count number, (c) the ant number, (d) the step number
(if the current run is successful) or "Failure" (if the current run is
unsuccessful), and for successful runs (f) the chosen fit statistics (from
\code{fit.indices}), the average of the gammas and betas (standardized regression
coefficients), and the overall variance explained of the current run.}

\item{max.run}{The maximum number of ants to run before the algorithm stops.
This includes failed iterations as well. Default is 1000.}

\item{parallel}{An option for using parallel processing. If \code{TRUE}, the
function will utilize all available cores (up to the number of ants). Default
is \code{TRUE}.}
}
\value{
A list with four elements: the first containing a named matrix with
 final model's best fit indices, the final pheromone level (either the mean
 of the standardized regression coefficients (gammas, betas, or both), or the mean variance
 explained), and a series of 0/1 values indicating the items selected in the
 final solution,  the second element containing tbe summary matrix of the
 best fit statistic value(s) for each run, the items chosen for said best fit,
 the mean gamma, beta, and variance explained for the best fit, and the item pheromone
 levels after each run, the third containing the best-fitting lavaan model
 object, and the fourth containing the best-fitting model syntax.
}
\description{
The Ant Colony Optimization (ACO) algorithm (Dorigo & Stutzle,
 2004) can produce short forms of scales that are optimized with respect to
 characteristics selected by the developer, such as model fit and predictive
 relationships with other variables. The algorithm is based on the foraging
 behavior of a group of ants, which start searching for food in a variety of
 directions and then eventually all ants converge to the shortest distance to
 the food source. This behavior occurs because ants leave a pheronome trail
 behind as they search for food and ants in shorter paths leave stronger
 pheronome trails, which are detected by other ants and that will lead them
 to follow the shortest trail.
}
\details{
This function sends a specified number of ants per iteration, which
 randomly select items to build a model, then evaluates the model based on
 pheromone levels. The pheromone levels are updated after each iteration
 according to the best-fitting model of that iteration. The algorithm's
 stopping rule is to end the search when a certain solution is the same for a
 given number of ants in a row.

 PREPARATORY STEPS: For the ACO algorithm implementation for short for
 selection, the following decisions are needed:

 1. Determine the target size for the short form.

 2. Determine which characteristics should be optimized.

 3. Define how the pheronome level will be computed: This is a function of
 the characteristics of the short form that will be optimized. In Leite,
 Huang and Marcoulides (2008), the pheromone level was zero if model fit
 indices did not meet Hu and Bentler's (1999) suggested thresholds, and equal
 to the sum of path coefficients of a predictor variable if model fit indices
 met thresholds. Currently, the package only implements pheromone calculation
 based on regression coefficients or variance explained, with user-selected
 model fit index thresholds.

 4. Define how many short forms should be evaluated before the best-so-far
 pheronome level is examined. Leite, Huang and Marcoulides (2008) used 10
 short forms.

 5. Define the percentage of pheronome evaporation, if any. Leite, Huang and
 Marcoulides (2008) used 5\%.

 6. Define convergence criterion. Leite, Huang and Marcoulides (2008) set the
 algorithm to converge if the short form did not improve in 100 x number of
 short forms in step 4.

 IMPLEMENTATION: Once these decisions are made, the ACO algorithm selects
 short forms with the following steps:

 Step 1. All items are assigned an initial weight of 1.

 Step 2. A set of n short forms is selected by sampling with probability
 proportional to the item weights.

 Step 3. Fit the latent variable model to the n short forms.

 Step 4. Calculate the pheromone levels for the n short forms. Define the
 best-so-far pheronome level (if iteration 1) or compare the current best
 pheronome from the set of n short forms to the best-so-far pheronome.

 Step 5. If the pheromone level of the best short form from step 4 exceeds
 the best-so-far pheronome level, update the best-so-far pheromone level and
 add it to the current weight of the items of the best short form.

 Step 6. Return to step 2 until convergence criterion is reached.
}
\examples{
# a 3-factor example using the HolzingerSwineford1939 data from `lavaan`

# some changes to the default values
# notice that in this example we are recreating the original model
abilityShortForm <- antcolony.lavaan(
  data = lavaan::HolzingerSwineford1939,
  ants = 2, evaporation = 0.7,
  antModel = " visual  =~ x1 + x2 + x3
             textual =~ x4 + x5 + x6
             speed   =~ x7 + x8 + x9 ",
  list.items = list(c(
    "x1",
    "x2", "x3"
  ), c("x4", "x5", "x6"), c("x7", "x8", "x9")), full = 9, i.per.f =
    c(3, 3, 3), factors = c("visual", "textual", "speed"), steps = 2, fit.indices =
    c("cfi"), fit.statistics.test = "(cfi > 0.6)", summaryfile =
    NULL, feedbackfile = NULL, max.run = 2, parallel = FALSE
)
\dontrun{
# using simulated test data and the default values for lavaan.model.specs
# first, read in the original or "full" model
data(exampleAntModel) # a character vector for a lavaan model

# then, create the list of the items by the factors
# in this case, all items load onto the general 'Ability' factor
list.items <- list(c(
  "Item1", "Item2", "Item3", "Item4", "Item5",
  "Item6", "Item7", "Item8", "Item9", "Item10",
  "Item11", "Item12", "Item13", "Item14", "Item15",
  "Item16", "Item17", "Item18", "Item19", "Item20",
  "Item21", "Item22", "Item23", "Item24", "Item25",
  "Item26", "Item27", "Item28", "Item29", "Item30",
  "Item31", "Item32", "Item33", "Item34", "Item35",
  "Item36", "Item37", "Item38", "Item39", "Item40",
  "Item41", "Item42", "Item43", "Item44", "Item45",
  "Item46", "Item47", "Item48", "Item49", "Item50",
  "Item51", "Item52", "Item53", "Item54", "Item55", "Item56"
))

# load the data
data(simulated_test_data)

# finally, call the function with some minor changes to the default values.
abilityShortForm <- antcolony.lavaan(
  data = simulated_test_data,
  ants = 5, evaporation = 0.7, antModel = exampleAntModel,
  list.items = list.items, full = 56, i.per.f = 20,
  factors = "Ability", steps = 3, fit.indices = c("cfi", "rmsea"),
  fit.statistics.test = "(cfi > 0.95)&(rmsea < 0.05)",
  summaryfile = "summary.txt",
  feedbackfile = "iteration.html",
  max.run = 500
)

abilityShortForm # print the results of the final short form

# an example using binary (ordered) data
# create the simulated full model and model data
sim_model <- "
f1 =~ x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9 + x10
f2 =~ x11 + x12 + x13 + x14 + x15 + x16 + x17 + x18 + x19 + x20
f3 =~ x21 + x22 + x23 + x24 + x25 + x26 + x27 + x28 + x29 + x30"

sim_data <-
  cbind(
    psych::sim.rasch(nvar = 10)$items,
    psych::sim.rasch(nvar = 10)$items,
    psych::sim.rasch(nvar = 10)$items
  )

colnames(sim_data) = paste0("x", 1:30)
# fit with antcolony.lavaan
# note that ONLY the estimator and ordered args
# of lavaan.model.specs are changed. This retains
# the default args, fitting a CFA but with ordered data.
example <-
antcolony.lavaan(
  data = sim_data,
  ants = 5, evaporation = 0.7,
  antModel = sim_model,
  lavaan.model.specs = 
    list(estimator = "wlsmv", ordered = T),
  list.items = list(paste0('x', 1:10),
                    paste0('x', 11:20),
                    paste0('x', 21:30)), 
  full = 30, 
  i.per.f = c(5, 5, 5), 
  factors = c("f1", "f2", "f3"),
  steps = 20, 
  fit.indices = c("cfi.scaled"), 
  fit.statistics.test = "(cfi.scaled > 0.90)", 
  summaryfile = NULL, 
  feedbackfile = NULL, 
  max.run = 500, 
  parallel = T
)
# note that this example will take a bit of time to run
# as ordered data factor analysis is computationally expensive.
}
}
\seealso{
\code{\link{antcolony.mplus}}

Other Ant Colony Algorithms:
\code{\link[=antcolony.mplus]{antcolony.mplus()}}
}
\author{
Anthony W Raborn, \email{anthony.w.raborn@gmail.com}
}
\concept{Ant Colony Algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Tabu_searchprep.R
\name{search.prep}
\alias{search.prep}
\title{Given a fitted lavaan model (e.g., CFA), prepares a table that contains
parameters that can be fixed/freed as part of a model specification search.}
\usage{
search.prep(fitted.model, loadings = TRUE, fcov = TRUE, errors = FALSE)
}
\arguments{
\item{fitted.model}{- an object of class "lavaan" that contains the initially fitted model for the search}

\item{loadings}{- a logical value that indicates whether cross-loadings will be part of the search}

\item{fcov}{- a logical value indicating whether factor covariances will be part of the search}

\item{errors}{- a logical value indicating whether error covariances will be part of the search}
}
\value{
A \code{data.frame} with lavaan-formatted parameter values.
}
\description{
Given a fitted lavaan model (e.g., CFA), prepares a table that contains
parameters that can be fixed/freed as part of a model specification search.
}
\examples{
\dontrun{
# load simulation data and select columns used in this example
data(simulated_test_data)
tabuData <- simulated_test_data[, c(1:10)]

# specify an improper model (improper because data is unidimensional)
tabuModel <- "
Ability =~ Item1 + Item2 + Item3 + Item4
FakeAbility =~ Item5 + Item6 + Item7 + Item8
Ability ~ Outcome
FakeAbility ~ 0*Outcome"

# run the initial misspecified model for Tabu

init.model <- lavaan::lavaan(
  model = tabuModel, data = tabuData,
  auto.var = TRUE, auto.fix.first = FALSE, std.lv = TRUE, auto.cov.lv.x = TRUE
)

# Use search.prep to prepare for the Tabu search
ptab <- search.prep(fitted.model = init.model, loadings = TRUE, fcov = TRUE, errors = FALSE)

# add an additional (mispecified) parameter
additional.param <- "Item1 ~~ 0.5*Item3"
ptab <- add.param(fitted.model = init.model, ptab = ptab, syntax = additional.param)

# Perform Tabu Search
trial <- tabu.sem(init.model = init.model, ptab = ptab, obj = AIC, niter = 2, tabu.size = 5)
}
}
\references{
\doi{10.1080/10705511.2017.1409074}
}
\seealso{
Other Tabu Search:
\code{\link[=add.param]{add.param()}},
\code{\link[=refit.model]{refit.model()}}
}
\author{
Carl F. Falk
}
\concept{Tabu Search}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{simulated_test_data}
\alias{simulated_test_data}
\title{A simulated data set based on a standardized test.}
\format{
An object of class \code{data.frame} with 1000 rows and 58 columns.
}
\usage{
data(simulated_test_data)
}
\description{
Simulated response patterns, abilities, and outcomes based on a uni-dimensional state-issued standardized test.
}
\details{
@format A data frame of 1000 rows (observations) and 58 columns (variables):
 \describe{
 \item{Outcome}{a binary external criterion variable correlated with TrueAbility}
 \item{TrueAbility}{the simulated true ability parameter used to generate response patterns}
 \item{Item1-Item56}{binary responses to items generated using the TrueAbility parameters and simulated 3PL item parameters generated from the distribution of parameters estimated from a state-issued standardized test}
 }
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TS.R
\docType{class}
\name{TS-class}
\alias{TS-class}
\title{An S4 class for the Tabu Search Algorithm}
\value{
An S4 object of class `TS`.
}
\description{
An S4 class for the Tabu Search Algorithm
}
\section{Slots}{

\describe{
\item{\code{function_call}}{The original function call.}

\item{\code{all_fit}}{A summary `vector` indicating the model fit results for
each iteration.}

\item{\code{best_fit}}{The best model fit result using the selected `fitStatistic`. A numeric value or vector, possibly named.}

\item{\code{best_model}}{A `lavaan` object of the final solution.}

\item{\code{best_syntax}}{A `character` vector of the final solution model syntax.}

\item{\code{runtime}}{A `difftime` object of the total run time of the function.}

\item{\code{final_tabu_list}}{The final list of Tabu models. Each element of the list is a `lavaan` object.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabu_sem_shortform.R
\name{tabuShortForm}
\alias{tabuShortForm}
\title{Short Form Tabu Search}
\usage{
tabuShortForm(
  originalData,
  initialModel,
  numItems,
  criterion = function(x) {
     tryCatch(-lavaan::fitmeasures(object = x, fit.measures =
    "cfi"), error = function(e) Inf)
 },
  niter = 20,
  tabu.size = 5,
  lavaan.model.specs = list(int.ov.free = TRUE, int.lv.free = FALSE, std.lv = TRUE,
    auto.fix.first = FALSE, auto.fix.single = TRUE, auto.var = TRUE, auto.cov.lv.x =
    TRUE, auto.th = TRUE, auto.delta = TRUE, auto.cov.y = TRUE, ordered = NULL,
    model.type = "cfa", estimator = "default"),
  bifactor = FALSE,
  verbose = FALSE,
  parallel = T
)
}
\arguments{
\item{originalData}{The original data frame with variable names.}

\item{initialModel}{The initial model (typically the full form) as a character vector with lavaan model.syntax.}

\item{numItems}{A numeric vector indicating the number of items to retain for each factor.}

\item{criterion}{A function calculating the objective criterion to minimize. Default is to use the built-in `rmsea` value from `lavaan::fitmeasures()`.}

\item{niter}{A numeric value indicating the number of iterations (model specification selections)
to perform. Default is 50.}

\item{tabu.size}{A numeric value indicating the size of Tabu list. Default is 5.}

\item{lavaan.model.specs}{A list which contains the specifications for the
lavaan model. The default values are the defaults for lavaan to perform a
CFA. See \link[lavaan]{lavaan} for more details.}

\item{bifactor}{Logical. Indicates if the latent model is a bifactor model. If `TRUE`, assumes that the last latent variable in the provided model syntax is the bifactor (i.e., all of the retained items will be set to load on the last latent variable).}

\item{verbose}{Logical. If `TRUE`, prints out the initial short form and the selected short form at the end of each iteration.}

\item{parallel}{An option for using parallel processing. If \code{TRUE}, the 
function will utilize all available cores. Default is \code{TRUE}.}
}
\value{
A named list with the best value of the objective function (`best.obj`) and the best lavaan model object (`best.mod`).
}
\description{
Given an initial (full) lavaan model string, the original data, a criterion
function to minimize, and some additional specifications,
performs a Tabu model specification search. Currently only supports
neighbors that are 1 move away from the current model.
}
\examples{
shortAntModel <- "
Ability =~ Item1 + Item2 + Item3 + Item4 + Item5 + Item6 + Item7 + Item8
Ability ~ Outcome
"

data(simulated_test_data)
tabuResult <- tabuShortForm(
  initialModel = shortAntModel,
  originalData = simulated_test_data, numItems = 7,
  niter = 1, tabu.size = 3, parallel = FALSE
)
summary(tabuResult) # shows the resulting model
\dontrun{
# create simulation data from the `psych` package
# four factors, 12 items each, 48 total items
# factor loading matrix - not quite simple structure
fxMatrix <-
  matrix(
    data = c(
      rep(x = c(.9, .7, .5, .3), times = 3),
      rep(0.2, times = 3 * 4 * 3), # first factor loadings

      rep(0.2, times = 3 * 4),
      rep(x = c(.9, .7, .5, .3), times = 3),
      rep(0.2, times = 3 * 4 * 2), # second factor loadings

      rep(0.2, times = 3 * 4 * 2),
      rep(x = c(.9, .7, .5, .3), times = 3),
      rep(0.2, times = 3 * 4), # third factor loadings

      rep(0.2, times = 3 * 4 * 3),
      rep(x = c(.9, .7, .5, .3), times = 3) # fourth factor loadings
    ),
    ncol = 4
  )
# factor correlation matrix - all factors uncorrelated
PhiMatrix <-
  matrix(data = c(
    1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, 1, 0,
    0, 0, 0, 1
  ), ncol = 4)
tabuData <-
  psych::sim(
    fx = fxMatrix,
    Phi = PhiMatrix,
    n = 1000,
    raw = TRUE
  )$observed # observed is the simulated observed data

# NOTE: you must specify the model such that each factor is on a single line!
# otherwise, the algorithm will not work correctly!
tabuModel <- "
Trait1 =~ Item1 + Item2 + Item3 + Item4 + Item5 + Item6 +
Item7 + Item8 + Item9 + Item10 + Item11 + Item12
Trait2 =~ Item13 + Item14 + Item15 + Item16 + Item17 +
Item18 + Item19 + Item20 + Item21 + Item22 + Item23 + Item24
Trait3 =~ Item25 + Item26 + Item27 + Item28 + Item29 + Item30 +
Item31 + Item32 + Item33 + Item34 + Item35 + Item36
Trait4 =~ Item37 + Item38 + Item39 + Item40 + Item41 +
Item42 + Item43 + Item44 + Item45 + Item46 + Item47 + Item48
"

colnames(tabuData) <- paste0("Item", 1:48)
# specify the criterion function that the Tabu Search minimizes
# wrap this in a tryCatch in case a model does not converge!
# specify an appropriate error value: if minimizing, error value must be large
tabuCriterion <- function(x) {
  tryCatch(lavaan::fitmeasures(object = x, fit.measures = "chisq"),
    error = function(e) Inf
  )
}

# use the tabuShortForm function
# reduce form to the best 12 items
tabuShort <- tabuShortForm(
  initialModel = tabuModel, originalData = tabuData,
  numItems = c(3, 3, 3, 3),
  criterion = tabuCriterion,
  niter = 20, tabu.size = 10
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TS.R
\name{plot,TS,ANY-method}
\alias{plot,TS,ANY-method}
\title{Plot method for class `TS`}
\usage{
\S4method{plot}{TS,ANY}(x, y, ...)
}
\arguments{
\item{x, y}{An S4 object of class `TS`.}

\item{...}{Not used.}
}
\description{
Plot method for class `TS`
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Tabu_sem.R
\name{tabu.sem}
\alias{tabu.sem}
\title{Given a fitted lavaan model, a search table, and an objective criterion,
performs a Tabu model specification search. Currently only supports
neighbors that are 1 move away from the current model.}
\usage{
tabu.sem(init.model, ptab, obj, niter = 30, tabu.size = 5)
}
\arguments{
\item{init.model}{initial fitted model of class lavaan}

\item{ptab}{search table (e.g., created by search.prep) that lists candidate
parameters that can be modified as part of the search and how the parameters
can be modified (fixed to what values)}

\item{obj}{objective function to be MINIMIZED. Any function that takes a
lavaan object as the sole argument and returns a numeric value can be used.}

\item{niter}{number of Tabu iterations to perform}

\item{tabu.size}{size of Tabu list}
}
\value{
A list with three elements: best.obj, the numerical value of the best (minimal) objective function achieved; best.mod, the final lavaan model, and best.binvec, a data.frame of the lavaan-formatted parameter table for the final model.
}
\description{
Given a fitted lavaan model, a search table, and an objective criterion,
performs a Tabu model specification search. Currently only supports
neighbors that are 1 move away from the current model.
}
\examples{
# load simulation data and select columns used in this example
data(simulated_test_data)
tabuData <- simulated_test_data[, c(1:10)]

# specify an improper model (improper because data is unidimensional)
tabuModel <- "
Ability =~ Item1 + Item2 + Item3 + Item4
FakeAbility =~ Item5 + Item6 + Item7 + Item8
Ability ~ Outcome
FakeAbility ~ 0*Outcome"

# run the initial misspecified model for Tabu

init.model <- lavaan::lavaan(
  model = tabuModel, data = tabuData,
  auto.var = TRUE, auto.fix.first = FALSE, std.lv = TRUE, auto.cov.lv.x = TRUE
)

# Use search.prep to prepare for the Tabu search
ptab <- search.prep(fitted.model = init.model, loadings = TRUE, fcov = TRUE, errors = FALSE)

# Perform Tabu Search
trial <- tabu.sem(init.model = init.model, ptab = ptab, obj = AIC, niter = 2, tabu.size = 5)
}
\references{
\doi{10.1080/10705511.2017.1409074}
}
\author{
Carl F. Falk
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Tabu_addparam.R
\name{add.param}
\alias{add.param}
\title{Adds a parameter to the given search table. Checks whether parameter
is involved in any (in)equality constraints in a fitted lavaan model}
\usage{
add.param(
  fitted.model,
  ptab,
  syntax,
  nullval = NULL,
  free = NULL,
  block = NULL
)
}
\arguments{
\item{fitted.model}{fitted lavaan model}

\item{ptab}{search table}

\item{syntax}{model.syntax specifying the parameter to add to the current table}

\item{nullval}{optional numeric value specifying what the parameter should be fixed to (when fixed)}

\item{free}{optional logical value specifying whether the parameter should initially be set free (or not)}

\item{block}{optional numeric value specifying the group number to which the parameter corresponds}
}
\value{
A \code{data.frame} with lavaan-formatted parameter values.
}
\description{
Adds a parameter to the given search table. Checks whether parameter
is involved in any (in)equality constraints in a fitted lavaan model
}
\examples{
\dontrun{
# load simulation data and select columns used in this example
data(simulated_test_data)
tabuData <- simulated_test_data[, c(1:10)]

# specify an improper model (improper because data is unidimensional)
tabuModel <- "
Ability =~ Item1 + Item2 + Item3 + Item4
FakeAbility =~ Item5 + Item6 + Item7 + Item8
Ability ~ Outcome
FakeAbility ~ 0*Outcome"

# run the initial misspecified model for Tabu

init.model <- lavaan::lavaan(
  model = tabuModel, data = tabuData,
  auto.var = TRUE, auto.fix.first = FALSE, std.lv = TRUE, auto.cov.lv.x = TRUE
)

# Use search.prep to prepare for the Tabu search
ptab <- search.prep(fitted.model = init.model, loadings = TRUE, fcov = TRUE, errors = FALSE)

# add an additional (mispecified) parameter
additional.param <- "Item1 ~~ 0.5*Item3"
ptab <- add.param(fitted.model = init.model, ptab = ptab, syntax = additional.param)

# Perform Tabu Search
trial <- tabu.sem(init.model = init.model, ptab = ptab, obj = AIC, niter = 2, tabu.size = 5)
}

}
\references{
\doi{10.1080/10705511.2017.1409074}
}
\seealso{
Other Tabu Search:
\code{\link[=refit.model]{refit.model()}},
\code{\link[=search.prep]{search.prep()}}
}
\author{
Carl F. Falk
}
\concept{Tabu Search}
